/* 
 * Copyright (c) 2000 ISP RAS.
 * B.Communisticheskaya,25, Moscow, Russia.
 * All rights reserved.
 *
 */

package jatva.engines;

import jatva.lang.AbstractAction;
import jatva.lang.GraphExplorer;
import jatva.lang.TestingModel;
import jatva.lang.synchronization.TestManager;

import java.util.ArrayList;
import java.util.HashMap;

//import jatva.lang.TestException;

/**
 * Test Engine class for testing deterministic finite state machines
 * with mutually connected state-transition graph 
 */
public class DFSMExplorer extends GraphExplorer
{
  /** the path - contains incompletely tested states */    
  protected ArrayList path = new ArrayList();
  
  protected int pathSize = 0;
  
  /** the table with all tested states information */
  protected HashMap passedStates = new HashMap();
  
  /**
   * Default constructor
   */
  public DFSMExplorer()
  {
  }
  
  public DFSMExplorer(int numberOfNoncriticalErrorsToIgnore)
  {
    super(numberOfNoncriticalErrorsToIgnore);
  }
  
  public DFSMExplorer(  int numberOfNoncriticalErrorsToIgnore
                      , int maximumPathLength
                     )
  {
    super(numberOfNoncriticalErrorsToIgnore, maximumPathLength);
  }
  
  public DFSMExplorer(  int numberOfNoncriticalErrorsToIgnore
                      , int maximumPathLength
                      , boolean useReduction
                     )
  {
    super(numberOfNoncriticalErrorsToIgnore, maximumPathLength, useReduction);
  }
  
  public Object getInitialState()
  {
    if(path.size() == 0) return null;
    else                 return path.get(0);
  }
  
  /**
   * The main method of Test Engine for runing scenario driver
   */
  public void run(TestingModel model)
  {
    TestManager tm = TestManager.getRegistrar();

    try
    {
	    if(!initializeTest())
	    {
	      tm.stateEnter(0, retrieveState(model));
	      return;
	    }
	
	    Object st = null, est = null;
	    AbstractAction a = null;
	    DFSMStateInfo si = null, osi = null;
	    
	    while(true)
	    {
	      // get the current state
	      st = retrieveState(model);
	      
	      //tm.traceMessage("    EXPLORER : state = " + st);
	      
	      // get the state info
	      si = (DFSMStateInfo) passedStates.get(st);
	      
	      //tm.traceMessage("    EXPLORER : stateInfo = " + si);
	      
	      // look for current state in passed ones
	      if(si != null)  // state is passed
	      {
	        tm.stateEnter(si.id, st);
	        
	        if(si.path == null)                      // state is completely tested
	        {
	          tm.traceScenarioMessage("All transitions in this state are tested");
	          
	          if(si.backIndex == -1)
	            notifyOnUnreachableStates(model, est);
	          
	          // restore back arc
	          a = si.back._ts_actionClone();
	          
	          if(!performOldAction(tm, model, a, si.back_id)) return;
	          
	          checkDeterminism(model, est, si.back_id, si.prev);
	        }
	        else                                // state is tested only partially
	        {
	          // correct the last state on the path info if needed
	
	          if(osi != null && osi != si)
	          {
	            if(osi.backIndex == -1 || si.index < osi.backIndex)
	            {
	              for(int i = si.index + 1; i < osi.index; i++)
	              {
	                DFSMStateInfo msi = (DFSMStateInfo)passedStates.get(path.get(i));
	                
	                if(msi.backIndex == -1 || msi.backIndex > si.index)
	                {
	                  msi.backIndex = si.index;
	                  msi.back = msi.path._ts_actionClone();
	                  msi.back_id = msi.path_id;
	                  msi.prev = msi.next;
	                }
	              }
	              
	              osi.backIndex = si.index;
	              osi.back = osi.path._ts_actionClone();
	              osi.back_id = osi.path_id;
	              osi.prev = osi.next;
	            }
	            
	            osi = null;
	          }
	          
	          est = st;
	          
	          if(!st.equals(path.get(path.size()-1))) // state isn't the last on the path
	          {
	            // pass through this state along the path
	            a = si.path._ts_actionClone();
	            
	            if(!performOldAction(tm, model, a, si.path_id)) return;
	            
	            //check the model determinism
	            checkDeterminism(model, est, si.path_id, si.next);
	          }
	          else                              // state is the last on the path
	          {
	            // restore the arc to be passed 
	            a = si.last._ts_actionClone();
	            
	            if((a = model._ts_next(a)) != null)                 // untested arc found
	            {
	              // save this arc as a part of the path
	              si.path = a._ts_actionClone();
	              si.path_id = numberOfTransitions;
	              
	              if(!performNewAction(tm, model, a)) return;              
	              
	              // save the next arc to be passed as last
	              si.last = a._ts_actionClone();
	              
	              // save the next state
	              st = retrieveState(model);
	              
	              si.next = st;
	              
	              // save the state to correct the state info
	              osi = si;
	            }
	            else                            // there are no untested arcs in this state
	            {
	              tm.traceScenarioMessage("All transitions in this state are tested");
	              
	              if(path.size() == 1)          // all was tested
	              {
	                
	                finishReason = allTestedReason;
	                
	                return;
	              }
	              else                          // there are incompletely tested states
	              {
	                tm.traceScenarioMessage("Untested transitions exist for other states");
	
	                // remove this state from the path
	                path.remove(path.size()-1);
	                
	                //pathSize--;
	              
	                // remove the state info
	                si.last = null;
	                si.path = null;
	                si.next = null;
	                
	                // check connectness
	                if(si.backIndex == -1)
	                  notifyOnUnreachableStates(model, est);
	                
	                // restore back arc
	                a = (AbstractAction)si.back.clone();
	          
	                if(!performAction(tm, model, a, si.back_id)) return;
	                
	                checkDeterminism(model, est, si.back_id, si.prev);
	              }
	            }
	          }
	        }
	      }
	      else                                  // state is new
	      {
	        tm.stateEnter(numberOfStates, st);
	        
	        if(!checkStates(tm)) return;
	        
	        // initialize the state
	        a = model._ts_initialize();
	        if(a == null)
	        {
	          finishReason = emptyScenarioReason;
	          return;
	        }

	        // add the state info to the table
	        passedStates.put(st, (si = new DFSMStateInfo(a, numberOfStates++, path.size())));
	        si.path_id = numberOfTransitions;
	        
	        // add the state at the end of the path
	        path.add(st);
	        
	        // trying to pass the first arc
	        if(!performNewAction(tm, model, a)) return ;
	
	        // save the passed arc as last
	        si.last = a._ts_actionClone();
	          
	        // save the next state
	        si.next = retrieveState(model);
	        
	        // save the state to correct the state info
	        osi = si;
	      }
	    }
    }
    finally
    {
      tm.traceScenarioMessage(finishReason);
    }
  }
  
  protected void notifyOnUnreachableStates(TestingModel model, Object st)
  {
    NonconnectedModelException err = new NonconnectedModelException( model.getClass().getName(), st, path.get(0) );
    TestManager.getRegistrar().traceScenarioMessage("All the states reachable from it are completely tested");
    
    throw err;
  }
  
  protected void checkDeterminism(TestingModel model, Object est, int transition_id, Object prev)
  {
    Object st = retrieveState(model);
    
    if(!prev.equals(st))
    {
      NondeterministicModelException err = new NondeterministicModelException( model.getClass().getName(), est, transition_id, prev, st);

      //Tracer.getTracerObject().trace_scenario_info("This test engine can be used for deterministic scenarios only");
      //Tracer.getTracerObject().trace_scenario_info("The transition \"" + transition_id + "\" from \"" + est + "\" has been already tested");
      //Tracer.getTracerObject().trace_scenario_info("Previously it has led to: " + prev);
      //Tracer.getTracerObject().trace_scenario_info("Now it leads to: " + st);  
      
      if(st != null)
      {
        DFSMStateInfo si = (DFSMStateInfo)passedStates.get(st);
        
        if(si == null) TestManager.getRegistrar().traceScenarioMessage("This state has never been met before");  
        else
        {
          TestManager.getRegistrar().stateEnter(si.id , st);
        }
      }
       
      throw err;
    }
  }
  
  static class DFSMStateInfo
  {
    // state identifier
    int id = 0;
    
    // path transition
    AbstractAction path = null;
    
    // path transition identifier
    int path_id = 0;
    
    // last used transition
    AbstractAction last = null;
    
    // next state on the last transition
    Object next = null;

    // back transition
    AbstractAction back = null;
    
    // back transition identifier
    int back_id = 0;
    
    // next state on the back arc
    Object prev = null;
    
    // index of the state in the path ???
    int index = -1;
    
    // index of the state where we can return in the path ???
    int backIndex = -1;
       
    DFSMStateInfo(AbstractAction a, int id, int index)
    {
      path = a._ts_actionClone();
      this.id    = id;
      this.index = index;
    }
  }

}
