/*
 * Copyright (c) 2004 ISP RAS.
 * 109004, B. Communisticheskaya, 25, Moscow, Russia.
 * All rights reserved.
 *
 * $Id: AbstractAction.java,v 1.7 2006/10/05 11:00:32 barancev Exp $
 * Created on 25.02.2005
 *
 */
package jatva.lang;


/** 
 * AbstractAction class is a base class for all objects representing
 * test actions in test scenarios. Each scenario method has associated
 * class inheriting this one for representation of this method's actions. 
 * <p>
 * Action object controls the execution of scenario methods,
 * storing values of iteration variables and state variables.
 * In addition to sole values it also stores auxiliary flags.
 * <p>
 * Flags for an iteration variable are the following.
 * <p>
 * <b>Move flag</b> indicates the fact that the next value
 * of the variable should be used next time. 
 * <p>
 * <b>Last flag</b> indicates the fact that the last possible
 * value of the variable was used last time. </item>
 * <p>
 * If both flags are set, that means that the variable shouble be
 * initialized before its next use.
 * The single flag for a state variable indicates that the variable
 * is not initialized yet.
 * <p>
 * <code>AbstractAction</code> class supports uniform interface for 
 * manipulating flags for all variables and for manipulating action 
 * uncertainty.
 * 
 * @author Victor Kuliamin
 * @since  JavaTESK 2.0
 *
 */
public class AbstractAction 
//  extends DeepSubject
  implements Cloneable
{
  /**
   * Stores the number of scenario method associated to this
   * action object. 
   */
  int _ts_method = 0;

  /**
   * Returns the number of scenario method associated to this action object.
   */
  public int method() 
  {
    return _ts_method;
  }

  /**
   * Auxiliary method for debugging. It should not be used in descendants
   * for tracing.
   *
   * <returns>The representation of flags for all iteration variables.
   */
  public String toString()
  {
    StringBuffer b = new StringBuffer("" + _ts_ivars);
    b.append("{");

    for(int i = 0; i < _ts_ivars; i++)
    {
      if(i != 0) b.append(", ");
      switch((int)((_ts_flags[2*i/CHUNK_SIZE] & (3L << (2*i%CHUNK_SIZE))) >> (2*i%CHUNK_SIZE)))
      {
        case 1:  b.append("L0"); break;
        case 2:  b.append("0M"); break;
        case 3:  b.append("00"); break;
        default: b.append("II"); break;
      }
    }
    b.append("}");

    return b.toString();
  }
  
  /**
   * The size of blocks for in flag storage.
   */
  final static int CHUNK_SIZE = 64;

  /**
   * The mask to cut the values of last-flags.
   */
  final static long LGRID = 0xaaaaaaaaaaaaaaaaL;
  
  /**
   * Stores the number of iteration variables in the associated
   * scenario method.
   */
  protected int _ts_ivars = 0;

  /**
   * Stores flags for all iteration and state variables of the
   * associated scenario method.
   *
   * 
   * <remarks>
   * <para>First bits, from <code>0</code> to <code>2*_ts_ivars-1</code>, store flags for iteration
   * variables. Flags for i-th variable are stored in bits <code>2*i</code> (move-flag)
   * and <code>2*i+1</code> (last-flag). Both flags are stored inverted - bit value <code>0</code> means
   * that the flag is set, <code>1</code> means that the flag is not set.</para>
   * 
   * <para>Bits from <code>2*_ts_ivars</code> till the last used store init-flags for
   * state variables. This flags are stored in direct way - <code>1</code> for set flag,
   * <code>0</code> for removed one.
   * </para>
   */
  protected long[] _ts_flags = null;

  /**
   * Constructor of action object.
   * <param name="method">The number of the associated scenario method in its class.</param>
   * <param name="iterationVariables">The number of iteration variables in the associated scenario method.</param>
   * <param name="stateVariables">The number of state variables in the associated scenario method.</param>
   */
  public AbstractAction(int method, int iterationVariables, int stateVariables)
  {
    _ts_method = (int)method;
    _ts_ivars   = iterationVariables;
    _ts_flags   = (iterationVariables == 0 && stateVariables == 0)?null:new long[(2*iterationVariables+stateVariables+((iterationVariables > 1)?0:-1))/CHUNK_SIZE+1];
  }

  /**
   * The method to copy all the data of AbstractAction object.
   * Used in descendants to make a copy of an action.
   * <param name="r">The action to copy data from.</param>
   */
  protected void _ts_copyBase(AbstractAction r)
  {
    _ts_method = r._ts_method;
    _ts_ivars   = r._ts_ivars;
    _ts_flags   = (r._ts_flags == null)?null:(long[])r._ts_flags.clone();
  }

  /**
   * The method to compare the data fo two AbstractAction objects.
   * Used in descendants to compare two actions.
   * <param name="r">The action to compare with.</param>
   * <returns>True, if the data of two objects are the same.</returns>
   */
  protected boolean _ts_compareBase(AbstractAction r)
  {
    if     (_ts_method != r._ts_method)             return false;
    else if(_ts_ivars != r._ts_ivars)                 return false; 
    else if(_ts_flags == null && r._ts_flags != null) return false;
    else if(_ts_flags == null && r._ts_flags == null) return true;
    else if(r._ts_flags == null || _ts_flags.length != r._ts_flags.length) return false;
    
    for(int i = 0; i < _ts_flags.length; i++)
      if(_ts_flags[i] != r._ts_flags[i]) return false;

    return true;
  }

  /**
   * Copy constructor. Used in the copy constructors of descendants.
   * <param name="r">The action to copy.</param>
   */
  public AbstractAction(AbstractAction r)
  {
    _ts_copyBase(r);
  }

  /**
   * Implements the construction of shallow copy required by
   * template method pattern implemented in 
   * <cref="DeepSubject"><code>DeepSubject</code></cref> class
   * to construct deep copies.
   * <returns>The shallow copy of this object.</returns>
   */
  /*
  public DeepSubject _ts_ShallowClone()
  {
    return new AbstractAction(this);
  }*/
  public Object clone()
  {
    return new AbstractAction(this);
  }

  /**
   * Implements the comparison of field values required by
   * template method pattern implemented in 
   * <cref="DeepSubject"><code>DeepSubject</code></cref> class
   * to provide deep comparison.
   * <param name="obj">The object to compare with.</param>
   * <returns>True, if both object have the same field values.</returns>
   */
  /*
  public boolean     _ts_ShallowEquals (Object obj)
  {
    if(!(obj instanceof AbstractAction)) return false;
    AbstractAction a = (AbstractAction)obj;

    return _ts_CompareBase(a);
  }*/
  public boolean equals(Object o)
  {
    if(o == null) return false;
    if(o == this) return true;
    
    if(!(o instanceof AbstractAction)) return false;
    
    return _ts_compareBase((AbstractAction)o);
  }
  

  /**
   * Implements the assignment of field values required by 
   * template method pattern implemented in 
   * <cref="DeepSubject"><code>DeepSubject</code></cref> class
   * to provide deep assignment.
   * <param name="obj">The object to take values of fields from.</param>
   *//*
  public void        _ts_ShallowAssign (Object obj)
  {
    if(!(obj instanceof AbstractAction)) return;
    AbstractAction a = (AbstractAction)obj;

    _ts_CopyBase(a);
  }*/

  /**
   * Constructs the copy of this actions. Used by test engines to store
   * actions.
   * <returns>The copy of this action.</returns>
   */
  public AbstractAction _ts_actionClone()
  {
    return (AbstractAction)clone();
  }

  /**
   * Checks if the state variable number <code>i</code> is initialized.
   * <param name="i">The number of state variable.</param>
   * <returns>True, if the corresponding flag is set.</returns>
   */
  public boolean isStateVariableInitialized(int i)
  {
    i += 2*_ts_ivars+((_ts_ivars > 1)?1:0);
    int j = i/CHUNK_SIZE;
    i %= CHUNK_SIZE;
    return (_ts_flags[j] & (1L << i)) != 0;
  }
  
  /**
   * Sets the init-flag of the state variable number <code>i</code>.
   * <param name="i">The number of state variable.</param>
   */
  public void setStateVariableInitialized(int i)
  {
    i += 2*_ts_ivars+((_ts_ivars > 1)?1:0);
    int j = i/CHUNK_SIZE;
    i %= CHUNK_SIZE;
    _ts_flags[j] |= (1L << i);
  }

  /**
   * Checks the move-flag of the iteration variable number <code>i</code>.
   * <param name="i">The number of iteration variable.</param>
   * <returns>True, if the move-flag is set for the corresponding variable.</returns>
   */
  public boolean isIterationMove(int i)
  {
    i *= 2;
    int j = i/CHUNK_SIZE;
    i %= CHUNK_SIZE;
    return (_ts_flags[j] & (1L << i)) == 0;
  }

  /**
   * Sets the value of the move-flag of the iteration variable number <code>i</code>.
   * <param name="i">The number of iteration variable.</param>
   * <param name="value">The value to set.</param>
   */
  public void setIterationMove(int i, boolean value)
  {
    i *= 2;
    int j = i/CHUNK_SIZE;
    i %= CHUNK_SIZE;
    if(value) _ts_flags[j] &= ~(1L << i);
    else      _ts_flags[j] |=  (1L << i);
  }

  /**
   * Checks the last-flag of the iteration variable number <code>i</code>.
   * <param name="i">The number of iteration variable.</param>
   * <returns>True, if the last-flag is set for the corresponding variable.</returns>
   */
  public boolean isIterationLast(int i)
  {
    i *= 2;
    i++;
    int j = i/CHUNK_SIZE;
    i %= CHUNK_SIZE;
    return (_ts_flags[j] & (1L << i)) == 0;
  }

  /**
   * Sets the value of the last-flag of the iteration variable number <code>i</code>.
   * <param name="i">The number of iteration variable.</param>
   * <param name="value">The value to set.</param>
   */
  public void setIterationLast(int i, boolean value)
  {
    i *= 2;
    i++;
    int j = i/CHUNK_SIZE;
    i %= CHUNK_SIZE;
    if(value) _ts_flags[j] &= ~(1L << i);
    else      _ts_flags[j] |=  (1L << i);
  }

  /**
   * Checks if the iteration variable number <code>i</code> is initialized.
   * <param name="i">The number of iteration variable.</param>
   * <returns>True, if the corresponding variable is initialized.</returns>
   */
  public boolean isIterationInitialized(int i)
  {
    i *= 2;
    int j = i/CHUNK_SIZE;
    i %= CHUNK_SIZE;
    return (_ts_flags[j] & (3L << i)) != 0;
  }

  /**
   * Private method to set flags in such a way that the next time the <code>i</code>-th
   * iteration variable will be initialized.
   * <param name="i">The number of iteration variable.</param>
   */
  private void setIterationInitialized(int i)
  {
    i *= 2;
    int j = i/CHUNK_SIZE;
    i %= CHUNK_SIZE;
    _ts_flags[j] &= ~(3L << i);
  }

  /**
   * Calculates next values of all flags after the end of scenario method execution
   * on the base of their current values and the values of local flags for 
   * the variables of that method.
   * <param name="f">IterationFlags object storing local flags of iteration
   *  variables after the execution of the associated scenario method.</param>
   */
  public void retrieveFlags(IterationFlags f)
  {
    switch(_ts_ivars)
    {
      case 0: return;
      case 1: 
        setIterationLast(0, f.getLast(0));
        setIterationMove(0, true);
        return;
      default:
        boolean[] e = new boolean[_ts_ivars];
        boolean next = false, l, u;

        //      Chase.Trace.Tracer.TraceUserInfo("e[0] = " + e[0]);
        //      int s = (2*_ts_ivars)/CHUNK_SIZE, t = 2*_ts_ivars%CHUNK_SIZE, i;
        //      for(i = 0; i < s; i++) { _ts_flags[i] &= LGRID; _ts_flags[i] |= (f._ts_flags[i] & LGRID); }
        //      if(t != 0) { _ts_flags[s] &= ~((~LGRID) >> (CHUNK_SIZE - t)); _ts_flags[s] |= ((f._ts_flags[s] & LGRID) >> (CHUNK_SIZE - t)); }

        int i = _ts_ivars-1;
        while(i >= 0)
        {
          if(!e[i])
          {
            //          Chase.Trace.Tracer.TraceUserInfo("Processing " + i);
            l = f.getLast(i);
            u = f.getUsage(i);
            if(u && !l) next = true;
            setIterationLast(i, l);
            if(toMove(i, e, f))    setIterationMove(i, e, f);
            else if(u)             setIterationMove(i, false);
          }
          i--;
        }
        if(!next)
        {
          i = 2*_ts_ivars + 1;
          _ts_flags[i/CHUNK_SIZE] |= (1L << (i%CHUNK_SIZE));
        }
        return;
    }
  }

  /**
   * Private method setting move-flag for <code>i</code>-th iteration variable.
   * In addition sets init-flags for all dependent variables with less numbers.
   * <param name="i">The number of iteration variable.</param>
   * <param name="e">Filtering array.</param>
   * <param name="f">IterationFlags object storing local flags of iteration
   *  variables after the execution of the associated scenario method.</param>
   */
  private void setIterationMove(int i, boolean[] e, IterationFlags f)
  {
//      Chase.Trace.Tracer.TraceUserInfo("  Moving " + i);
    setIterationMove(i, true);
    for(int j = 0; j < i; j++)
    {
      if(!e[j] && !f.getIndependence(i, j))
      {
        setIterationInitialized(j); 
        e[j] = true;
      }
    }
  }

  /**
   * Private method checking if the next value of <code>i</code>-th iteration
   * variable should be used next time.
   * <param name="i">The number of iteration variable.</param>
   * <param name="e">Filtering array.</param>
   * <param name="f">IterationFlags object storing local flags of iteration
   *  variables after the execution of the associated scenario method.</param>
   * <returns></returns>
   */
  private boolean toMove(int i, boolean[] e, IterationFlags f)
  {
//      Chase.Trace.Tracer.TraceUserInfo("  Checking " + i);
    for(int j = 0; j < i; j++)
    {
//        Chase.Trace.Tracer.TraceUserInfo("    Checking " + i + ", " + j);
      if(!e[j] && !f.getIndependence(i, j) && f.getUsage(j) && !f.getLast(j)) return false;
//        if(e[j])                    { Chase.Trace.Tracer.TraceUserInfo("    Skip " + i + ", " + j); continue; }
//        if(f.GetIndependence(i, j)) { Chase.Trace.Tracer.TraceUserInfo("    Independent " + i + ", " + j); continue; }
//        if(!f.GetUsage(j))          { Chase.Trace.Tracer.TraceUserInfo("    Unused " + i + ", " + j); continue; }
//        if(f.GetLast(j))            { Chase.Trace.Tracer.TraceUserInfo("    Last " + i + ", " + j); continue; }

//        Chase.Trace.Tracer.TraceUserInfo("    Relevant, dependent, used and not last " + i + ", " + j);
//        return false;
    }

    return true;
  }


  /**
   * Checks if there exists the next unapplied action corresponding to the same method
   * in the current state. Used by the <code>_ts_Next()</code> method in scenario class.
   * <returns>True, if there are unused values of iteration variables for
   * the associated scenario method.</returns>
   */
  public boolean next()
  {
    switch(_ts_ivars)
    {
      case 0: return false;
      case 1: return (_ts_flags[0] & 3L) != 0;
      default: 
        int i = 2*_ts_ivars+1;
        return (_ts_flags[i/CHUNK_SIZE] & (1L << (i%CHUNK_SIZE))) == 0;
    }
  }

  public static class IterationFlags
  {
    long[] _ts_flags = null;
    int _ts_size = 0;

    public String toString()
    {
      StringBuffer b = new StringBuffer("" + _ts_size);
      b.append("{");
      int i;

      for(i = 0; i < _ts_size; i++)
      {
        if(i != 0) b.append(", ");
        if(getLast(i))  b.append('L');
        else            b.append('0');
        if(getUsage(i)) b.append('U');
        else            b.append('0');
  /*        switch((_ts_flags[2*i/CHUNK_SIZE] & (3L << (2*i%CHUNK_SIZE))) >> (2*i%CHUNK_SIZE))
          {
            case 1:  b.Append("LU"); break;
            case 2:  b.Append("00"); break;
            case 3:  b.Append("0U"); break;
            default: b.Append("L0"); break;
          }*/
      }
      b.append("}{");
      for(i = 0; i < _ts_size; i++)
        for(int j = 0; j < i; j++)
          b.append("[" + i + ", " + j + "]:" + get(i, j));
      b.append("}");

      return b.toString();
    }

    public IterationFlags()
    {}

    public IterationFlags(AbstractAction a)
    {
      _ts_size = a._ts_ivars;
      if(_ts_size > 0) _ts_flags = new long[(2*_ts_size + _ts_size*(_ts_size-1)/2-1)/CHUNK_SIZE+1];

      int s = (2*_ts_size)/CHUNK_SIZE, t = 2*_ts_size%CHUNK_SIZE;
//        Chase.Trace.Tracer.TraceUserInfo("s = " + s);
//        Chase.Trace.Tracer.TraceUserInfo("t = " + t);
      for(int i = 0; i < s; i++) _ts_flags[i] = a._ts_flags[i] & LGRID;

//        Chase.Trace.Tracer.TraceUserInfo("flags[t] = " + (a._ts_flags[s] & (LGRID >> (CHUNK_SIZE - t))));
      if(t != 0) _ts_flags[s] = a._ts_flags[s] & (LGRID >> (CHUNK_SIZE - t));
    }

    public void setUsage(int i)
    {
      i *= 2;
      int j = i/CHUNK_SIZE;
      i %= CHUNK_SIZE;

      _ts_flags[j] |= (1L << i);
    }

    public boolean getUsage(int i)
    {
      i *= 2;
      int j = i/CHUNK_SIZE;
      i %= CHUNK_SIZE;

      return (_ts_flags[j] & (1L << i)) != 0;
    }

    public void setLast(int i, boolean value)
    {
      i *= 2;
      i++;
      int j = i/CHUNK_SIZE;
      i %= CHUNK_SIZE;

      if(value) _ts_flags[j] &= ~(1L << i);
      else      _ts_flags[j] |=  (1L << i);
    }

    public boolean getLast(int i)
    {
      i *= 2;
      i++;
      int j = i/CHUNK_SIZE;
      i %= CHUNK_SIZE;

      return (_ts_flags[j] & (1L << i)) == 0;
    }

    public void setIndependence(int i, int j, boolean value)
    {
      if(i == j) return;
    
      int t;
      if(i < j) { t = i; i = j; j = t; }
      t = 2*_ts_size + i*(i-1)/2+j;

      i = t/CHUNK_SIZE;
      j = t%CHUNK_SIZE;
      if(value) _ts_flags[i] |= (1L << j);
      else      _ts_flags[i] &= ~(1L << j);
    }

    public boolean getIndependence(int i, int j)
    {
      if(i == j) return false;

      int t;
      if(i < j) { t = i; i = j; j = t; }
      t = 2*_ts_size + i*(i-1)/2+j;

      return (_ts_flags[t/CHUNK_SIZE] & (1L <<  (t%CHUNK_SIZE))) != 0;
    }

    public boolean get(int i, int j)
    {
      return getIndependence(i, j);
    }
    
    public void set(int i, int j, boolean value)
    {
      setIndependence(i, j, value);
    }

    public void clear()
    {
      for(int i = 0; i < _ts_flags.length; i++) _ts_flags[i] = 0;
      _ts_size = 0;
    }

    public void fill()
    {
      for(int i = 0; i < _ts_flags.length; i++) _ts_flags[i] = ~0L;
    }

    public boolean isEmpty()
    {
      for(int i = 0; i < _ts_flags.length; i++)
        if(_ts_flags[i] != 0) return false;

      return true;
    }

    public boolean equals(Object o)
    {
      if(!(o instanceof IterationFlags)) return false;
      IterationFlags r = (IterationFlags)o;
      if(_ts_size != r._ts_size) return false;
      if(r == this)        return true;
            
      for(int i = 0; i < _ts_flags.length; i++)
        if(_ts_flags[i] != r._ts_flags[i]) return false;

      return true;
    }

    public int hashCode()
    {
      int res = 0;

      for(int i = 0; i < _ts_flags.length; i++)
        res += ((int)(_ts_flags[i] >> CHUNK_SIZE/2) + (int)(_ts_flags[i] & 0xffffffff));
    
      return res;
    }

    public IterationFlags(IterationFlags r)
    {
      _ts_size = r._ts_size;
      if(_ts_size > 0)
      {
        _ts_flags = new long[r._ts_flags.length];
        System.arraycopy(r._ts_flags, 0, _ts_flags, 0, _ts_flags.length);
      }
    }

    public Object clone()
    {
      return new IterationFlags(this);
    }
  }
}

