/* 
 * Copyright (c) 2000 ISP RAS.
 * B.Communisticheskaya,25, Moscow, Russia.
 * All rights reserved.
 *
 * $Id: GraphExplorer.java,v 1.8 2007/08/28 12:04:30 kuliamin Exp $
 * Created on 23.03.2001
 */

package jatva.lang;

import jatva.lang.synchronization.TestManager;

/**
 * GraphExplorer class is the base class of classes representing 
 * graph exploration algorithms.
 * The main method of such a class <code>run</code> should be
 * overriden to implement specific algorithm.  
 *
 * @author Victor Kuliamin
 * @version 2.0      
 */
public class GraphExplorer
{
  /**
   * Default number of noncritical errors that are ignored
   */
  protected final static int defaultNumberOfNoncriticalErrorsToIgnore = 20;
  protected final static int defaultMaximumPathLength                 = 15000;
  protected final static int defaultMaximumNumberOfStates             = 300;
  protected final static int defaultMaximumNumberOfTransitions        = 3000;
  
  protected final static String allTestedReason             = "Test is complete. All available transitions are tested";
  protected final static String tooLongPathReason           = "Test is stopped because the constructed path's length exceeds the maximum allowed";
  protected final static String criticalErrorReason         = "Test is stopped after a critical error";
  protected final static String tooManyErrorsReason         = "Test is stopped because the number of errors detected exceeds the maximum allowed";
  protected final static String tooManyStatesReason         = "Test is stopped because the number of different states found exceeds the maximum allowed";
  protected final static String tooManyTransitionsReason    = "Test is stopped because the number of different transitions found exceeds the maximum allowed";
  protected final static String stateCalculationErrorReason = "Test is stoped because of an exception during state calculation";
  protected final static String reductionErrorReason        = "Test is stopped because of an exception in reduction method";
  protected final static String emptyScenarioReason         = "Test is stopped because the scenario used has no scenario methods";
  
  /**
   * Number of noncritical errors that are ignored
   */
  protected int numberOfNoncriticalErrorsToIgnore = defaultNumberOfNoncriticalErrorsToIgnore;
  protected int maximumPathLength                 = defaultMaximumPathLength;
  protected int maximumNumberOfStates             = defaultMaximumNumberOfStates;
  protected int maximumNumberOfTransitions        = defaultMaximumNumberOfTransitions;
  
  protected boolean useReduction                  = false;
  
  /**
   * Current number of noncritical errors
   */
  protected int numberOfNoncriticalErrors = 0;
  protected int currentPathLength         = 0;
  protected int numberOfStates            = 0;
  protected int numberOfTransitions       = 0;
  
  protected boolean hasStarted            = false;
  protected String  finishReason          = allTestedReason;      
  
  public GraphExplorer()
  {
  }
  
  public GraphExplorer(int numberOfNoncriticalErrorsToIgnore)
  {
    this.numberOfNoncriticalErrorsToIgnore = numberOfNoncriticalErrorsToIgnore;
  }
  
  public GraphExplorer(  int numberOfNoncriticalErrorsToIgnore
                       , int maximumPathLength
                      )
  {
    this.numberOfNoncriticalErrors = numberOfNoncriticalErrorsToIgnore;
    this.maximumPathLength         = maximumPathLength;
  }
  
  public GraphExplorer(  int numberOfNoncriticalErrorsToIgnore
                       , int maximumPathLength
                       , boolean useReduction
                      )
  {
    this.numberOfNoncriticalErrors  = numberOfNoncriticalErrorsToIgnore;
    this.maximumPathLength          = maximumPathLength;
    this.useReduction               = useReduction;
  }

  /**
   * The main method of graph exploration.
   * In this class simply calls scenario methods one by one.
   * <p>
   * Should be overrden to implement another graph exlploration algorithm.
   * 
   * @param model   scenario to be executed. Represents 
   *                a graph to be explored. 
   */
  public void run(TestingModel model)
  {
    TestManager tm = TestManager.getRegistrar();
    
    try
    {  
      tm.stateEnter(0, "Single");
      
      if(!initializeTestPrivate()) return;
      
      // initialize the scenario
       AbstractAction a = model._ts_initialize();
       if(a == null)
       {
       	 finishReason = emptyScenarioReason;
       	 return;
       }
      
      do
      { 
        if(!performActionPrivate(tm, model, a)) return;
        
        if(!tm.transitionPerformed()) 
        { 
          /*tm.traceScenarioMessage("EXPLORER: FAKED TRANSITION ");*/
          currentPathLength--; 
        }
        else
        {
          /*tm.traceScenarioMessage("EXPLORER: ACTUAL TRANSITION ");*/
          if(useReduction) performReduction(tm, model);

          tm.stateEnter(0, "Single");
        }
      }
      while((a = model._ts_next(a)) != null);
      
      finishReason = allTestedReason;
    }
    finally
    {
      tm.traceScenarioMessage(finishReason);
    }
  }
  
  private boolean initializeTestPrivate()
  {
    hasStarted = true;
    
    if(maximumPathLength == 0)
    {
      finishReason = tooLongPathReason + " (0)";
      return false;
    }
    return true;
  }

  private boolean performActionPrivate(TestManager m, TestingModel model, AbstractAction a)
  {
    if(maximumPathLength >= 0 && currentPathLength >= maximumPathLength)
    {
      finishReason = tooLongPathReason + " (" + maximumPathLength + ")";
      return false;
    }
    
    m.transitionStart(currentPathLength++, model, a);
        
    // trying to pass the arc
    try
    {
      model._ts_execute(a);
      return true;
    }
    catch(TestException error)
    {
      checkException(m, error);
      return true;
    }
    finally
    {
      m.transitionEnd();
    }
  }
  
  protected void performReduction(TestManager m, TestingModel model)
  {
    m.traceScenarioMessage("Performing reduction");
    try
    {
      model.reduce();
    }
    catch(TestException error)
    {
      finishReason = reductionErrorReason;
      throw error;
    }
    catch(Throwable error)
    {
      finishReason = reductionErrorReason;
      throw new ReductionException(model, error);
    }
    finally
    {
      m.traceScenarioMessage("Reduction complete");
    }
  }
  
  protected boolean initializeTest()
  {
    hasStarted = true;
    
    if(maximumPathLength == 0)
    {
      finishReason = tooLongPathReason + " (0)";
      return false;
    }
    else if(maximumNumberOfTransitions == 0)
    {
      finishReason = tooManyTransitionsReason + " (0)";
      return false;
    }
    return true;
  }
  
  public Object getInitialState()
  {
    return "Single";
  }
  
  protected boolean performNewAction(TestManager m, TestingModel model, AbstractAction a)
  {
    if(maximumPathLength >= 0 && currentPathLength++ >= maximumPathLength)
    {
      finishReason = tooLongPathReason + " (" + maximumPathLength + ")";
      m.traceScenarioMessage(finishReason);
      return false;
    }
    else if(maximumNumberOfTransitions >= 0 && numberOfTransitions >= maximumNumberOfTransitions)
    {
      finishReason = tooManyTransitionsReason + " (" + maximumNumberOfTransitions + ")";
      m.traceScenarioMessage(finishReason);
      return false;
    }
    
    return performAction(m, model, a, numberOfTransitions++);
  }
  
  protected boolean performOldAction(TestManager m, TestingModel model, AbstractAction a, int id)
  {
    if(maximumPathLength >= 0 && currentPathLength++ >= maximumPathLength)
    {
      finishReason = tooLongPathReason + " (" + maximumPathLength + ")";
      m.traceScenarioMessage(finishReason);
      return false;
    }
    
    return performAction(m, model, a, id);
  }
  
  protected boolean performAction(TestManager m, TestingModel model, AbstractAction a, int id)
  {
    m.transitionStart(id, model, a);
        
    // trying to pass the arc
    try
    {
      model._ts_execute(a);
      return true;
    }
    catch(TestException error)
    {
      checkException(m, error);
      return true;
    }
    finally
    {
      m.transitionEnd();
    }
  }
  
  protected boolean checkStates(TestManager m)
  {
    if(maximumNumberOfStates >= 0 && numberOfStates >= maximumNumberOfStates)
    {
      finishReason = tooManyStatesReason + " (" + maximumNumberOfStates + ")";
      return false;
    }
    return true;
  }
  
  protected void checkException(TestManager m, TestException error)
  {
    if(error.isCritical())
    {
      finishReason = criticalErrorReason;
      throw error;
    }
    
    if(    numberOfNoncriticalErrorsToIgnore >= 0
        && ++numberOfNoncriticalErrors > numberOfNoncriticalErrorsToIgnore
      )
    {
      finishReason = tooManyErrorsReason; 
      throw error;
    }
  }
  
  protected Object retrieveState(TestingModel model)
  {
    try
    {
      //TestManager.getRegistrar().traceMessage("    EXPLORER STATE : " + model._ts_state());
      
      return model._ts_state();
    }
    catch(TestException error)
    {
      finishReason = stateCalculationErrorReason;
      throw error;
    }
    catch(Throwable error)
    {
      finishReason = stateCalculationErrorReason;
      throw new GenStateCalculationException(model, error);
    }
  }

  /**
   * @param maximumNumberOfStates The maximumNumberOfStates to set.
   */
  public void setMaximumNumberOfStates (int maximumNumberOfStates)
  {
    if(!hasStarted)
    {
      this.maximumNumberOfStates = maximumNumberOfStates;
    }
  }

  /**
   * @param maximumNumberOfTransitions The maximumNumberOfTransitions to set.
   */
  public void setMaximumNumberOfTransitions (int maximumNumberOfTransitions)
  {
    if(!hasStarted)
    {
      this.maximumNumberOfTransitions = maximumNumberOfTransitions;
    }
  }

  /**
   * @param maximumPathLength The maximumPathLength to set.
   */
  public void setMaximumPathLength (int maximumPathLength)
  {
    if(!hasStarted)
    {
      this.maximumPathLength = maximumPathLength;
    }
  }

  /**
   * @param numberOfNoncriticalErrorsToIgnore The numberOfNoncriticalErrorsToIgnore to set.
   */
  public void setNumberOfNoncriticalErrorsToIgnore (int numberOfNoncriticalErrorsToIgnore)
  {
    if(!hasStarted)
    {
      this.numberOfNoncriticalErrorsToIgnore = numberOfNoncriticalErrorsToIgnore;
    }
  }

  /**
   * @param useReduction The useReduction to set.
   */
  public void setUseReduction (boolean useReduction)
  {
    if(!hasStarted)
    {
      this.useReduction = useReduction;
    }
  }

  /**
   * @return Returns the finishReason.
   */
  public String getFinishReason ()
  {
    return finishReason;
  }
}