/*
 * Copyright (c) 2004 ISP RAS.
 * 109004, B. Communisticheskaya, 25, Moscow, Russia.
 * All rights reserved.
 *
 * $Id: ModelEvent.java,v 1.24 2008/08/20 12:56:12 almo Exp $
 * Created on 11.02.2005
 *
 */
package jatva.lang;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import jatva.lang.coverage.SEOperationInfo;
import jatva.lang.synchronization.TestManager;
import jatva.lang.synchronization.SynchronizationObjectInterface;
import jatva.util.IdentityList;
import jatva.util.Utils;

/**
 * @author Victor Kuliamin
 *
 */
public class ModelEvent
{
  private static int counter = 0;
  private        int id;

  protected TestManager   manager   = null;
  protected Object        channelId = null;
  protected ModelObject   object    = null;
  protected SEOperationInfo operationInfo = null;
  protected Object[]      arguments = null;
  protected Object        result    = null;

  private IdentityList    synchObjects = new IdentityList();

  protected Throwable exception;
  protected Throwable implException;
  private boolean implExceptionFlag;

  private ModelEvent returnEvent;
  private long time;

  /**
   * Used to determine at what level of refining oracles invariant checks should
   * be triggered.
   * This counter is set to default value (operationInfo.classLevel) in the beginning of each
   * 'check' operation.
   * Also it's reduced at the moment of each partial precondition failure but only if no
   * refining preconditions completed successfully.
   */
  private int invariantCheckClassLevel;

  /**
   * @param object      Object, in which the event occurs
   * @param operationInfo   Operation corresponding to teh event
   * @param arguments   Arguments (data) of the event
   */
  public ModelEvent( ModelObject object
                   , SEOperationInfo operationInfo
                   , Object[] arguments
                   )
  {
    synchronized( ModelEvent.class )
    {
        this.id = counter ++;
    }
    this.object    = object;
    this.operationInfo = operationInfo;
    this.arguments = arguments;
  }

  public String toString( boolean printChannelId )
  {
    StringBuffer s = new StringBuffer(20);

    if( printChannelId )
    {
        s.append( channelId );
        s.append( ": " );
    }
    s.append(id);
    s.append(": ");
    s.append(Utils.getSimpleString(object));
    s.append(": ");
    s.append(operationInfo.toString());

    return s.toString();
  }

  @Override
public String toString()
  {
      return toString( false );
  }

  public void execute()
  {
    Method imth = null;
    try
    {
      imth = operationInfo.getImplementationMethod();
      imth.setAccessible(true);

      if( imth.getReturnType() == Void.TYPE )
        imth.invoke( object, arguments );
      else
        result = imth.invoke( object, arguments );
    }
    catch( IllegalAccessException error )
    {
      throw new InsufficientAccessibilityError(   "Method is inaccessible through reflection"
                                                , error
                                                , new String[]{"class", "method"}
                                                , new String[]{object.getClass().getName(), imth.toString()}
                                              );
    }
    catch( IllegalArgumentException error )
    {
      throw new TranslationError(  "Incorrect translation of event operation name or arguments"
                                 , new String[]{"object", "called method", "arguments"}
                                 , new String[]{object.toString(), imth.toString(), Utils.getArraySimpleString(arguments)}
                                );
    }
    // exceptions from implementation
    catch( ExceptionInInitializerError error )
    {
      Throwable ex = error.getCause();
      if(ex == null)
      {
        throw new LostExceptionError(  error
                                     , new String[]{"object", "called method", "arguments"}
                                     , new String[]{object.toString(), imth.toString(), Utils.getArraySimpleString(arguments)}
                                    );
      }
      else if (ex instanceof TestException)
      {
        throw (TestException)ex;
      }
      else exception = ex;
    }
    catch( InvocationTargetException error )
    {
      Throwable ex = error.getCause();
      if(ex == null)
      {
        throw new LostExceptionError(  error
                                     , new String[]{"object", "called method", "arguments"}
                                     , new String[]{object.toString(), imth.toString(), Utils.getArraySimpleString(arguments)}
                                    );
      }
      else if (ex instanceof TestException)
      {
        throw (TestException)ex;
      }
      else exception = ex;
    }
  }

  public boolean check()
  {
    if( object._ts_isOracleSet() )
    {

      getObject()._ts_removeAnyPreHolds();
      getObject()._ts_removeAnyPostFails();
      preVerdictFailureComment = null;
      postVerdictFailureComment = null;

      this.invariantCheckClassLevel = operationInfo.getClassLevel();

      Method omth = operationInfo.getOracleMethod();
      try
      {
        omth.setAccessible(true);
        return ((Boolean)omth.invoke( object, arguments )).booleanValue();
      }
      catch( IllegalAccessException error )
      {
        throw new InsufficientAccessibilityError(   "Method is inaccessible through reflection"
                                                  , error
                                                  , new String[]{"class", "method"}
                                                  , new String[]{object.getClass().getName(), omth.toString()}
                                                );
      }
      catch( IllegalArgumentException error )
      {
        throw new TranslationError(  "Incorrect translation of event operation name or arguments"
                                   , new String[]{"object", "called method", "arguments"}
                                   , new String[]{object.toString(), omth.toString(), Utils.getArraySimpleString(arguments)}
                                  );
      }
      catch( ExceptionInInitializerError error )
      {
        throw new RuntimeTestException(   "Exception in static initializer of specification class"
                                        , error.getCause()
                                        , new String[]{"class"}
                                        , new String[]{object.getClass().getName()}
                                       );
      }
      // exceptions from oracle
      catch( InvocationTargetException error )
      {
        Throwable ex = error.getCause();
        if(ex == null)
        {
          throw new LostExceptionError(  error
                                       , new String[]{"object", "called method", "arguments"}
                                       , new String[]{object.toString(), omth.toString(), Utils.getArraySimpleString(arguments)}
                                      );
        }
        else if (ex instanceof TestException)
        {
          throw (TestException)ex;
        }
        else
        {
          throw new TestSystemInternalError(  "Incorrect exception created during oracle execution"
                                            , ex
                                            , new String[]{"object", "called method", "arguments"}
                                            , new String[]{object.toString(), omth.toString(), Utils.getArraySimpleString(arguments)}
                                           );
        }
      }
    }
    else return true;
  }

  public void update()
  {
    for(int i = 0; i < synchObjects.size(); i++)
        ((SynchronizationObjectInterface)synchObjects.get(i)).update();
  }

  public void addSynchronizationObject(SynchronizationObjectInterface o)
  {
    if( o != null )
        synchObjects.addLast(o);
  }

  public void storeResult(Object o)
  {
    result = o;
  }

  /**
   * @return Returns the manager.
   */
  public TestManager getManager ()
  {
    return manager;
  }
  /**
   * @param manager The manager to set.
   */
  public void setManager (TestManager manager)
  {
    this.manager = manager;
  }
  /**
   * @return Returns the arguments.
   */
  public Object[] getArguments ()
  {
    return arguments;
  }
  /**
   * @return Returns the object.
   */
  public ModelObject getObject ()
  {
    return object;
  }
  /**
   * @return Returns the operationInfo.
   */
  public SEOperationInfo getOperation ()
  {
    return operationInfo;
  }

  /**
   * @return    Returns the exception.
   */
  public Throwable getException ()
  {
    return exception;
  }

  /**
   * @param exception    The exception to set.
   */
  public void setException (Throwable exception)
  {
    this.exception = exception;
  }

  /**
   * @return Returns the implExceptionFlag.
   */
  public boolean isImplException ()
  {
    return implExceptionFlag;
  }

  /**
   * @param implExceptionFlag The implExceptionFlag to set.
   */

  public void setImplExceptionFlag ()
  {
    this.implExceptionFlag = true;
  }

  public void setImplException(Throwable e)
  {
    implException = e;
  }

  public Throwable getImplException()
  {
    return implException;
  }


  public Object getResult()
  {
    return result;
  }

  /**
   * @return Returns the returnEvent.
   */
  public ModelEvent getReturnEvent ()
  {
    return returnEvent;
  }

  /**
   * @param returnEvent The returnEvent to set.
   */
  public void setReturnEvent (ModelEvent returnEvent)
  {
    this.returnEvent = returnEvent;
    //result = returnEvent.arguments[0];
  }

  /**
   * @return    Returns the time.
   */
  public long getTime ()
  {
    return time;
  }

  /**
   * @param time    The time to set.
   */
  public void setTime (long time)
  {
    this.time = time;
  }
  public int getId()
  {
    return id;
  }
  public Object getChannelId()
  {
      return channelId;
  }
  public void setChannelId( Object channelId )
  {
      this.channelId = channelId;
  }

  public int getInvariantCheckClassLevel()
  {
      return invariantCheckClassLevel;
  }

  public void reducedInvariantCheckClassLevel()
  {
      invariantCheckClassLevel --;
  }
//
//  private boolean bounded = false;
//  public boolean isBounded()
//  { return bounded; }
//
//  public void bound()
//  { bounded = true; }
//
//  public void unbound()
//  { bounded = false; }

  StringBuffer preVerdictFailureComment = null;
  public void addPreVerdictFailureComment( Object fc )
  {
      if( preVerdictFailureComment == null )
          preVerdictFailureComment = new StringBuffer();
      else
          preVerdictFailureComment.append( Utils.endl );

      preVerdictFailureComment.append( "" + fc );
  }

  public String getPreVerdictFailureComment()
  {
      if( preVerdictFailureComment == null )
          return null;

      String res = preVerdictFailureComment.toString();
      preVerdictFailureComment = null;
      return res;
  }

  StringBuffer postVerdictFailureComment = null;
  public void addPostVerdictFailureComment( Object fc )
  {
      if( postVerdictFailureComment == null )
          postVerdictFailureComment = new StringBuffer();
      else
          postVerdictFailureComment.append( Utils.endl );

      postVerdictFailureComment.append( "" + fc );
  }

  public String getPostVerdictFailureComment()
  {
      if( postVerdictFailureComment == null )
          return null;

      String res = postVerdictFailureComment.toString();
      postVerdictFailureComment = null;
      return res;
  }


}
