/*
 * Copyright (c) 2000 ISP RAS.
 * B.Communisticheskaya,25, Moscow, Russia.
 * All rights reserved.
 *
 */

package jatva.lang;

import jatva.lang.coverage.SEOperationInfo;
import jatva.lang.services.DeepSubject;
import jatva.lang.synchronization.SynchronizationManager;

/**
 * The <code>ModelObject</code> class is the base class of all model classes.
 * <p>
 * It provides default implementations for all the methods of model objects.
 * <p>
 * This class can be used to represent objects with no details of internal
 * structure or objects of Java <code>Object</code> class.
 * <p>
 *
 * @author  Victor Kuliamin
 * @version 0.1, 12/03/2002
 * @see     jatva.lang.services.DeepSubject
 * @see     java.lang.Object
 * @since   JavaTESK 0.2
 */
public class ModelObject
  extends DeepSubject
{
  //---------------------------------------------------------------------------
  // Constructors
  public ModelObject()
  {
    SynchronizationManager.register(this);
  }

  public ModelObject(ModelObject r)
  {
    this();
  }
  //---------------------------------------------------------------------------

  //---------------------------------------------------------------------------
  // Methods overrding the ones of DeepSubject class
  /**
   * Method overrides the corresponding method in DeepSubject class.
   *
   * @return  the clone created.
   * @see     jatva.lang.services.DeepSubject
   */
  public DeepSubject _ts_shallowClone ()
  {
    return new ModelObject(this);
  }
  //---------------------------------------------------------------------------

  //---------------------------------------------------------------------------
  // Group of flags and counters

  /**
   * Oracle flag - if it is set, all model methods are checked
   */
  private final static long _TS_ORACLE_SET      = 0x0000000000000001L;
  /**
   * Read-only-check flag - if it is set, all read-only accesses are checked
   */
  private final static long _TS_RO_ACCESS_CHECK = 0x0000000000000002L;

  /**
   * Oracle-call flag - marks states where oracle is called
   */
  //private final static long _TS_ORACLE_CALLED   = 0x0000000000000004L;
  /**
   * Any-pre-holds flag - marks states where some
   * precondition holds in an inheritance series of methods
   */
  private final static long _TS_ANY_PRE_HOLDS   = 0x0000000000000004L;
  
  /**
   * Any post fails - marks states where some of postconditions failed
   */
  private final static long _TS_ANY_POST_FAILS  = 0x0000000000000008L;
  
  /**
   * Implementation-called flag - marks states where implementation of model
   * method is already called
   */
  //private final static long _TS_IMPL_CALLED     = 0x0000000000000010L;

  /** Inheritance counter used to redirect model call into one of
   * inherited oracles calls.
   * Inheritance chain length is bounded by 2^16 */
  //private final static long _TS_INH_COUNTER     = 0x0000ffff00000000L;
  //private final static int  _TS_INH_SHIFT       = 32;
  /** Call counter used to count model calls of one method.
   * Inheritance and refinement chain length is bounded by 2^16 */
  //private final static long _TS_CALL_COUNTER    = 0xffff000000000000L;
  //private final static int  _TS_CALL_SHIFT      = 48;
  /**
   * Invalid flag shows whether the implementation object is removed from memory
   */
  private final static long _TS_INVALID = 0x000000000000010L;

  private long _ts_flags = _TS_ORACLE_SET | _TS_RO_ACCESS_CHECK;

  protected static enum ExecContext
  {
        None( null )
      , CommonBlock( "common block" )
      , PreCondition( "precondition" )
      , PostCondition( "postcondition" );

      private String txt;

      ExecContext( String txt)
      {
          this.txt = txt;
      }
      public String toString()
      {
          return txt;
      }
  }
  //------------------------------------------------------
  // public methods to switch oracle on and off
  /**
   * Method is used to switch on the oracle verifying the behavior of this object.
   * The parameter code>checkReadOnlyAccess</code> determines whether to check that
   * read-only objects keep their state.
   * <p>
   * User may use this method in mediators or test scenarios, but not redefine it.
   * <p>
   *
   * @param  checkReadOnlyAccess  flag saying whether to check that read-only objects keep
   *                              their state in all methods.
   */
  public final void attachOracle(boolean checkReadOnlyAccess)
  {
    if(checkReadOnlyAccess) _ts_flags |= (_TS_ORACLE_SET | _TS_RO_ACCESS_CHECK);
    else                    _ts_flags |= _TS_ORACLE_SET;
  }

  /**
   * This method switches on the oracle with <code>checkReadOnlyAccess</code> set to
   * <code>true</code>.
   * <p>
   * User may use this method in mediators or test scenarios, but not redefine it.
   * <p>
   *
   */
  public final void attachOracle()
  {
    _ts_flags |= (_TS_ORACLE_SET | _TS_RO_ACCESS_CHECK);
  }

  /**
   * Used to switch off the oracle of this object.
   * <p>
   * User may use this method in mediators or test scenarios, but not redefine it.
   * <p>
   *
   */
  public final void removeOracle()
  {
    _ts_flags &= ~(_TS_ORACLE_SET | _TS_RO_ACCESS_CHECK);
  }

  //------------------------------------------------------
  // protected methods to deal with various flags
  //
  protected final boolean _ts_isROAccessChecked()
  {
    return ((_ts_flags & _TS_RO_ACCESS_CHECK) != 0);
  }

  public final boolean _ts_isOracleSet()
  {
    return ((_ts_flags & _TS_ORACLE_SET) != 0);
  }
  /*
  protected final boolean _ts_isOracleCalled()
  {
    return ((_ts_flags & _TS_ORACLE_CALLED) != 0);
  }

  protected final void _ts_setOracleCalled()
  {
    _ts_flags |= _TS_ORACLE_CALLED;
  }

  protected final void _ts_removeOracleCalled()
  {
    _ts_flags &= ~_TS_ORACLE_CALLED;
  }
  */
  public final boolean _ts_isAnyPreHolds()
  {
    return ((_ts_flags & _TS_ANY_PRE_HOLDS) != 0);
  }

  public final boolean _ts_isAnyPostFails()
  {
    return ((_ts_flags & _TS_ANY_POST_FAILS) != 0);
  }

  public final void _ts_setAnyPreHolds()
  {
    _ts_flags |= _TS_ANY_PRE_HOLDS;
  }

  public final void _ts_setAnyPostFails()
  {
      _ts_flags |= _TS_ANY_POST_FAILS;
  }
  
  public final void _ts_removeAnyPreHolds()
  {
    _ts_flags &= ~_TS_ANY_PRE_HOLDS;
  }
  
  public final void _ts_removeAnyPostFails()
  {
    _ts_flags &= ~_TS_ANY_POST_FAILS;
  }
  /*
  protected final boolean _ts_isImplCalled()
  {
    return ((_ts_flags & _TS_IMPL_CALLED) != 0);
  }

  protected final void _ts_setImplCalled()
  {
    _ts_flags |= _TS_IMPL_CALLED;
  }

  protected final void _ts_removeImplCalled()
  {
    _ts_flags &= ~_TS_IMPL_CALLED;
  }
  */
  //------------------------------------------------------
  // protected methods to deal with inheritance counter
  /*
  protected final int _ts_inheritanceCounter()
  {
    return (int)((_ts_flags & _TS_INH_COUNTER) >> _TS_INH_SHIFT);
  }

  protected final int _ts_increaseInheritanceCounter()
  {
    int res = (int)((_ts_flags & _TS_INH_COUNTER) >> _TS_INH_SHIFT);
    res++;
    _ts_flags &= ~_TS_INH_COUNTER;
    _ts_flags |= ((long)res << _TS_INH_SHIFT);
    return res;
  }

  protected final int _ts_decreaseInheritanceCounter()
  {
    int res = (int)((_ts_flags & _TS_INH_COUNTER) >> _TS_INH_SHIFT);
    res--;
    _ts_flags &= ~_TS_INH_COUNTER;
    _ts_flags |= ((long)res << _TS_INH_SHIFT);
    return res;
  }

  protected final int _ts_setInheritanceCounter(int v)
  {
    v &= 65535;
    _ts_flags &= ~_TS_INH_COUNTER;
    _ts_flags |= ((long)v << _TS_INH_SHIFT);
    return v;
  }

  //------------------------------------------------------
  // protected methods to deal with call counter
  //
  protected final int _ts_callCounter()
  {
    return (int)((_ts_flags & _TS_CALL_COUNTER) >> _TS_CALL_SHIFT);
  }

  protected final int _ts_increaseCallCounter()
  {
    int res = (int)((_ts_flags & _TS_CALL_COUNTER) >> _TS_CALL_SHIFT);
    res++;
    _ts_flags &= ~_TS_CALL_COUNTER;
    _ts_flags |= ((long)res << _TS_CALL_SHIFT);
    return res;
  }

  protected final int _ts_decreaseCallCounter()
  {
    int res = (int)((_ts_flags & _TS_CALL_COUNTER) >> _TS_CALL_SHIFT);
    res--;
    _ts_flags &= ~_TS_CALL_COUNTER;
    _ts_flags |= ((long)res << _TS_CALL_SHIFT);
    return res;
  }
  */
  /** Class level in the inheritance hierarchy  */
  //!!!protected static int _ts_classLevel = 0;
  //------------------------------------------------------
  // class state support
  //
  private static ModelClass _ts_classModel = new ModelClass("jatva.lang.ModelObject", null);

  protected static ModelClass _ts_getClassModel()
  {
    return _ts_classModel;
  }

  protected static void _ts_setClassModel(ModelClass _ts_model)
  {
    _ts_classModel = _ts_model;
  }

  public static DeepSubject _ts_cloneClass()
  {
    return _ts_classModel._ts_shallowClone();
  }

  //------------------------------------------------------
  // public methods to switch calss oracle on and off
  //
  public static void attachClassOracle(boolean checkReadOnlyAccess)
  {
    _ts_classModel.attachOracle(checkReadOnlyAccess);
  }

  public static void attachClassOracle()
  {
    attachClassOracle(true);
  }

  public static void removeClassOracle()
  {
    _ts_classModel.removeOracle();
  }

  //------------------------------------------------------
  // object state synchronization methods
  //
  /**
   * Used for synchronization of the model state with the implementation one.
   * May update only a part of model state.
   * Throws <code>StateSynchronizationException</code> to exhibit the impossibility
   * to construct correct model state corresponding to the current implementation one.
   * <p>
   * The general procedure of synchronizing two states can have the following steps
   * <ol>
   *   <li>If the implementation is not a low-level model and has no mechanism
   *   for invariants checking, then the first step is to check the implementation-specific
   *   invariants that may be broken due to loose structure of implementation data.
   *   These invariants describe the domain of the <i>retrieve mapping</i>.
   *   </li>
   *   <li>Proper <i>retrieve mapping</i>, which construct some part of model state
   *   (<i>retrievable state</i>) from the current state of implementation.
   *   </li>
   *   <li>Last step is the checking of invariants specific to the constructed model.
   *   </li>
   * </ol>
   * <p>
   * User may redefine this method in mediator, if it is needed, but should never use it.
   * <p>
   *
   * @throws StateSynchronizationException if the new state cannot be constructed correctly.
   */
  public void _ts_mapStateUp ()
  {
  }

  /**
   * This method is used to synchronize the state of implementation with the model
   * state of this object. It may be used, for example, to construct implementation
   * values of arguments from model ones.
   * <p>
   * User may redefine this method in mediator, if it is needed. He may use it
   * in mediators or test scenarios.
   * <p>
   *
   * @throws StateSynchronizationException if the new state cannot be constructed correctly.
   */
  public void mapStateDown ()
  {
    throw new MapStateDownNotSupportedException( getClass().getName() );
  }

  /**
   * Used to synchronize the complete model state of an object with
   * its implementation.
  * <p>
   * User may redefine this method in mediator, if it is needed, but should never use it.
   * <p>
   *
   * @throws StateSynchronizationException if the new state cannot be constructed correctly.

  public void _ts_mapQuiescentStateUp()
  {
    throw new MapQuiescentStateUpNotSupportedException( getClass().getName() );
  }*/

  //------------------------------------------------------
  // class state synchronization methods
  //
  public static void _ts_mapClassStateUp()
  {
    _ts_classModel._ts_mapStateUp();
  }

  public static void mapClassStateDown()
  {
    _ts_classModel.mapStateDown();
  }

  /*
  public static void _ts_mapQuiescentlClassStateUp()
  {
    _ts_classModel._ts_mapQuiescentStateUp();
  }*/

  /** Used to point out the necesary mediator classes in scenario initilization */
  public static void initClassState()
  {
  }

  //------------------------------------------------------
  // invariants checking methods
  //
  /**
   * Checks object quiescent-only invariants. In case of invariant violation or
   * exception during invariant execution it throws <code>InvariantException</code>
   * with identification of the broken invariant.
   * <p>
   * User should never use this method, nor redefine it.
   * <p>
   * Future releases may add path identification for broken invariant.
   * <p>
   *
   * @throws InvariantException if some invariant fails or throws an exception.
   */
  public void _ts_checkInvariants ()
  {
  }

  /**
   * Checks static quiescent-only invariants. In case of invariant violation or
   * exception during invariant execution it throws <code>InvariantException</code>
   * with identification of the broken invariant.
   * <p>
   * User should never use this method, nor redefine it.
   * <p>
   * Future releases may add path identification for broken invariant.
   * <p>
   *
   * @throws InvariantException if some invariant fails or throws an exception.
   */
  public static void _ts_checkClassInvariants ()
  {
    _ts_classModel._ts_checkInvariants();
  }

  public void _ts_checkPrimeInvariants ()
  {

  }

  public static void _ts_checkClassPrimeInvariants ()
  {

  }

  public final void _ts_checkAllInvariants ()
  {
    _ts_checkPrimeInvariants();
    _ts_checkInvariants();
  }
  //------------------------------------------------------
  // operationInfo descriptors manipulation
  //
  //!!!private static OperationMap _ts_specificationOperations = null;

  //!!!private static void _ts_initSpecificationOperations()
  //{}

  public static SEOperationInfo _ts_getClassOperationInfo(String signature)
  {
    return null;
  }

  public SEOperationInfo _ts_getOperationInfo(String signature)
  {
    return null;
  }

  //------------------------------------------------------
  // comparison of implementation fields
  //

  //!!!protected static MediatorSet _ts_mediatorSet = null;

  public boolean _ts_compareImplementationFields(Object[] implFields)
  {
    return false;
  }

  public void    _ts_initializeImplementation()
  {
  }

  /*public boolean _ts_compareImplementationFields(ModelObject o)
  {
    return false;
  }*/

  //------------------------------------------------------
  // unmanaged objects manipulation
  //

  /**
   * The <code>isValidModel</code> helps to check whether the implementation
   * represented by this model object already destroyed with the help of some
   * memory management procedures.
   * <p>
   * User may use this method, but not redefine it.
   * <p>
   *
   * @return  <code>true</code> if and only if this object corresponds to
   *          undestroyed implementation object.
   */
  public boolean isValidModel()
  {
    return ((_ts_flags & _TS_INVALID) == 0);
  }

  /**
   * The <code>invalidateModel</code> method is to be called when
   * the implementation object modeled was destroyed. This may be done
   * by user if there are no automatic mechanism to control memory deallocation
   * in the system under test, or automatically if such a mechanism exists.
   * <p>
   * User may use this method, and may redefine it, if he knows the correct
   * procedure to check whether an implementation object was destroyed.
   * <p>
   */
  public void invalidateModel()
  {
    _ts_flags |= _TS_INVALID;
    SynchronizationManager.registerForCleaning(this);
  }

  /**
   * This method is called by test system synchronization mechanism
   * to remove the model object related with the destroyed
   * implementation one from the system.
   * <p>
   * User should never use this method, nor redefine it.
   * <p>
   */
  public void cleanUpModel()
  {
  }



  //------------------------------------------------------
  // miscellaneous general-purpose methods
  //

  public Object clone()
  {
    return new ModelObject(this);
  }
}
