/*
 * Copyright (c) 2004 ISP RAS.
 * 109004, B. Communisticheskaya, 25, Moscow, Russia.
 * All rights reserved.
 *
 * $Id: DeepOperationManager.java,v 1.7 2007/11/27 13:08:58 almo Exp $
 * Created on 02.06.2004.
 * Modified on 16.02.2005.
 */
package jatva.lang.services;

import jatva.iterators.lang.IIntegerIterator;
import jatva.iterators.lang.IObjectIntPairIterator;
import jatva.lang.InsufficientAccessibilityError;
import jatva.util.IdentityList;
import jatva.util.IdentityMap;
import jatva.util.IdentitySet;
import jatva.util.IdentityToIntMap;
import jatva.util.IntList;
import jatva.util.IntSet;
import jatva.util.IntToIntMap;
import jatva.util.IntToObjectMap;
import jatva.util.IntToStringMap;
import jatva.util.StringToObjectMap;
import jatva.util.Utils;
import jatva.util.IList;
import jatva.util.ISet;
import jatva.util.IMap;
import jatva.util.IntToStringMap.IntToStringMapEnumerator;
import jatva.util.StringToObjectMap.StringToObjectMapEnumerator;

import java.lang.reflect.Array;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

/**
 * The class managing deep cloning and equality checks. 
 * 
 * @author Victor Kuliamin
 */
public class DeepOperationManager
{
  private final static ThreadLocal tclones  = new ThreadLocal();
  private final static ThreadLocal tequals  = new ThreadLocal();
  private final static ThreadLocal tstrings = new ThreadLocal();

  /**
   * The method used to finish deep cloning process.
   */
  private final static void finishCloning()
  {
    if(tclones.get() != null) tclones.set(null);
  }

  /**
   * The method used to finish deep conversion to string. 
   */
  private final static void finishToString()
  {
    if(tstrings.get() != null) tstrings.set(null);
  }
  
  /**
   * The method used to finish deep equality check. 
   */
  private final static void finishEquals()
  {
    if(tequals.get() != null) tequals.set(null);
  }
  
  /**
   * The method to clone deeply a set of objects starting
   * from the root objects sepcified.
   * 
   * @param o    The root objects
   * @return     The array of root objects' clones
   */
  public static Object[] deepCloneIndependent(Object[] o)
  {
    if(o != null)
    {
      switch(o.length)
      {
        case 0: return new Object[0];
        case 1: return new Object[] { getDeepCloneIndependent(o[0]) };
        default:
        {
          IdentityMap map = null;
          Object      t   = tclones.get();

          if(t != null)   map = (IdentityMap)t;
          else          { map = new IdentityMap(); tclones.set(map); }
          
          Object[] res = new Object[o.length];
          for(int i = 0; i < o.length; i++)
            res[i] = getDeepCloneIndependent(o[i], map);
          
          finishCloning();
          return res;
        }
      }
    }
    else return null;
  }
  
  public static Object deepCloneIndependent(Object o, IdentityMap m)
  {
    return o; //!!!!!!!!! see store()
  }
  
  /**
   * The method returns the clone of its parameter and
   * starts its deep cloning. If the clone already exists
   * the method returns it as a result. This method is called
   * only from {@link DeepSubject#_ts_implDeepClone()}.
   * 
   * @param o    The object to be cloned.
   * @return     The existing clone of the object <code>o</code>
   *             or a new one.
   */
  static DeepSubject getCloneIndependent(DeepSubject o)
  {
    DeepSubject res = o._ts_shallowClone();
    IdentityMap map = (IdentityMap)tclones.get();

    map.put(o, res);

    return res;
  }

  /**
   * The method return deep clone of its parameter.
   * 
   * @param o    The object to clone
   * @return     A deep clone of <code>o</code>
   */
  public static Object getDeepCloneIndependent(Object o)
  {
    IdentityMap map = null;
    Object      t   = tclones.get();

    if(t != null)   map = (IdentityMap)t;
    else          { map = new IdentityMap(); tclones.set(map); }

    t = getDeepCloneIndependent(o, map);
    finishCloning();
    return t;
  }
  
  /**
   * The methods returns a deep clone of its parameter in the process of
   * deep cloning of some object. It should be used only from implementations 
   * of the method {@link DeepSubject#_ts_fieldsDeepCloneIndependent(DeepSubject)}.
   * 
   * @param o    The object to find clone for
   * @return     The stored clone of the parameter or the result of its deep cloning.
   */
  public static Object getSecondaryObjectDeepCloneIndependent(Object o)
  {
    IdentityMap map = (IdentityMap)tclones.get();
    
    return getDeepCloneIndependent(o, map);
  }
  
  /**
   * The method returns a deep clone of its parameter.
   * It uses already defined map of clones.
   * 
   * Makes separate object form <code>StringBuffer</code>.
   * Immutable objects such as instances of <code>String</code>, 
   * <code>Number</code>, <code>Object</code> are considered as their own
   * clones. 
   * 
   * @param o      The object to find clone for
   * @param map    The map of already cloned objects
   * @return       The stored clone of the parameter or a result of deep 
   *               cloning
   */
  private static Object getDeepCloneIndependent(Object o, IdentityMap map)
  {
    if(o == null) return null;

    if(map.containsKey(o))                return map.get(o);
    else if(o instanceof DeepSubject) return ((DeepSubject)o)._ts_implDeepClone();
    else
    {
      Class cl = o.getClass();
      
      if(cl.isArray())                   return getArrayDeepCloneIndependent(o, map, cl);
      else if(o instanceof List)         return getListDeepCloneIndependent((List)o, map, cl);
      else if(o instanceof Map)          return getMapDeepCloneIndependent((Map)o, map, cl);
      else if(o instanceof Collection)   return getCollectionDeepCloneIndependent((Collection)o, map, cl);
      else if(o instanceof Number)       return o; // getNumberClone((Number)o);
      else if(o instanceof String)       return o;
      else if(cl.getName().equals("java.lang.Object")) return o; // new Object();
      else if(o instanceof StringBuffer) 
      {
        Object res = new StringBuffer(o.toString());
        map.put(o, res);
        return res;
      }
      else
      {
        try
        {
          Method mth; 
          if((mth = cl.getMethod("clone", new Class[]{})) != null) ;
          else if(o instanceof Cloneable)
            mth = cl.getDeclaredMethod("clone", new Class[]{});
          else
          {
            map.put(o, null);
            return null;
          }
          
          Object res = mth.invoke(o, new Object[]{});
          map.put(o, res);
          return res;
        }
        catch(Exception e)
        {
          map.put(o, null);
          return null;
        }
      }
    }
  }
 
  /**
   * The method returns a deep clone of an array.
   * 
   * @param a      The array object to clone
   * @param map    The map of already cloned objects
   * @param cl     Class object of the array <code>o</code>
   * @return       A deep clone of the array <code>o</code>
   */
  private static Object getArrayDeepCloneIndependent(Object a, IdentityMap map, Class cl)
  {
    if(a == null) return null;
    
    int l = Array.getLength(a);

    if(!cl.getComponentType().isArray())
    {
      cl = cl .getComponentType();
      if(cl.isPrimitive())
      {
        if(cl.equals(Boolean.TYPE))
        {
          boolean[] res = new boolean[l];
          map.put(a, res);
          System.arraycopy(a, 0, res, 0, l);
          
          return res;
        }
        else if(cl.equals(Character.TYPE))
        {
          char[] res = new char[l];
          map.put(a, res);
          System.arraycopy(a, 0, res, 0, l);
          
          return res;
        }
        else if(cl.equals(Byte.TYPE))
        {
          byte[] res = new byte[l];
          map.put(a, res);
          System.arraycopy(a, 0, res, 0, l);
          
          return res;
        }
        else if(cl.equals(Short.TYPE))
        {
          short[] res = new short[l];
          map.put(a, res);
          System.arraycopy(a, 0, res, 0, l);
          
          return res;
        }
        else if(cl.equals(Integer.TYPE))
        {
          int[] res = new int[l];
          map.put(a, res);
          System.arraycopy(a, 0, res, 0, l);
          
          return res;
        }
        else if(cl.equals(Long.TYPE))
        {
          long[] res = new long[l];
          map.put(a, res);
          System.arraycopy(a, 0, res, 0, l);
          
          return res;
        }
        else if(cl.equals(Float.TYPE))
        {
          float[] res = new float[l];
          map.put(a, res);
          System.arraycopy(a, 0, res, 0, l);
          
          return res;
        }
        else if(cl.equals(Double.TYPE))
        {
          double[] res = new double[l];
          map.put(a, res);
          System.arraycopy(a, 0, res, 0, l);
          
          return res;
        }
        else// if(cl.equals(Void.TYPE))
        {
          map.put(a, null);
          return null;
        }
      }
      else
      {
        Object res = Array.newInstance(cl, l);
        map.put(a, res);

        for(int j = 0; j < l; j++)
          Array.set(res, j, getDeepCloneIndependent(Array.get(a, j), map));
        
        return res;
      }
    }
    else
    {
      cl = cl.getComponentType();
      Object res = Array.newInstance(cl, l);
      map.put(a, res);
      for(int j = 0; j < l; j++)
        Array.set(res, j, getArrayDeepCloneIndependent(Array.get(a, j), map, cl));
      
      return res;
    }
  }

  /**
   * The method returns clone of <code>Number</code> object.
   * 
   * @param n    The <code>Number</code> object to clone
   * @return     A clone of the number <code>n</code>
   */
  public static Number getNumberClone(Number n)
  {
    if     (n ==         null   )              return null;
    if     (n instanceof Byte   )              return new Byte(n.byteValue());
    else if(n instanceof Short  )              return new Short(n.shortValue());
    else if(n instanceof Integer)              return new Integer(n.intValue()); 
    else if(n instanceof Long   )              return new Long(n.longValue()); 
    else if(n instanceof Float  )              return new Float(n.floatValue()); 
    else if(n instanceof Double )              return new Double(n.doubleValue()); 
    else if(n instanceof java.math.BigInteger) return new java.math.BigInteger(((java.math.BigInteger)n).toByteArray()); 
    else if(n instanceof java.math.BigDecimal)
    {
      java.math.BigDecimal d = (java.math.BigDecimal)n;
      return new java.math.BigDecimal(d.unscaledValue(), d.scale());
    }
    else                                       return null; 
  }
  
  /**
   * The method returns a deep clone of a list.
   * 
   * @param l      The list to clone
   * @param map    The map of already cloned objects
   * @param cl     The class object of the list <code>l</code>
   * @return       A deep clone of the list <code>l</code>
   */
  private static Object getListDeepCloneIndependent(List l, IdentityMap map, Class cl)
  {
    List res;
    try                { res = (List)(cl.newInstance()); }
    catch(Throwable e) { map.put(l, null); return null; }
    map.put(l, res);
    for(int i = 0; i < l.size(); i++)
      res.add(getDeepCloneIndependent(l.get(i), map));
    
    return res;
  }

  /**
   * The method returns a deep clone of a map.
   * 
   * @param m      The map to clone
   * @param map    The map of already cloned objects
   * @param cl     The class object of the map <code>l</code>
   * @return       A deep clone of the map <code>l</code>
   */
  private static Object getMapDeepCloneIndependent(Map m, IdentityMap map, Class cl)
  {
    Map res;
    try                { res = (Map)(cl.newInstance()); }
    catch(Throwable e) { map.put(m, null); return null; }
    map.put(m, res);
    for(Iterator i = m.keySet().iterator(); i.hasNext(); )
    {
      Object o = i.next();
      res.put(getDeepCloneIndependent(o, map), getDeepCloneIndependent(m.get(o), map));
    }
    
    return res;
  }

  /**
   * The method returns a deep clone of a collection.
   * 
   * @param c      The collection to clone
   * @param map    The map of already cloned objects
   * @param cl     The class object of the collection <code>l</code>
   * @return       A deep clone of the collection <code>l</code>
   */
  private static Object getCollectionDeepCloneIndependent(Collection c, IdentityMap map, Class cl)
  {
    Collection res;
    try                { res = (Collection)(cl.newInstance()); }
    catch(Throwable e) { map.put(c, null); return null; }
    map.put(c, res);
    for(Iterator i = c.iterator(); i.hasNext(); )
      res.add(getDeepCloneIndependent(i.next(), map));
    
    return res;
  }

  /**
   * The method for deep tracing of a set of objects. 
   * 
   * @param o    The array of objects to trace
   * @return     The array of trace strings
   */
  public static String[] deepToString(Object[] o)
  {
    if(o != null)
    {
      switch(o.length)
      {
        case 0: return new String[0];
        case 1: return new String[] { getDeepString(o[0]) };
        default:
        {
          IdentityMap map = null;
          Object      t   = tstrings.get();

          if(t != null)   map = (IdentityMap)t;
          else          { map = new IdentityMap(); tstrings.set(map); }
          
          String[] res = new String[o.length];
          StringBuffer buf = new StringBuffer();
          for(int i = 0; i < o.length; i++)
          {
            buf.delete(0, buf.length());
            appendDeepString(o[i], map, buf);
            res[i] = buf.toString();
          }
          
          finishToString();
          return res;
        }
      }
    }
    else return null;
  }
  
  /**
   * Method for deep tracing of an object to string. 
   * 
   * @param o    The object to trace
   * @return     The resulting string 
   */
  public static String getDeepString(Object o)
  {
    IdentityMap map = null;
    Object      t   = tstrings.get();

    if(t != null)   map = (IdentityMap)t;
    else          { map = new IdentityMap(); tstrings.set(map); }

    StringBuffer buf = new StringBuffer();
    appendDeepString(o, map, buf);
    finishToString();
    return buf.toString();
  }
  
  /**
   * Append the <code>StringBuffer</code> object specified with
   * deep trace string of the object <code>o</code>.
   * Should be used only from implementations 
   * of the method {@link DeepSubject#_ts_appendDeepFields(StringBuffer)}.
   * 
   * @param o      The object to trace
   * @param buf    The <code>StringBuffer</code> to append
   
  public static void appendSecondaryObject(Object o, StringBuffer buf)
  {
    IdentityMap map = (IdentityMap)tstrings.get();

    appendDeepString(o, map, buf);
  }
  */
  /**
   * Appends the <code>StringBuffer</code> object specified with
   * deep trace string of the object <code>o</code>.
   * 
   * @param o      The object to trace
   * @param map    The map of already traced objects
   * @param buf    The <code>StringBuffer</code> to append
   */
  public static void appendDeepString(Object o, IdentityMap map, StringBuffer buf)
  {
    if(o == null) { buf.append("null"); return; }

    if(map.containsKey(o)) buf.append(map.get(o));
    else
    {
      Class cl = o.getClass();
      String bb = Utils.getSimpleString(o);
      buf.append(bb);
      
      map.put(o, bb + '!');
      
      if(o instanceof DeepSubject)   { buf.append(':'); ((DeepSubject)o)._ts_implDeepAppend(buf, map); }
      else if(cl.isArray())              { buf.append(':'); appendArrayDeepString(o, map, cl, buf); }
      else if(o instanceof List)         { buf.append(':'); appendListDeepString((List)o, map, buf); }
      else if(o instanceof IList)        { buf.append(':'); appendIListDeepString((IList)o, map, buf); }
      else if(o instanceof IdentityList) { buf.append(':'); appendIdListDeepString((IdentityList)o, map, buf); }
      else if(o instanceof Map)          { buf.append(':'); appendMapDeepString((Map)o, map, buf); }
      else if(o instanceof IMap)         { buf.append(':'); appendIMapDeepString((IMap)o, map, buf); }
      else if(o instanceof IdentityToIntMap)  { buf.append(':'); appendIdIntMapDeepString((IdentityToIntMap)o, map, buf); }
      else if(o instanceof IntToObjectMap)    { buf.append(':'); appendIntObjMapDeepString((IntToObjectMap)o, map, buf); }
      else if(o instanceof StringToObjectMap) { buf.append(':'); appendStrObjMapDeepString((StringToObjectMap)o, map, buf); }
      else if(o instanceof ISet)         { buf.append(':'); appendISetDeepString((ISet)o, map, buf); }
      else if(o instanceof Collection)   { buf.append(':'); appendCollectionDeepString((Collection)o, map, buf); }
      else if(o instanceof Class)        { buf.append(':'); appendTypeString((Class)o, buf); } 
      else
      {
        Class[] cls = new Class[]{};
        Method mth1 = null, mth2 = null;
        try
        {
          mth1 = cl.getMethod("toString", cls);
          mth2 = Class.forName("java.lang.Object").getMethod("toString", cls);
        }
        catch (Throwable e) { }
        
        if(mth2.equals(mth1)) { buf.append(':'); appendCommonDeepString(o, map, cl, buf); }
        else { buf.append(':'); buf.append(o.toString()); }
      }
    }
  }
  
  private static void appendCommonDeepString(Object o, IdentityMap map, Class cl, StringBuffer buf)
  {
    Field[] ff;
    
    buf.append("{ ");
    boolean flag = false;
    
    try
    {
      while(cl != null)
      {
        ff = cl.getDeclaredFields();
        for(int i = 0; i < ff.length; i++)
        {
          if(Modifier.isStatic(ff[i].getModifiers())) continue;
          
          ff[i].setAccessible(true);

          if(flag) buf.append(", ");
          flag = true;
          
          buf.append(ff[i].getName()).append(" = ");
          
          if(ff[i].getType().isPrimitive())
          {
            buf.append(ff[i].get(o));
          }
          else
          {
            appendDeepString(ff[i].get(o), map, buf);
          }
        }
        
        cl = cl.getSuperclass();
      }
    }
    catch (Throwable e)
    {
    }
    
    buf.append(" }"); 
  }
  
  /**
   * Converts Java type object into suitable string.
   * Differs from standard conversion by converting array types
   * into strings ending with "[]". 
   * 
   * @param t    The type to convert
   * @return     The resulting string
   
  public static String typeToString(Class t)
  {
    StringBuffer buf = new StringBuffer();
    while(t.isArray()) { buf.append("[]"); t = t.getComponentType(); }
    buf.insert(0, t.getName());
    return buf.toString();
  }
  
  /**
   * Appends the <code>StringBuffer</code> specified with the name
   * of the type <code>t</code>.
   * 
   * @param t      The type to trace
   * @param b    The <code>StringBuffer</code> to append
   */
  private static void appendTypeString(Class t, StringBuffer b)
  {
    int l = b.length();
    while(t.isArray()) { b.append("[]"); t = t.getComponentType(); }
    b.insert(l, t.getName());
  }
  
  /**
   * Appends the <code>StringBuffer</code> specified with the string
   * representing the contents of an array.
   * 
   * @param o      The array to trace
   * @param map    The map of already traced objects
   * @param cl     The type of array object
   * @param buf    The <code>StringBuffer</code> to append
   */
  private static void appendArrayDeepString(Object o, IdentityMap map, Class cl, StringBuffer buf)
  {
    if(o == null) { buf.append("null"); return; }
    
    int l = Array.getLength(o);
    int j = 0;

    buf.append("[ ");
    
    if(!cl.getComponentType().isArray())
    {
      cl = cl.getComponentType();
      if(cl.isPrimitive())
      {
        if(cl.equals(Boolean.TYPE))
          for(; j < l; j++) { if(j != 0) buf.append(", "); buf.append(Array.getBoolean(o, j)); }
        else if(cl.equals(Character.TYPE))
          for(; j < l; j++) { if(j != 0) buf.append(", "); buf.append(Array.getChar(o, j));    }
        else if(cl.equals(Byte.TYPE))
          for(; j < l; j++) { if(j != 0) buf.append(", "); buf.append(Array.getByte(o, j));    }
        else if(cl.equals(Short.TYPE))
          for(; j < l; j++) { if(j != 0) buf.append(", "); buf.append(Array.getShort(o, j));   }
        else if(cl.equals(Integer.TYPE))
          for(; j < l; j++) { if(j != 0) buf.append(", "); buf.append(Array.getInt(o, j));     }
        else if(cl.equals(Long.TYPE))
          for(; j < l; j++) { if(j != 0) buf.append(", "); buf.append(Array.getLong(o, j));    }
        else if(cl.equals(Float.TYPE))
          for(; j < l; j++) { if(j != 0) buf.append(", "); buf.append(Array.getFloat(o, j));   }
        else if(cl.equals(Double.TYPE))
          for(; j < l; j++) { if(j != 0) buf.append(", "); buf.append(Array.getDouble(o, j));  }
        else// if(cl.equals(Void.TYPE))
          for(; j < l; j++) { if(j != 0) buf.append(',');                                     }
      }
      else
        for(; j < l; j++)   { if(j != 0) buf.append(", "); appendDeepString(Array.get(o, j), map, buf); }
    }
    else
      for(; j < l; j++)     { if(j != 0) buf.append(", "); appendArrayDeepString(Array.get(o, j), map, cl.getComponentType(), buf); }
    
    buf.append(" ]");   
  }
  
  /**
   * Appends the <code>StringBuffer</code> specified with the string
   * representing the contents of a list.
   * 
   * @param l      The list to trace
   * @param map    The map of already traced objects
   * @param buf    The <code>StringBuffer</code> to append
   */
  private static void appendListDeepString(List l, IdentityMap map, StringBuffer buf)
  {
    buf.append("[ ");
    for(int i = 0; i < l.size(); i++)
    {
      if(i > 0) buf.append(", ");
      appendDeepString(l.get(i), map, buf);
    }
    buf.append(" ]");
  }
  
  private static void appendIListDeepString(IList l, IdentityMap map, StringBuffer buf)
  {
    buf.append("[ ");
    for(int i = 0; i < l.size(); i++)
    {
      if(i > 0) buf.append(", ");
      appendDeepString(l.get(i), map, buf);
    }
    buf.append(" ]");
  }

  private static void appendIdListDeepString(IdentityList l, IdentityMap map, StringBuffer buf)
  {
    buf.append("[ ");
    for(int i = 0; i < l.size(); i++)
    {
      if(i > 0) buf.append(", ");
      appendDeepString(l.get(i), map, buf);
    }
    buf.append(" ]");
  }

  /**
   * Appends the <code>StringBuffer</code> specified with the string
   * representing the contents of a map.
   * 
   * @param m      The map to trace
   * @param map    The map of already traced objects
   * @param buf    The <code>StringBuffer</code> to append
   */
  private static void appendMapDeepString(Map m, IdentityMap map, StringBuffer buf)
  {
    buf.append("{ ");
    boolean first = true;
    for(Iterator i = m.keySet().iterator(); i.hasNext(); )
    {
      Object k = i.next();
      if(!first) buf.append(", ");
      first = false;
      buf.append("( ");
      appendDeepString(k, map, buf);
      buf.append(" : ");
      appendDeepString(m.get(k), map, buf);
      buf.append(" )");
    }
    buf.append(" } ");
  }
  
  private static void appendIMapDeepString(IMap m, IdentityMap map, StringBuffer buf)
  {
    buf.append("{ ");
    boolean first = true;
    for(Iterator i = m.keys().iterator(); i.hasNext(); )
    {
      Object k = i.next();
      if(!first) buf.append(", ");
      first = false;
      buf.append("( ");
      appendDeepString(k, map, buf);
      buf.append(" : ");
      appendDeepString(m.get(k), map, buf);
      buf.append(" )");
    }
    buf.append(" } ");
  }

  private static void appendIdIntMapDeepString(IdentityToIntMap m, IdentityMap map, StringBuffer buf)
  {
    buf.append("{ ");
    boolean first = true;
    for(IObjectIntPairIterator i = m.iterator(); !i.stop(); i.shift() )
    {
      if(!first) buf.append(", ");
      first = false;
      buf.append("( ");
      appendDeepString(i.value1(), map, buf);
      buf.append(" : " + i.value2() + " )");
    }
    buf.append(" } ");
  }

  private static void appendIntObjMapDeepString(IntToObjectMap m, IdentityMap map, StringBuffer buf)
  {
    buf.append("{ ");
    boolean first = true;
    for(IObjectIntPairIterator i = m.iterator(); !i.stop(); i.shift() )
    {
      if(!first) buf.append(", ");
      first = false;
      buf.append("( " + i.value2() + " : ");
      appendDeepString(i.value1(), map, buf);
      buf.append(" )");
    }
    buf.append(" } ");
  }
  
  private static void appendStrObjMapDeepString(StringToObjectMap m, IdentityMap map, StringBuffer buf)
  {
    buf.append("{ ");
    boolean first = true;
    for(StringToObjectMapEnumerator i = m.getEnumerator(); !i.moveNext(); )
    {
      if(!first) buf.append(", ");
      first = false;
      buf.append("( " + i.current().key + " : ");
      appendDeepString(i.current().value, map, buf);
      buf.append(" )");
    }
    buf.append(" } ");
  }

  /**
   * Appends the <code>StringBuffer</code> specified with the string
   * representing the contents of a collection.
   * 
   * @param c      The collection to trace
   * @param map    The map of already traced objects
   * @param buf    The <code>StringBuffer</code> to append
   */
  private static void appendCollectionDeepString(Collection c, IdentityMap map, StringBuffer buf)
  {
    buf.append("{ ");
    boolean first = true;
    for(Iterator i = c.iterator(); i.hasNext(); )
    {
      if(!first) buf.append(", ");
      first = false;
      appendDeepString(i.next(), map, buf);
    }
    buf.append(" }");
  }
  
  private static void appendISetDeepString(ISet c, IdentityMap map, StringBuffer buf)
  {
    buf.append("{ ");
    boolean first = true;
    for(Iterator i = c.iterator(); i.hasNext(); )
    {
      if(!first) buf.append(", ");
      first = false;
      appendDeepString(i.next(), map, buf);
    }
    buf.append(" }");
  }

  /**
   * The method used to compare two arrays of objects on the full deep.  
   * 
   * @param o1    The first object array
   * @param o2    The second object array
   * @return      The result of arrays deep equality check 
   */
  public static boolean deepEquals(Object[] o1, Object[] o2)
  {
    if(o1 != null)
    {
      if(o2 != null)
      {
        if(o1.length != o2.length) return false;
        
        switch(o1.length)
        {
          case 0: return true;
          case 1: return areEqual(o1[0], o2[0]);
          default:
          {
            IdentityMap map = null;
            Object      t   = tequals.get();

            if(t != null)   map = (IdentityMap)t;
            else          { map = new IdentityMap(); tequals.set(map); }
            
            boolean res = true;
            for(int i = 0; i < o1.length; i++)
              if(!areDeeplyEqual(o1[i], o2[i], map)) { res = false; break; }
            
            finishEquals();
            return res;
          }
        }
      }
      else return false;
    }
    else return (o2 == null); 
  }
  
  public static boolean areStatesDeeplyEqual(IdentityMap s1, IdentityMap s2)
  {
    if(s1 == null) return (s2 == null);
    if(s2 == null) return false;
    if(s1.size() != s2.size()) return false;
    
    IdentityMap m = new IdentityMap();
    Iterator i = s1.keys().iterator(), j = s2.keys().iterator();
    Object o1, o2; 
    
    for( ; i.hasNext();)
    {
      o1 = i.next();
      o2 = j.next();
      
      if(o1 != o2) return false;
      if(!areDeeplyEqual(o1, s2.get(o2), m)) return false;
    }
    
    return true;
  }
  
  /**
   * Checks whether reference structures of two objects are
   * isomorphic.
   * 
   * @param o1    The first object
   * @param o2    The second object
   * @return      The result of check
   */
  public static boolean areEqual(Object o1, Object o2)
  {
    IdentityMap map;
    Object t = tequals.get();
    if(t != null)   map = (IdentityMap)t;
    else          { map = new IdentityMap(); tequals.set(map); }
    
    boolean res = areDeeplyEqual(o1, o2, map);
    finishEquals();
    return res;
  }
  
  /**
   * Checks whether reference structures of two objects are
   * isomorphic.
   * Should be used only from implementations 
   * of the method {@link DeepSubject#_ts_fieldsDeepEquals(DeepSubject)}.
   * 
   * @param o1    The first object
   * @param o2    The second object
   * @return      The result of check
   */
  public static boolean areSecondaryEqual(Object o1, Object o2)
  {
    IdentityMap map = (IdentityMap)tequals.get();
    
    return areDeeplyEqual(o1, o2, map);
    
  }
  
  /**
   * Checks whether two objects has isomorphic reference structures.
   * 
   * @param o1     The first object  
   * @param o2     The second object
   * @param map    The map of already matched objects    
   * @return       The result of checking
   */
  public static boolean areDeeplyEqual(Object o1, Object o2, IdentityMap map)
  {
    if(o1 == null)                 return (o2 == null);
    else if(o2 == null)            return false;
    else if(map.containsKey(o1))   return (map.get(o1) == o2);
    else if(map.containsValue(o2)) return false;
    else
    {
      if(o1 instanceof DeepSubject)
      {
        map.put(o1, o2);

        return ((DeepSubject)o1)._ts_fieldsDeepEquals((DeepSubject)o2, map);
      }

      Class cl = o1.getClass();
      if(!cl.equals(o2.getClass())) return false;
      
      map.put(o1, o2);
      
      if(cl.isArray())                    return areArraysDeeplyEqual(o1, o2, map, cl);
      else if(o1 instanceof List)         return areListsDeeplyEqual((List)o1, (List)o2, map);
      else if(o1 instanceof IList)        return areIListsDeeplyEqual((IList)o1, (IList)o2, map);
      else if(o1 instanceof IdentityList) return areIdListsDeeplyEqual((IdentityList)o1, (IdentityList)o2, map);
      else if(o1 instanceof Map)          return areMapsDeeplyEqual((Map)o1, (Map)o2, map);
      else if(o1 instanceof IMap)         return areIMapsDeeplyEqual((IMap)o1, (IMap)o2, map);
      else if(o1 instanceof IdentityToIntMap)  return areIdIntMapsDeeplyEqual((IdentityToIntMap)o1, (IdentityToIntMap)o2, map);
      else if(o1 instanceof IntToObjectMap)    return areIntObjMapsDeeplyEqual((IntToObjectMap)o1, (IntToObjectMap)o2, map);
      else if(o1 instanceof StringToObjectMap) return areStrObjMapsDeeplyEqual((StringToObjectMap)o1, (StringToObjectMap)o2, map);
      else if(o1 instanceof ISet)       return areISetsDeeplyEqual((ISet)o1, (ISet)o2, map);
      else if(o1 instanceof Collection) return areCollectionsDeeplyEqual((Collection)o1, (Collection)o2, map);
      else
      {
        Method mth1 = null, mth2 = null;
        try
        {
          Class obc = Class.forName("java.lang.Object");
          Class[] cls = new Class[]{ obc };
          mth1 = cl.getMethod("equals", cls);
          mth2 = obc.getMethod("equals", cls);
        }
        catch (Throwable e) { }
        
        if(mth2.equals(mth1)) { map.put(o1, o2); return true; }
        else if(o1.equals(o2))
        {
          map.put(o1, o2);
          return true;
        }
        else return false;
      }
    }
  }
  
  /**
   * Checks whether the two arrays specified are isomorphic according to 
   * their reference structure.
   * 
   * @param a1     The first array 
   * @param a2     The second array
   * @param map    The map of already matched objects
   * @param cl     The class of arrays
   * @return       The result of checking
   */
  private static boolean areArraysDeeplyEqual(Object a1, Object a2, IdentityMap map, Class cl)
  {
    if(a1 == null) return (a2 == null);
    else if(a2 == null) return false;
    
    int l = Array.getLength(a1), i = 0;
    if(l != Array.getLength(a2)) return false;

    if(!cl.getComponentType().isArray())
    {
      cl = cl .getComponentType();
      if(cl.isPrimitive())
      {
        if(cl.equals(Boolean.TYPE))
        {
          for(; i < l; i++) 
            if(Array.getBoolean(a1, i) != Array.getBoolean(a2, i)) return false;
          
          return true;
        }
        else if(cl.equals(Character.TYPE))
        {
          for(; i < l; i++) 
            if(Array.getChar(a1, i) != Array.getChar(a2, i)) return false;
          
          return true;
        }
        else if(cl.equals(Byte.TYPE))
        {
          for(; i < l; i++) 
            if(Array.getByte(a1, i) != Array.getByte(a2, i)) return false;
          
          return true;
        }
        else if(cl.equals(Short.TYPE))
        {
          for(; i < l; i++) 
            if(Array.getShort(a1, i) != Array.getShort(a2, i)) return false;
          
          return true;
        }
        else if(cl.equals(Integer.TYPE))
        {
          for(; i < l; i++) 
            if(Array.getInt(a1, i) != Array.getInt(a2, i)) return false;
          
          return true;
        }
        else if(cl.equals(Long.TYPE))
        {
          for(; i < l; i++) 
            if(Array.getLong(a1, i) != Array.getLong(a2, i)) return false;
          
          return true;
        }
        else if(cl.equals(Float.TYPE))
        {
          for(; i < l; i++) 
            if(Array.getFloat(a1, i) != Array.getFloat(a2, i)) return false;
          
          return true;
        }
        else if(cl.equals(Double.TYPE))
        {
          for(; i < l; i++) 
            if(Array.getDouble(a1, i) != Array.getDouble(a2, i)) return false;
          
          return true;
        }
        else// if(cl.equals(Void.TYPE))
        {
          return true;
        }
      }
      else
      {
        for(; i < l; i++)
          if(!areDeeplyEqual(Array.get(a1, i), Array.get(a2, i), map)) return false;
        
        return true;
      }
    }
    else
    {
      cl = cl.getComponentType();
      for(; i < l; i++)
        if(!areArraysDeeplyEqual(Array.get(a1, i), Array.get(a2, i), map, cl)) return false;
      
      return true;
    }
  }
  
  /**
   * Checks whether the two lists specified are isomorphic according to 
   * their reference structure.
   * 
   * @param l1     The first list 
   * @param l2     The second list
   * @param map    The map of already matched objects
   * @return       The result of checking
   */
  private static boolean areListsDeeplyEqual(List l1, List l2, IdentityMap map)
  {
    int s = l1.size();
    if(s != l2.size()) return false;
    
    for(int i = 0; i < s; i++)
      if(!areDeeplyEqual(l1.get(i), l2.get(i), map)) return false;

    return true;
  }
  
  private static boolean areIListsDeeplyEqual(IList l1, IList l2, IdentityMap map)
  {
    int s = l1.size();
    if(s != l2.size()) return false;
    
    for(int i = 0; i < s; i++)
      if(!areDeeplyEqual(l1.get(i), l2.get(i), map)) return false;

    return true;
  }

  private static boolean areIdListsDeeplyEqual(IdentityList l1, IdentityList l2, IdentityMap map)
  {
    int s = l1.size();
    if(s != l2.size()) return false;
    
    for(int i = 0; i < s; i++)
      if(!areDeeplyEqual(l1.get(i), l2.get(i), map)) return false;

    return true;
  }

  /**
   * Checks whether the two maps specified are isomorphic according to 
   * their reference structure.
   * 
   * Not implemented completely yet.
   * Just now simply checks that the sizes of maps are equal.
   * 
   * @param m1     The first map
   * @param m2     The second map
   * @param map    The map of already matched objects
   * @return       The result of checking
   */
  private static boolean areMapsDeeplyEqual(Map m1, Map m2, IdentityMap map)
  {
    if(m1.size() != m2.size()) return false;

    int max = 1;
    IdentityMap keys = new IdentityMap();
    IdentitySet set;
    boolean shallow = true;
    Iterator i, j;
    Object o, p;
    
    for(i = m1.keySet().iterator(); i.hasNext(); )
    {
      o = i.next();
      
      shallow = false;
      for(j = m2.keySet().iterator(); j.hasNext(); )
      {
        p = j.next();
        
        if(areCommonEqual(o, p) && areCommonEqual(m1.get(o), m2.get(p)))
        {
          shallow = true;
          if(keys.containsKey(o))
          {
            set = (IdentitySet)keys.get(o);
            set.add(p);
            max = Math.max(max, set.size());
          }
          else
          {
            set = new IdentitySet();
            keys.put(o, set);
            set.add(p);
          }
        }
      }
      
      if(!shallow) return false;
    }
    
    for(i = keys.keys().iterator(); i.hasNext(); )
    {
      o = i.next();
      set = (IdentitySet)keys.get(o);
      j = set.iterator();
      
      p = j.next();
      
      /* works only for one-to-one correspondences */
      if(!areDeeplyEqual(o, p, map)) return false;
      if(!areDeeplyEqual(m1.get(o), m2.get(p), map)) return false;
    }
      
    return true;
  }
  
  private static boolean areIMapsDeeplyEqual(IMap m1, IMap m2, IdentityMap map)
  {
    if(m1.size() != m2.size()) return false;

    int max = 1;
    IdentityMap keys = new IdentityMap();
    IdentitySet set;
    boolean shallow = true;
    Iterator i, j;
    Object o, p;
    
    for(i = m1.keys().iterator(); i.hasNext(); )
    {
      o = i.next();
      
      shallow = false;
      for(j = m2.keys().iterator(); j.hasNext(); )
      {
        p = j.next();
        
        if(areCommonEqual(o, p) && areCommonEqual(m1.get(o), m2.get(p)))
        {
          shallow = true;
          if(keys.containsKey(o))
          {
            set = (IdentitySet)keys.get(o);
            set.add(p);
            max = Math.max(max, set.size());
          }
          else
          {
            set = new IdentitySet();
            keys.put(o, set);
            set.add(p);
          }
        }
      }
      
      if(!shallow) return false;
    }
    
    for(i = keys.keys().iterator(); i.hasNext(); )
    {
      o = i.next();
      set = (IdentitySet)keys.get(o);
      j = set.iterator();
      
      p = j.next();
      
      /* works only for one-to-one correspondences */
      if(!areDeeplyEqual(o, p, map)) return false;
      if(!areDeeplyEqual(m1.get(o), m2.get(p), map)) return false;
    }
      
    return true;
  }
  
  private static boolean areIdIntMapsDeeplyEqual(IdentityToIntMap m1, IdentityToIntMap m2, IdentityMap map)
  {
    if(m1.size() != m2.size()) return false;

    int max = 1;
    IdentityMap keys = new IdentityMap();
    IdentitySet set;
    boolean shallow = true;
    Iterator i, j;
    Object o, p;
    
    for(i = m1.keySet().iterator(); i.hasNext(); )
    {
      o = i.next();
      
      shallow = false;
      for(j = m2.keySet().iterator(); j.hasNext(); )
      {
        p = j.next();
        
        if(areCommonEqual(o, p) && m1.get(o) == m2.get(p))
        {
          shallow = true;
          if(keys.containsKey(o))
          {
            set = (IdentitySet)keys.get(o);
            set.add(p);
            max = Math.max(max, set.size());
          }
          else
          {
            set = new IdentitySet();
            keys.put(o, set);
            set.add(p);
          }
        }
      }
      
      if(!shallow) return false;
    }
    
    for(i = keys.keys().iterator(); i.hasNext(); )
    {
      o = i.next();
      set = (IdentitySet)keys.get(o);
      j = set.iterator();
      
      p = j.next();
      
      /* works only for one-to-one correspondences */
      if(!areDeeplyEqual(o, p, map)) return false;
    }
      
    return true;
  }
  
  private static boolean areIntObjMapsDeeplyEqual(IntToObjectMap m1, IntToObjectMap m2, IdentityMap map)
  {
    if(m1.size() != m2.size()) return false;

    int max = 1;
    IntToObjectMap keys = new IntToObjectMap();
    IdentitySet set;
    boolean shallow = true;
    Iterator i, j;
    int o, p;
    
    for(i = m1.keySet().iterator(); i.hasNext(); )
    {
      o = (Integer)i.next();
      
      shallow = false;
      for(j = m2.keySet().iterator(); j.hasNext(); )
      {
        p = (Integer)j.next();
        
        if(o == p && areCommonEqual(m1.get(o), m2.get(p)))
        {
          shallow = true;
          if(keys.containsKey(o))
          {
            set = (IdentitySet)keys.get(o);
            set.add( m2.get(p) );
            max = Math.max(max, set.size());
          }
          else
          {
            set = new IdentitySet();
            keys.put(o, set);
            set.add( m2.get(p) );
          }
        }
      }
      
      if(!shallow) return false;
    }
    
    for(i = keys.keySet().iterator(); i.hasNext(); )
    {
      o = (Integer)i.next();
      set = (IdentitySet)keys.get(o);
      j = set.iterator();
      
      /* works only for one-to-one correspondences */
      if(!areDeeplyEqual(m1.get(o), j.next(), map)) return false;
    }
      
    return true;
  }
  
  private static boolean areStrObjMapsDeeplyEqual(StringToObjectMap m1, StringToObjectMap m2, IdentityMap map)
  {
    if(m1.size() != m2.size()) return false;

    int max = 1;
    StringToObjectMap keys = new StringToObjectMap();
    IdentitySet set;
    boolean shallow = true;
    StringToObjectMapEnumerator i, j;
    String o, p;
    
    for(i = m1.getEnumerator(); i.moveNext(); )
    {
      o = i.current().key;
      
      shallow = false;
      for(j = m2.getEnumerator(); j.moveNext(); )
      {
        p = j.current().key;
        
        if(o.equals(p) && areCommonEqual(m1.get(o), m2.get(p)))
        {
          shallow = true;
          if(keys.containsKey(o))
          {
            set = (IdentitySet)keys.get(o);
            set.add( m2.get(p) );
            max = Math.max(max, set.size());
          }
          else
          {
            set = new IdentitySet();
            keys.put(o, set);
            set.add( m2.get(p) );
          }
        }
      }
      
      if(!shallow) return false;
    }
    
    for(i = keys.getEnumerator(); i.moveNext(); )
    {
      o = i.current().key;
      set = (IdentitySet)keys.get(o);
      
      /* works only for one-to-one correspondences */
      if(!areDeeplyEqual(m1.get(o), set.iterator().next(), map)) return false;
    }
      
    return true;
  }
  
  /**
   * Checks whether the two collections specified are isomorphic according to 
   * their reference structure.
   * 
   * Not implemented completely yet.
   * Just now simply checks that the sizes of collections are equal.
   * 
   * @param c1     The first collection
   * @param c2     The second collection
   * @param map    The map of already matched objects
   * @return       The result of checking
   */
  private static boolean areCollectionsDeeplyEqual(Collection c1, Collection c2, IdentityMap map)
  {
    if(c1.size() != c2.size()) return false;

    int max = 1;
    IdentityMap keys = new IdentityMap();
    IdentitySet set;
    boolean shallow = true;
    Iterator i, j;
    Object o, p;
    
    for(i = c1.iterator(); i.hasNext(); )
    {
      o = i.next();
      
      shallow = false;
      for(j = c2.iterator(); j.hasNext(); )
      {
        p = j.next();
        
        if(areCommonEqual(o, p))
        {
          shallow = true;
          if(keys.containsKey(o))
          {
            set = (IdentitySet)keys.get(o);
            set.add(p);
            max = Math.max(max, set.size());
          }
          else
          {
            set = new IdentitySet();
            keys.put(o, set);
            set.add(p);
          }
        }
      }
      
      if(!shallow) return false;
    }
    
    for(i = keys.keys().iterator(); i.hasNext(); )
    {
      o = i.next();
      set = (IdentitySet)keys.get(o);
      j = set.iterator();
      
      p = j.next();
      
      /* works only for one-to-one correspondences */
      if(!areDeeplyEqual(o, p, map)) return false;
    }
      
    return true;
  }
  
  private static boolean areISetsDeeplyEqual(ISet c1, ISet c2, IdentityMap map)
  {
    if(c1.size() != c2.size()) return false;

    int max = 1;
    IdentityMap keys = new IdentityMap();
    IdentitySet set;
    boolean shallow = true;
    Iterator i, j;
    Object o, p;
    
    for(i = c1.iterator(); i.hasNext(); )
    {
      o = i.next();
      
      shallow = false;
      for(j = c2.iterator(); j.hasNext(); )
      {
        p = j.next();
        
        if(areCommonEqual(o, p))
        {
          shallow = true;
          if(keys.containsKey(o))
          {
            set = (IdentitySet)keys.get(o);
            set.add(p);
            max = Math.max(max, set.size());
          }
          else
          {
            set = new IdentitySet();
            keys.put(o, set);
            set.add(p);
          }
        }
      }
      
      if(!shallow) return false;
    }
    
    for(i = keys.keys().iterator(); i.hasNext(); )
    {
      o = i.next();
      set = (IdentitySet)keys.get(o);
      j = set.iterator();
      
      p = j.next();
      
      /* works only for one-to-one correspondences */
      if(!areDeeplyEqual(o, p, map)) return false;
    }
      
    return true;
  }

  
  private static boolean areCommonEqual(Object o1, Object o2)
  {
    if(o1 == null)      return (o2 == null);
    else if(o2 == null) return false;
    
    Class c = o1.getClass(), d;
    
    if(c != o2.getClass()) return false;
    
    Field[] fl;
    int i;
    
    do
    {
      fl = c.getDeclaredFields();
      for(i = 0; i < fl.length; i++)
      {
        try
        {
          fl[i].setAccessible(true);        
          if(fl[i].getDeclaringClass().isPrimitive() && !fl[i].get(o1).equals(fl[i].get(o2))) return false;
        }
        catch(SecurityException e)
        {
          throw new InsufficientAccessibilityError("Insufficient rights to get field value"
              , e
              , new String[]{"class", "field"}
              , new String[]{c.getName(), fl[i].getName()}
             );
        }
        catch(IllegalAccessException e)
        {
          throw new InsufficientAccessibilityError("Insufficient rights to get field value"
              , e
              , new String[]{"class", "field"}
              , new String[]{c.getName(), fl[i].getName()}
             );
        }
      }
    } while ((c = c.getSuperclass()) != null);
    
    return true;
  }
  
  public static boolean areDeeplyEqual(Object o1, Object o2, IdentityMap map, IdentityMap tmp)
  {
    if(o1 == null)                 return (o2 == null);
    else if(o2 == null)            return false;
    else if(map.containsKey(o1))   return (map.get(o1) == o2);
    else if(map.containsValue(o2)) return false;
    else if(tmp != null)
    {
      if(tmp.containsKey(o1))        return (tmp.get(o1) == o2);
      else if(tmp.containsValue(o2)) return false;
    }
    
    Class cl = o1.getClass();
    if(!cl.equals(o2.getClass())) return false;
    
    if(tmp != null) tmp.put(o1, o2);
    else            map.put(o1, o2);
    
    if(o1 instanceof DeepSubject)
      return ((DeepSubject)o1)._ts_fieldsDeepEquals((DeepSubject)o2, map, tmp);
    else if(cl.isArray())
      return areArraysDeeplyEqual(o1, o2, map, tmp, cl);
    else if(o1 instanceof List)
      return areListsDeeplyEqual((List)o1, (List)o2, map, tmp);
    else if(o1 instanceof Map)
      return areMapsDeeplyEqual((Map)o1, (Map)o2, map, tmp);
    else if(o1 instanceof Collection)
      return areCollectionsDeeplyEqual((Collection)o1, (Collection)o2, map, tmp);
    else
    {
      return areCommonsDeeplyEqual(o1, o2, map, tmp, cl);
      /*Method mth1 = null, mth2 = null;
      try
      {
        Class obc = Class.forName("java.lang.Object");
        Class[] cls = new Class[]{ obc };
        mth1 = cl.getMethod("equals", cls);
        mth2 = obc.getMethod("equals", cls);
      }
      catch (Throwable e) { }
      
      if(mth2.equals(mth1)) { return true; }
      else if(o1.equals(o2))
      {
        return true;
      }
      else return false;*/
    }
  }
  
  private static boolean areCommonsDeeplyEqual(Object o1, Object o2, IdentityMap map, IdentityMap tmp, Class cl)
  {
    Field[] ff;
    
    try
    {
      while(cl != null)
      {
        ff = cl.getDeclaredFields();
        for(int i = 0; i < ff.length; i++)
        {
          if(Modifier.isStatic(ff[i].getModifiers())) continue;
          
          ff[i].setAccessible(true);

          if(ff[i].getType().isPrimitive())
          {
            if(!ff[i].get(o1).equals(ff[i].get(o2))) return false;
          }
          else
          {
            if(!areDeeplyEqual(ff[i].get(o1), ff[i].get(o2), map, tmp)) return false;
          }
        }
        
        cl = cl.getSuperclass();
      }
    }
    catch (Throwable e)
    {
      return false;
    }
    
    return true;
  }
  
  private static boolean areArraysDeeplyEqual(Object a1, Object a2, IdentityMap map, IdentityMap tmp, Class cl)
  {
    if(a1 == null) return (a2 == null);
    else if(a2 == null) return false;
    
    int l = Array.getLength(a1), i = 0;
    if(l != Array.getLength(a2)) return false;

    if(!cl.getComponentType().isArray())
    {
      cl = cl .getComponentType();
      if(cl.isPrimitive())
      {
        if(cl.equals(Boolean.TYPE))
        {
          for(; i < l; i++) 
            if(Array.getBoolean(a1, i) != Array.getBoolean(a2, i)) return false;
          
          return true;
        }
        else if(cl.equals(Character.TYPE))
        {
          for(; i < l; i++) 
            if(Array.getChar(a1, i) != Array.getChar(a2, i)) return false;
          
          return true;
        }
        else if(cl.equals(Byte.TYPE))
        {
          for(; i < l; i++) 
            if(Array.getByte(a1, i) != Array.getByte(a2, i)) return false;
          
          return true;
        }
        else if(cl.equals(Short.TYPE))
        {
          for(; i < l; i++) 
            if(Array.getShort(a1, i) != Array.getShort(a2, i)) return false;
          
          return true;
        }
        else if(cl.equals(Integer.TYPE))
        {
          for(; i < l; i++) 
            if(Array.getInt(a1, i) != Array.getInt(a2, i)) return false;
          
          return true;
        }
        else if(cl.equals(Long.TYPE))
        {
          for(; i < l; i++) 
            if(Array.getLong(a1, i) != Array.getLong(a2, i)) return false;
          
          return true;
        }
        else if(cl.equals(Float.TYPE))
        {
          for(; i < l; i++) 
            if(Array.getFloat(a1, i) != Array.getFloat(a2, i)) return false;
          
          return true;
        }
        else if(cl.equals(Double.TYPE))
        {
          for(; i < l; i++) 
            if(Array.getDouble(a1, i) != Array.getDouble(a2, i)) return false;
          
          return true;
        }
        else// if(cl.equals(Void.TYPE))
        {
          return true;
        }
      }
      else
      {
        for(; i < l; i++)
          if(!areDeeplyEqual(Array.get(a1, i), Array.get(a2, i), map, tmp)) return false;
        
        return true;
      }
    }
    else
    {
      cl = cl.getComponentType();
      for(; i < l; i++)
        if(!areArraysDeeplyEqual(Array.get(a1, i), Array.get(a2, i), map, tmp, cl)) return false;
      
      return true;
    }
  }
  
  /**
   * Checks whether the two lists specified are isomorphic according to 
   * their reference structure.
   * 
   * @param l1     The first list 
   * @param l2     The second list
   * @param map    The map of already matched objects
   * @return       The result of checking
   */
  private static boolean areListsDeeplyEqual(List l1, List l2, IdentityMap map, IdentityMap tmp)
  {
    int s = l1.size();
    if(s != l2.size()) return false;
    
    for(int i = 0; i < s; i++)
      if(!areDeeplyEqual(l1.get(i), l2.get(i), map, tmp)) return false;

    return true;
  }
  
  /**
   * Checks whether the two maps specified are isomorphic according to 
   * their reference structure.
   * 
   * Not implemented completely yet.
   * Just now simply checks that the sizes of maps are equal.
   * 
   * @param m1     The first map
   * @param m2     The second map
   * @param map    The map of already matched objects
   * @return       The result of checking
   */
  private static boolean areMapsDeeplyEqual(Map m1, Map m2, IdentityMap map, IdentityMap tmp)
  {
    int s = m1.size();
    if(s != m2.size()) return false;
    
    IdentityMap tmpMap = new IdentityMap();
    if(tmp != null)
    {
      tmpMap.putAll(tmp);
    }
    
    Object o, p;
    Iterator i, j;
    boolean f;
    
    for(i = m1.keySet().iterator(); i.hasNext(); )
    {
      o = i.next();
      f = false;
      for(j = m2.keySet().iterator(); j.hasNext(); )
      {
        p = j.next();
        
        if(    map.contains(o, p)
            || tmp != null && tmp.contains(o, p)
          )                     
        { 
          if(!areDeeplyEqual(m1.get(o), m2.get(p), map, tmpMap))
          {
            tmpMap.clear();
            tmpMap.putAll(tmp);
            continue;
          }
          else
          {
            f = true;
            if(tmp != null) tmp.putAll(tmpMap);
            else            map.putAll(tmpMap);
            tmpMap.clear();
            break;
          }
        }
        if(map.containsValue(p))                continue;
        if(tmp != null && tmp.containsValue(p)) continue;
        if(   !areDeeplyEqual(o, p, map, tmpMap)
           || !areDeeplyEqual(m1.get(o), m2.get(p), map, tmpMap)
          )
        {
          tmpMap.clear();
          tmpMap.putAll(tmp);
          continue;
        }
        
        f = true;
        if(tmp != null) tmp.putAll(tmpMap);
        else            map.putAll(tmpMap);
        tmpMap.clear();
        break;
      }
      if(!f) return false;
    }
    
    return true;
  }
  
  /**
   * Checks whether the two collections specified are isomorphic according to 
   * their reference structure.
   * 
   * Not implemented completely yet.
   * Just now simply checks that the sizes of collections are equal.
   * 
   * @param c1     The first collection
   * @param c2     The second collection
   * @param map    The map of already matched objects
   * @return       The result of checking
   */
  private static boolean areCollectionsDeeplyEqual(Collection c1, Collection c2, IdentityMap map, IdentityMap tmp)
  {
    int s = c1.size();
    if(s != c2.size()) return false;
    
    IdentityMap tmpMap = new IdentityMap();
    if(tmp != null)
    {
      tmpMap.putAll(tmp);
    }
    
    Object o, p;
    Iterator i, j;
    boolean f;
    
    for(i = c1.iterator(); i.hasNext(); )
    {
      o = i.next();
      f = false;
      for(j = c2.iterator(); j.hasNext(); )
      {
        p = j.next();
        
        if(map.contains(o, p))                     { f = true; break; }
        else if(tmp != null && tmp.contains(o, p)) { f = true; break; }
        else if(map.containsValue(p))                continue;
        else if(tmp != null && tmp.containsValue(p)) continue;
        else if(!areDeeplyEqual(o, p, map, tmpMap))
        {
          tmpMap.clear();
          tmpMap.putAll(tmp);
          continue;
        }
        
        f = true;
        if(tmp != null) tmp.putAll(tmpMap);
        else            map.putAll(tmpMap);
        tmpMap.clear();
        break;
      }
      if(!f) return false;
    }
    
    return true;
  }
  
  /**
   * The method is used to compare two objects on their correspondence it
   * current deep equality check and to start their deep equality check.
   * This method is called only from 
   * {@link DeepSubject#_ts_implDeepEquals(DeepSubject)}.
   * 
   * @param o1    The first object
   * @param o2    The second object
   * @return      The result of object correspondence check
   */
  static boolean areEqual(DeepSubject o1, DeepSubject o2)
  {
    if(!o1._ts_shallowEquals(o2)) return false;

    IdentityMap map = (IdentityMap)tequals.get();
    map.put(o1, o2);
    
    return true;
  }

  /**
   * @param subject
   * @param m
   */
  
  public static <T> T getDeepClone(T o, IdentityMap m)
  {
    if(o == null)
    {
      return null;
    }
    
    if(m.containsKey(o))
    {
      return (T)m.get(o);
    }
    else if(o instanceof DeepSubject)
    {
      DeepSubject s = (DeepSubject)o;
      DeepSubject res = s._ts_shallowClone();
      m.put(o, res);
      
      s._ts_fieldsDeepClone(res, m);
      return (T)res;
    }
    else
    {
      Class cl = o.getClass();
      
      if(cl.isArray())                        return (T)getDeepArrayClone(o, m, cl);
      else if(o instanceof List)              return (T)getDeepListClone((List)o, m, cl);
      else if(o instanceof IList)             return (T)getDeepIListClone((IList)o, m, cl);
      else if(o instanceof IdentityList)      return (T)getDeepIdListClone((IdentityList)o, m);
      else if(o instanceof Map)               return (T)getDeepMapClone((Map)o, m, cl);
      else if(o instanceof IMap)              return (T)getDeepIMapClone((IMap)o, m, cl);
      else if(o instanceof IdentityToIntMap)  return (T)getDeepIdIntMapClone((IdentityToIntMap)o, m);
      else if(o instanceof IntToObjectMap)    return (T)getDeepIntObjMapClone((IntToObjectMap)o, m);
      else if(o instanceof StringToObjectMap) return (T)getDeepStrObjMapClone((StringToObjectMap)o, m);
      else if(o instanceof ISet)              return (T)getDeepISetClone((ISet)o, m, cl);
      else if(o instanceof Collection)        return (T)getDeepCollectionClone((Collection)o, m, cl);
      else if(   o instanceof Number                      // unmodifiable objects
              || o instanceof String 
              || cl.getName().equals("java.lang.Object")
             ) 
      {
        m.put(o, o);
        
        return (T)o;                     
      }
      else if(o instanceof StringBuffer)              // the same for StringBuilder !!!
      {
        Object res = new StringBuffer(o.toString());
        m.put(o, res);
        
        return (T)res;
      }                                
      else
      {
        try
        {
          Method mth; 
          if((mth = cl.getMethod("clone", new Class[]{})) != null)
          {
            ;
          }
          else if(o instanceof Cloneable)
          {
            mth = cl.getDeclaredMethod("clone", new Class[]{});
          }
          else
          {
            return (T)getDeepCommonClone(o, m, cl);
          }
          
          Object res = mth.invoke(o, new Object[]{});
          m.put(o, res);
          
          return (T)res;
        }
        catch(Exception e)
        {
          return (T)getDeepCommonClone(o, m, cl);
        }
      }
    }
  }
    
  private static Object getDeepArrayClone(Object a, IdentityMap map, Class cl)
  {
    if(a == null)
    {
      return null;
    }
    
    int l = Array.getLength(a);

    if(!cl.getComponentType().isArray())
    {
      cl = cl.getComponentType();
      if(cl.isPrimitive())
      {
        if(cl.equals(Boolean.TYPE))
        {
          boolean[] res = new boolean[l];
          map.put(a, res);
          System.arraycopy(a, 0, res, 0, l);
          
          return res;
        }
        else if(cl.equals(Character.TYPE))
        {
          char[] res = new char[l];
          map.put(a, res);
          System.arraycopy(a, 0, res, 0, l);
          
          return res;
        }
        else if(cl.equals(Byte.TYPE))
        {
          byte[] res = new byte[l];
          map.put(a, res);
          System.arraycopy(a, 0, res, 0, l);
          
          return res;
        }
        else if(cl.equals(Short.TYPE))
        {
          short[] res = new short[l];
          map.put(a, res);
          System.arraycopy(a, 0, res, 0, l);
          
          return res;
        }
        else if(cl.equals(Integer.TYPE))
        {
          int[] res = new int[l];
          map.put(a, res);
          System.arraycopy(a, 0, res, 0, l);
          
          return res;
        }
        else if(cl.equals(Long.TYPE))
        {
          long[] res = new long[l];
          map.put(a, res);
          System.arraycopy(a, 0, res, 0, l);
          
          return res;
        }
        else if(cl.equals(Float.TYPE))
        {
          float[] res = new float[l];
          map.put(a, res);
          System.arraycopy(a, 0, res, 0, l);
          
          return res;
        }
        else if(cl.equals(Double.TYPE))
        {
          double[] res = new double[l];
          map.put(a, res);
          System.arraycopy(a, 0, res, 0, l);
          
          return res;
        }
        else// if(cl.equals(Void.TYPE)) - impossible ?!?
        {
          map.put(a, null);
          
          return null;
        }
      }
      else
      {
        Object res = Array.newInstance(cl, l);
        map.put(a, res);

        for(int j = 0; j < l; j++)
          Array.set(res, j, getDeepClone(Array.get(a, j), map));
        
        return res;
      }
    }
    else
    {
      cl = cl.getComponentType();
      Object res = Array.newInstance(cl, l);
      map.put(a, res);
      for(int j = 0; j < l; j++)
        Array.set(res, j, getDeepArrayClone(Array.get(a, j), map, cl));
      
      return res;
    }
  }

  private static List getDeepListClone(List l, IdentityMap map, Class cl)
  {
    List res;
    
    try                
    { 
      res = (List)(cl.newInstance()); 
    }
    catch(Throwable e) 
    { 
      map.put(l, null); 
      return null; 
    }
    
    map.put(l, res);
    for(int i = 0; i < l.size(); i++)
    {
      res.add(getDeepClone(l.get(i), map));
    }
    
    return res;
  }

  private static IList getDeepIListClone(IList l, IdentityMap map, Class cl)
  {
    IList res;
    
    try                
    { 
      res = (IList)(cl.newInstance()); 
    }
    catch(Throwable e) 
    { 
      map.put(l, null); 
      return null; 
    }
    
    map.put(l, res);
    for(int i = 0; i < l.size(); i++)
    {
      res.addLast(getDeepClone(l.get(i), map));
    }
    
    return res;
  }

  private static IdentityList getDeepIdListClone(IdentityList l, IdentityMap map)
  {
    IdentityList res = new IdentityList();
    
    map.put(l, res);
    for(int i = 0; i < l.size(); i++)
    {
      res.addLast(getDeepClone(l.get(i), map));
    }
    
    return res;
  }
  
  private static Map getDeepMapClone(Map m, IdentityMap map, Class cl)
  {
    Map res;
    try                
    { 
      res = (Map)(cl.newInstance()); 
    }
    catch(Throwable e) 
    { 
      map.put(m, null); 
      return null; 
    }
    
    map.put(m, res);
    for(Iterator i = m.keySet().iterator(); i.hasNext(); )
    {
      Object o = i.next();
      res.put(getDeepClone(o, map), getDeepClone(m.get(o), map));
    }
    
    return res;
  }
  
  private static IMap getDeepIMapClone(IMap m, IdentityMap map, Class cl)
  {
    IMap res;
    try                
    { 
      res = (IMap)(cl.newInstance()); 
    }
    catch(Throwable e) 
    { 
      map.put(m, null); 
      return null; 
    }
    
    map.put(m, res);
    for(Iterator i = m.keys().iterator(); i.hasNext(); )
    {
      Object o = i.next();
      res.put(getDeepClone(o, map), getDeepClone(m.get(o), map));
    }
    
    return res;
  }

  private static IdentityToIntMap getDeepIdIntMapClone(IdentityToIntMap m, IdentityMap map)
  {
    IdentityToIntMap res = new IdentityToIntMap();
    
    map.put(m, res);
    for(Iterator i = m.keySet().iterator(); i.hasNext(); )
    {
      Object o = i.next();
      res.put(getDeepClone(o, map), m.get(o));
    }
    
    return res;
  }

  private static IntToObjectMap getDeepIntObjMapClone(IntToObjectMap m, IdentityMap map)
  {
    IntToObjectMap res = new IntToObjectMap();
    
    map.put(m, res);
    for(Iterator i = m.keySet().iterator(); i.hasNext(); )
    {
      int o = (Integer)i.next();
      res.put(o, getDeepClone(m.get(o), map));
    }
    
    return res;
  }

  private static StringToObjectMap getDeepStrObjMapClone(StringToObjectMap m, IdentityMap map)
  {
    StringToObjectMap res = new StringToObjectMap();
    
    map.put(m, res);
    for(StringToObjectMapEnumerator i = m.getEnumerator(); i.moveNext(); )
    {
      String s = (String)i.current().key;
      res.put(s, getDeepClone(m.get(s), map));
    }
    
    return res;
  }
  
  private static Collection getDeepCollectionClone(Collection c, IdentityMap map, Class cl)
  {
    Collection res;
    try                
    { 
      res = (Collection)(cl.newInstance()); 
    }
    catch(Throwable e) 
    { 
      map.put(c, null); 
      return null; 
    }
    
    map.put(c, res);
    for(Iterator i = c.iterator(); i.hasNext(); )
    {
      res.add(getDeepClone(i.next(), map));
    }
    
    return res;
  }

  private static ISet getDeepISetClone(ISet c, IdentityMap map, Class cl)
  {
    ISet res;
    try                
    { 
      res = (ISet)(cl.newInstance()); 
    }
    catch(Throwable e) 
    { 
      map.put(c, null); 
      return null; 
    }
    
    map.put(c, res);
    for(Iterator i = c.iterator(); i.hasNext(); )
    {
      res.add(getDeepClone(i.next(), map));
    }
    
    return res;
  }
  
  public static Object getDeepCommonClone(Object o, IdentityMap map, Class cl)
  {
    Object res = null;
    
    Constructor[] ctrs = cl.getDeclaredConstructors();
    Constructor ctr = null;
    Class[] pars = null;
    Object[] args = null;
    
    for(int i = 0; i < ctrs.length; i++)
    {
      ctr = ctrs[i];
      pars = ctrs[i].getParameterTypes(); 
      if(pars.length == 0)
      {
        args = new Object[]{};
        break;
      }
      else if(pars.length == 1 && pars[0].equals(cl))
      {
        args = new Object[]{ o };
        break;
      }
      else if(pars.length == 1 && pars[0].isPrimitive())
      {
        if(pars[0].equals(Boolean.TYPE))
        {
          args = new Object[]{ new Boolean(true) };
        }
        else if(pars[0].equals(Character.TYPE))
        {
          args = new Object[]{ new Character(' ') };
        }
        else if(pars[0].equals(Byte.TYPE))
        {
          args = new Object[]{ new Byte((byte)1) };
        }
        else if(pars[0].equals(Short.TYPE))
        {
          args = new Object[]{ new Short((short)1) };
        }
        else if(pars[0].equals(Integer.TYPE))
        {
          args = new Object[]{ new Integer(1) };
        }
        else if(pars[0].equals(Long.TYPE))
        {
          args = new Object[]{ new Long(1) };
        }
        else if(pars[0].equals(Float.TYPE))
        {
          args = new Object[]{ new Float(1.0f) };
        }
        else if(pars[0].equals(Double.TYPE))
        {
          args = new Object[]{ new Double(1.0) };
        }
        break;
      }
      else if(pars.length == 1 && pars[0].equals(new Object().getClass()))
      {
        args = new Object[]{ new Object() };
        break;
      }
      else if(pars.length == 1 && pars[0].equals("".getClass()))
      {
        args = new Object[]{"a"};
        break;
      }
    }
    
    if(args != null)
    {
      try
      {
        ctr.setAccessible(true);
        res = ctr.newInstance(args);
        map.put(o, res);
      }
      catch(Throwable e)
      {
        map.put(o, null);
        
        return null;
      }
    }
    else
    {
      map.put(o, null);
      
      return null;
    }
    
    Field[] ff;
    int i;
    Class cf;
    
    try
    {
      while(cl != null)
      {
        ff = cl.getDeclaredFields();
        for(i = 0; i < ff.length; i++)
        {
          if(Modifier.isStatic(ff[i].getModifiers()))
            continue;
          
          ff[i].setAccessible(true);
          
          if((cf = ff[i].getType()).isPrimitive())
          {
            if(cf.equals(Boolean.TYPE))
            {
              ff[i].setBoolean(res, ff[i].getBoolean(o));
            }
            else if(cf.equals(Byte.TYPE))
            {
              ff[i].setByte(res, ff[i].getByte(o));
            }
            else if(cf.equals(Character.TYPE))
            {
              ff[i].setChar(res, ff[i].getChar(o));
            }
            else if(cf.equals(Short.TYPE))
            {
              ff[i].setShort(res, ff[i].getShort(o));
            }
            else if(cf.equals(Integer.TYPE))
            {
              ff[i].setInt(res, ff[i].getInt(o));
            }
            else if(cf.equals(Long.TYPE))
            {
              ff[i].setLong(res, ff[i].getLong(o));
            }
            else if(cf.equals(Float.TYPE))
            {
              ff[i].setFloat(res, ff[i].getFloat(o));
            }
            else if(cf.equals(Double.TYPE))
            {
              ff[i].setDouble(res, ff[i].getDouble(o));
            }
            else // cf.equals(Void.TYPE) - impossible ?!?!
            {
              ;
            }
          }
          else
          {
            ff[i].set(res, getDeepClone(ff[i].get(o), map));
          }
        }
        
        cl = cl.getSuperclass();
      }
    }
    catch (Throwable e)
    {
      map.put(o, null);
      
      return null;
    }
    
    return res; 
  }
  
  public static ClassData getClassDeepClone(String name, IdentityMap m)
  {
    if(name == null || name.equals("")) return null;
    
    Class cl;
    
    try
    {
      cl = Class.forName(name);
    }
    catch (ClassNotFoundException e)
    {
      return null;
    }
    
    ClassData cd = new ClassData(name);
    m.put(cd, cd);
    Field ff[];
    Object o;
    
    ff = cl.getDeclaredFields();
    for(int i = 0; i < ff.length; i++)
    {
      if(!Modifier.isStatic(ff[i].getModifiers()))
          continue;
      
      ff[i].setAccessible(true);
      
      try
      {
        o = ff[i].get(null);
        cd.setField(ff[i].getName(), getDeepClone(o, m));
      }
      catch (Throwable e)
      {
        cd.setField(ff[i].getName(), null);
      }
    }
    
    return cd;
  }
  /*
  public static void deepAppend(Object o, StringBuffer b, IdentityMap m)
  {
    if     (o == null)                  { b.append("null"); return; }
    else if(m.containsKey(o))           { b.append(m.get(o)).append('!'); return; }
    else
    {
      Class        cl = o.getClass();
      int          cd = System.identityHashCode(o);
      StringBuffer bb = new StringBuffer(20);
      appendStringWithType(cl, bb);
      bb.append('@');
      bb.append(Integer.toHexString(cd));
      m.put(o, bb);
    
      if(o instanceof DeepSubject)      { b.append(':'); ((DeepSubject)o)._ts_implDeepAppend(b, m); }
      else if(cl.isArray())             { b.append(':'); deepArrayAppend(o, m, cl, b); }
      else if(o instanceof List)        { b.append(':'); deepListAppend((List)o, m, b); }
      else if(o instanceof Map)         { b.append(':'); deepMapAppend((Map)o, m, b); }
      else if(o instanceof Collection)  { b.append(':'); deepCollectionAppend((Collection)o, m, b); }
      else if(o instanceof Class)       { b.append(':'); appendTypeString((Class)o, b); } 
      else if(   o instanceof String
              || o instanceof StringBuffer
             )                          { b.append(':'); b.append(o); } 
      else
      {
        Class[] cls = new Class[]{};
        Method mth1 = null, mth2 = null;
        try
        {
          mth1 = cl.getMethod("toString", cls);
          mth2 = Class.forName("java.lang.Object").getMethod("toString", cls);
        }
        catch (Throwable e) { }
        
        if(mth2.equals(mth1)) ;
        else { b.append(':'); b.append(o.toString()); }
      }
    }
    
      else
      {
        try
        {
          Method mth; 
          if((mth = cl.getMethod("clone", new Class[]{})) != null)
          {
            ;
          }
          else if(o instanceof Cloneable)
          {
            mth = cl.getDeclaredMethod("clone", new Class[]{});
          }
          else
          {
            return getCommonDeepClone(o, m, cl);
          }
          
          Object res = mth.invoke(o, new Object[]{});
          m.put(o, res);
          
          return res;
        }
        catch(Exception e)
        {
          return getCommonDeepClone(o, m, cl);
        }
      }
    }
    
  }*/
  
  

  
  public static Object restoreFromClone(Object o, IdentityMap m)
  {
    if(o == null || !m.containsKey(o))
    {
      return o;
    }
    else if(o instanceof DeepSubject)
    {
      DeepSubject res = (DeepSubject)m.get(o);
      
      ((DeepSubject)o)._ts_fieldsDeepAssign(res, m);
      return o;
    }
    else
    {
      Class cl = o.getClass();
      
      if(cl.isArray())                        return restoreArrayClone(o, m, cl);
      else if(o instanceof List)              return restoreListClone((List)o, m);
      else if(o instanceof IList)             return restoreIListClone((IList)o, m);
      else if(o instanceof IdentityList)      return restoreIdListClone((IdentityList)o, m);
      else if(o instanceof IntList)           return restoreIntListClone((IntList)o, m);
      else if(o instanceof Map)               return restoreMapClone((Map)o, m);
      else if(o instanceof IMap)              return restoreIMapClone((IMap)o, m);
      else if(o instanceof IdentityToIntMap)  return restoreIdIntMapClone((IdentityToIntMap)o, m);
      else if(o instanceof IntToIntMap)       return restoreIntIntMapClone((IntToIntMap)o, m);
      else if(o instanceof IntToObjectMap)    return restoreIntObjMapClone((IntToObjectMap)o, m);
      else if(o instanceof IntToStringMap)    return restoreIntStrMapClone((IntToStringMap)o, m);
      else if(o instanceof StringToObjectMap) return restoreStrObjMapClone((StringToObjectMap)o, m);
      else if(o instanceof ISet)              return restoreISetClone((ISet)o, m);
      else if(o instanceof IntSet)            return restoreIntSetClone((IntSet)o, m);      
      else if(o instanceof Collection)        return restoreCollectionClone((Collection)o, m);
      else if(   o instanceof Number                      // unmodifiable objects
              || o instanceof String 
              || cl.getName().equals("java.lang.Object")
             ) 
      {
        return o;                     
      }
      else if(o instanceof StringBuffer)              // the same for StringBuilder !!!
      {
        ((StringBuffer)o).delete(0, ((StringBuffer)o).length());
        ((StringBuffer)o).append((StringBuffer)m.get(o));
        
        return o;
      }                                
      else
      {
        return o;
      }
    }
  }
  
  private static Object restoreArrayClone(Object a, IdentityMap map, Class cl)
  {
    if(a == null)
    {
      return null;
    }
    
    int l = Array.getLength(a);
  
    if(!cl.getComponentType().isArray())
    {
      cl = cl.getComponentType();
      if(cl.isPrimitive())
      {
        if(   cl.equals(Boolean.TYPE)
           || cl.equals(Character.TYPE)
           || cl.equals(Byte.TYPE)
           || cl.equals(Short.TYPE)
           || cl.equals(Integer.TYPE)
           || cl.equals(Long.TYPE)
           || cl.equals(Float.TYPE)
           || cl.equals(Double.TYPE)
          )
        {
          Object res = map.get(a);
          System.arraycopy(res, 0, a, 0, l);
          
          return a;
        }
        else// if(cl.equals(Void.TYPE)) - impossible ?!?
        {
          return null;
        }
      }
      else
      {
        Object res = map.get(a);
  
        for(int j = 0; j < l; j++)
          Array.set(a, j, restoreFromClone(Array.get(res, j), map));
        
        return a;
      }
    }
    else
    {
      cl = cl.getComponentType();
      Object res = map.get(a);
      for(int j = 0; j < l; j++)
        Array.set(a, j, restoreArrayClone(Array.get(res, j), map, cl));
      
      return a;
    }
  }
  
  private static List restoreListClone(List l, IdentityMap map)
  {
    List res = (List)map.get(l);
    l.clear();
    
    for(int i = 0; i < res.size(); i++)
    {
      l.add(restoreFromClone(res.get(i), map));
    }
    
    return l;
  }

  private static IList restoreIListClone(IList l, IdentityMap map)
  {
    IList res = (IList)map.get(l);
    l.clear();
    
    for(int i = 0; i < res.size(); i++)
    {
      l.addLast(restoreFromClone(res.get(i), map));
    }
    
    return l;
  }

  private static IdentityList restoreIdListClone(IdentityList l, IdentityMap map)
  {
    IdentityList res = (IdentityList)map.get(l);
    l.clear();
    
    for(int i = 0; i < res.size(); i++)
    {
      l.addLast(restoreFromClone(res.get(i), map));
    }
    
    return l;
  }

  private static IntList restoreIntListClone(IntList l, IdentityMap map)
  {
    IntList res = (IntList)map.get(l);
    l.clear();
    
    for(int i = 0; i < res.size(); i++)
    {
      l.addLast(res.get(i));
    }
    
    return l;
  }
  
  private static Map restoreMapClone(Map m, IdentityMap map)
  {
    Map res  = (Map)map.get(m); 
    m.clear();
    
    for(Iterator i = res.keySet().iterator(); i.hasNext(); )
    {
      Object o = i.next();
      m.put(restoreFromClone(o, map), restoreFromClone(res.get(o), map));
    }
    
    return m;
  }
  
  private static IMap restoreIMapClone(IMap m, IdentityMap map)
  {
    IMap res  = (IMap)map.get(m); 
    m.clear();
    
    for(Iterator i = res.keys().iterator(); i.hasNext(); )
    {
      Object o = i.next();
      m.put(restoreFromClone(o, map), restoreFromClone(res.get(o), map));
    }
    
    return m;
  }

  private static IdentityToIntMap restoreIdIntMapClone(IdentityToIntMap m, IdentityMap map)
  {
    IdentityToIntMap res  = (IdentityToIntMap)map.get(m); 
    m.clear();
    
    for(Iterator i = res.keySet().iterator(); i.hasNext(); )
    {
      Object o = i.next();
      m.put(restoreFromClone(o, map), res.get(o));
    }
    
    return m;
  }

  private static IntToIntMap restoreIntIntMapClone(IntToIntMap m, IdentityMap map)
  {
    IntToIntMap res  = (IntToIntMap)map.get(m); 
    m.clear();
    
    for(Iterator i = res.keySet().iterator(); i.hasNext(); )
    {
      int o = (Integer)i.next();
      m.put(o, res.get(o));
    }
    
    return m;
  }

  private static IntToObjectMap restoreIntObjMapClone(IntToObjectMap m, IdentityMap map)
  {
    IntToObjectMap res  = (IntToObjectMap)map.get(m); 
    m.clear();
    
    for(Iterator i = res.keySet().iterator(); i.hasNext(); )
    {
      int o = (Integer)i.next();
      m.put(o, restoreFromClone(res.get(o), map));
    }
    
    return m;
  }

  private static IntToStringMap restoreIntStrMapClone(IntToStringMap m, IdentityMap map)
  {
    IntToStringMap res  = (IntToStringMap)map.get(m); 
    m.clear();
    
    for(Iterator i = res.keys().iterator(); i.hasNext(); )
    {
      int o = (Integer)i.next();
      m.put(o, res.get(o));
    }
    
    return m;
  }

  private static StringToObjectMap restoreStrObjMapClone(StringToObjectMap m, IdentityMap map)
  {
    StringToObjectMap res  = (StringToObjectMap)map.get(m); 
    m.clear();
    
    for(StringToObjectMapEnumerator i = res.getEnumerator(); i.moveNext(); )
    {
      String o = i.current().key;
      m.put(o, restoreFromClone(res.get(o), map));
    }
    
    return m;
  }
  
  private static Collection restoreCollectionClone(Collection c, IdentityMap map)
  {
    Collection res = (Collection)map.get(c);
    c.clear();
    
    for(Iterator i = res.iterator(); i.hasNext(); )
    {
      c.add(restoreFromClone(i.next(), map));
    }
    
    return c;
  }

  private static ISet restoreISetClone(ISet c, IdentityMap map)
  {
    ISet res = (ISet)map.get(c);
    c.clear();
    
    for(Iterator i = res.iterator(); i.hasNext(); )
    {
      c.add(restoreFromClone(i.next(), map));
    }
    
    return c;
  }

  private static IntSet restoreIntSetClone(IntSet c, IdentityMap map)
  {
    IntSet res = (IntSet)map.get(c);
    c.clear();
    
    for(Iterator i = res.iterator(); i.hasNext(); )
    {
      c.add((Integer)i.next());
    }
    
    return c;
  }
  
  /**
   * @param subject
   * @param m
   */
  public static void restore (DeepSubject o, IdentityMap m)
  {
    restoreFromClone(o, m);
  }

  /**
   * @param subject
   * @param m
   * @return
   */
  public static boolean compare (DeepSubject subject, IdentityMap m)
  {
    // TODO Auto-generated method stub
    return true;
  }
}
