package jatva.lang.synchronization;

import jatva.lang.AbstractAction;
import jatva.lang.CompositeOracleException;
import jatva.lang.InsufficientAccessibilityError;
import jatva.lang.LostExceptionError;
import jatva.lang.ModelCall;
import jatva.lang.ModelEvent;
import jatva.lang.ModelEventProperties;
import jatva.lang.ModelObject;
import jatva.lang.ModelReaction;
import jatva.lang.NoCorrectSeriesException;
import jatva.lang.PostconditionViolationException;
import jatva.lang.PreconditionViolationException;
import jatva.lang.ResultTracingException;
import jatva.lang.TestException;
import jatva.lang.TestingModel;
import jatva.lang.TranslationError;
import jatva.lang.coverage.SEOperationInfo;
import jatva.tracer.Tracer;
import jatva.tracer.messages.BranchMessage;
import jatva.tracer.messages.CoverageStructureMessage;
import jatva.tracer.messages.DeferredBranchMessage;
import jatva.tracer.messages.ImplementationCallEnterMessage;
import jatva.tracer.messages.ImplementationCallExitMessage;
import jatva.tracer.messages.ImplementationExceptionMessage;
import jatva.tracer.messages.ImplementationMessage;
import jatva.tracer.messages.ImplementationResultMessage;
import jatva.tracer.messages.InstanceInvariantStartMessage;
import jatva.tracer.messages.InternalErrorMessage;
import jatva.tracer.messages.InvariantEndMessage;
import jatva.tracer.messages.JumpEndMessage;
import jatva.tracer.messages.JumpStartMessage;
import jatva.tracer.messages.MarkMessage;
import jatva.tracer.messages.ModelEventEndMessage;
import jatva.tracer.messages.ModelEventMessage;
import jatva.tracer.messages.ModelMessage;
import jatva.tracer.messages.OracleExitMessage;
import jatva.tracer.messages.OraclePostcodnitionEnterMessage;
import jatva.tracer.messages.OraclePrecodnitionEnterMessage;
import jatva.tracer.messages.PreconditionExitMessage;
import jatva.tracer.messages.PrimeFormulaMessage;
import jatva.tracer.messages.ScenarioEndMessage;
import jatva.tracer.messages.ScenarioMessage;
import jatva.tracer.messages.ScenarioMethodMessage;
import jatva.tracer.messages.ScenarioStartMessage;
import jatva.tracer.messages.ScenarioValueMessage;
import jatva.tracer.messages.SerializationEndMessage;
import jatva.tracer.messages.SerializationStartMessage;
import jatva.tracer.messages.SimpleUserInfoMessage;
import jatva.tracer.messages.StateMessage;
import jatva.tracer.messages.StaticInvariantEnterMessage;
import jatva.tracer.messages.TaggedUserInfoMessage;
import jatva.tracer.messages.TestExceptionMessage;
import jatva.tracer.messages.TestFrameEndMessage;
import jatva.tracer.messages.TestFrameEnterMessage;
import jatva.tracer.messages.TestFrameExitMessage;
import jatva.tracer.messages.TestFrameStartMessage;
import jatva.tracer.messages.TestMessage;
import jatva.tracer.messages.TestThreadStartMessage;
import jatva.tracer.messages.TracerMessage;
import jatva.tracer.messages.TransitionEndMessage;
import jatva.tracer.messages.TransitionStartMessage;
import jatva.tracer.messages.ValidExceptionMessage;
import jatva.util.CtorInfo;
import jatva.util.IdentitySet;
import jatva.util.MethodInfo;
import jatva.util.OperationInfo;
import jatva.util.Utils;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.LinkedList;
import java.util.Stack;

/**
 *
 * @author Victor Kuliamin
 *
 */
public class TestManager
{
  private static ThreadLocal<TestManager> threadManager = new ThreadLocal<TestManager>();

  private Test   test   = null;
  private Tracer tracer = null;

  /**
   * debug only field
   */
  private String threadName = null;

  private TestManager ()
  {
    this.threadName = Thread.currentThread().getName();
    test   = Test.getInitTest(this);
    tracer = test.getTracer();

    if( test.isManagerTheMain( this ) )
    {
        setMainMessageBuffer();
    } else {
        setNewBufferredMessageBuffer();
    }
    pushSelfAsChannelId();
    traceMessagesBuffer.addNew( new TestThreadStartMessage( this ) );
  }

  //-----------------------------------------------------------
  // General methods for test system management

  /** Get test manager for the current thread */
  public final static TestManager getRegistrar ()
  {
    TestManager m;

    synchronized (TestManager.class)
    {
      m = threadManager.get();
      if(m == null)
      {
          threadManager.set(m = new TestManager());
      }
    }

    return m;
  }

  // ????
  public boolean registerObjectForTest(Object o)
  {
    return test.registerObject(o);
  }

  // ????
  public static TestManager startTest ()
  {
    TestManager m = getRegistrar();

    return m;
  }

  // ????
  public static void endTest ()
  {

  }

  /**
   *  Returns status of the including test
   *  Status can be EXECUTING or CHECKING
   */
  public int getStatus ()
  {
    return test.getStatus();
  }

  /**
   * Returns mode of the including test.
   * Mode can be SYNCHRONOUS, ASYNCHRONOUS and CONCURRENT
   * @return
   */
  public int getMode()
  {
    return test.getMode();
  }

  public void setSeekAllSeries( boolean s )
  {
      test.setSeekAllSeries( s );
  }

  public void setAsynchronousFrameMode()
  {
      test.setFrameExecutionMode( Test.ASYNCHRONOUS );
  }

  public void setAsynchronousMode()
  {
    test.setMode( Test.ASYNCHRONOUS );
  }

  public void setSynchronousMode()
  {
    test.setMode( Test.SYNCHRONOUS );
  }

  /**
   *  Returns true if the current status of the including test is EXECUTING
   */
  public boolean isExecuting()
  {
    return test.getStatus() == Test.EXECUTING;
  }

  /**
   *   Returns the tracer object used by this test manager
   */
  public Tracer getTracer ()
  {
    return tracer;
  }

  /**
   * Returns the number identifier of this test thread in JVM process
   */
  public int getId()
  {
    return test.getManagerId(this);
  }

  /**
   * Returns the number identifier of test including this test thread in JVM process
   */
  public int getTestId()
  {
    return test.getId();
  }

  /**
   *  Returns the number identifier of this test thread inside the including test
   */
  public int getLocalId()
  {
    return test.getLocalId(this);
  }

  /**
   * Returns the including test
   */
  public Test getTest()
  {
    return test;
  }

  /**
   * Returns timeout value set for this test frame
   * @return
   */
  public long getFrameTimeout()
  {
    return test.getFrameTimeout();
  }

  /**
   * Sets timeout value for this frame in milliseconds
   * @param millis
   */
  public void setFrameTimeout(long millis)
  {
    test.setFrameTimeout(millis);
  }

  public void setCertaintyBasedOnModelCalls()
  { test.setCertaintyBasedOnModelCalls(); }

  public void setCertaintyBasedOnFrameStarts()
  { test.setCertaintyBasedOnFrameStarts(); }

  //------------------------------------------------------
  // Implementation of postponed transition actions tracing

  // Identifier of the current transition
  int          transitionId;

  TraceMessagesBuffer traceMessagesBuffer = new TraceMessagesBuffer();

  private void setMainMessageBuffer()
  { this.traceMessagesBuffer = test.getTraceMessageBuffer(); }

  private void setNewBufferredMessageBuffer()
  {
      this.traceMessagesBuffer = new TraceMessagesBuffer();
      this.traceMessagesBuffer.switchToBufferredTracing();
  }

  //------------------------------------------------------------
  // Indicators of test events

  // ???? Indicates start of the test scenario with no corresponding TestingModel object
  public void scenarioStart ()
  {

  }

  /**
   * Indicates start of the test scenario corresponding to tm object
   *
   */
  public void scenarioStart (TestingModel tm)
  {
    restoreManagerState();
    traceMessagesBuffer.addNew( new ScenarioStartMessage( this, tm ) );
  }

  /**
   * Restores TestManager's state to the initial one.
   * That is to the state just after this manager was created.
   */
  private void restoreManagerState()
  {
      test.setTransitionPerformed( true );
  }

  /* ????
  public void scenarioEnd ()
  {
    tracer.traceScenarioEnd(this);
  }*/

  /**
   * Indicates entering into a state
   */
  public void stateEnter (int id, Object state)
  {
    if( test.isTransitionPerformed() )
    {
      traceMessagesBuffer.addNew( new StateMessage( this, id, state ) );
      test.setTransitionPerformed( false );
    }
  }

  /**
   * Informs TestManager thar transition is starting to do something useful
   * so it should be properly traced.
   */
  public void confirmTransition()
  {
      test.setTransitionPerformed( true );
  }

  /**
   * Indicates start of transition execution
   * @param id
   * @param model
   * @param action
   */
  public void transitionStart (int id, TestingModel model, AbstractAction action)
  {
    assert test.isManagerTheMain( this );
    assert test.getTraceMessageBuffer() == this.traceMessagesBuffer;

    transitionId = id;

    traceMessagesBuffer.switchToBufferredTracing();
    traceMessagesBuffer.addNew( new TransitionStartMessage( this, transitionId ) );

    model._ts_traceAction(action);
  }

  /**
   * Indicates end of transition
   */
  public void transitionEnd ()
  {
    assert test.isManagerTheMain( this );
    assert test.getTraceMessageBuffer() == this.traceMessagesBuffer;

    if( test.isTransitionPerformed() )
    {
        traceMessagesBuffer.switchToImmediateTracing();

        traceMessagesBuffer.addNew( new TransitionEndMessage( this ) );
    }
    else
    {
        // here despite of the fact that all messages will be ignored we still
        // need to buffer 'TransitionEnd' message. the reason is that
        // formatters usually perform some indent at the start of the transinition
        // and we should allow them to make the corresponding outdent at transition end.

        traceMessagesBuffer.addNew( new TransitionEndMessage( this ) );
        traceMessagesBuffer.clear();

        traceMessagesBuffer.switchToImmediateTracing();
        traceMessagesBuffer.addNew( new ScenarioMessage( this, "Empty transition omitted" ) );
    }

  }

  /**
   * Indicates jump transition start
   * @param kind
   */
  public void jumpStart (String kind)
  {
    traceMessagesBuffer.addNew( new JumpStartMessage( this, kind ) );
  }

  /**
   * Indicates jump transition end
   *
   */
  public void jumpEnd ()
  {
    traceMessagesBuffer.addNew( new JumpEndMessage( this ) );
  }

  /**
   * Register model object in test pool for model state restoration
   * @param r
   */
  public void registerFrameStorage(Object r)
  {
    assert test.isManagerTheMain( this );

    test.registerFrameStorage(r);

    if( test.getMode() == Test.ASYNCHRONOUS )
    {
        // we really start the frame only in frame execution mode is ASYNHCORNOUS
        traceMessagesBuffer.addNew( new TestFrameStartMessage( this ) );

        if( test.isTransitionPerformed() )
        {
            traceMessagesBuffer.switchToImmediateTracing();
        }

        // we should substitute message buffer only for the main test thread
        // see assertion above
        setNewBufferredMessageBuffer();
    }
  }

  /**
   * Indicates enter of this test thread into test frame
   * This method is not currently executed by the main
   * test thread. It's executed only by supplementary test threads
   * such as actors. The same is true for 'testFrameExit'.
   */
  public void testFrameEnter ()
  {
    assert !(test.isManagerTheMain( this ));

    traceMessagesBuffer.addNew( new TestFrameEnterMessage( this ) );
    //test.registerFrame(this);
    //tracer.traceTestFrameEnter(this);
  }

  /**
   * Indicates exit of this test thread from test frame
   *
   */
  public void testFrameExit ()
  {
    assert !(test.isManagerTheMain( this ));

    traceMessagesBuffer.addNew( new TestFrameExitMessage( this ) );
    //test.unregisterFrame(this);
    /*
    if(test.unregisterFrame(this))
    {
      while(true)
      {
        try
        {
          test.getFrameLatch().acquire();  // wait util the test frame end - all other threads entered it should exit
          break;
        }
        catch (InterruptedException e)
        {
          tracer.traceException(this, e, "Test thread is interrupted while waiting test frame end", null, null);
          // ?
        }
      }
    }*/
  }

  /**
   * This method completes messages stored during the frame such that
   * the complete messages list is correct according to the trace format.
   * For example, this method inserts fake 'return-s' (i.e. model_operation_end)
   * for stimuli that hadn't finished by the end of the frame.
   *
   * @param messages   trace messages accumulated during the test frame
   * @param originalBuffer   the buffer from which those message were extracted. Original
   *                         buffer is required 'cause some messages may be
   *                         transferred to the next frame thus being returned back to
   *                         the buffer.
   */
  private void completeTraceBuffer( LinkedList<TracerMessage> messages
                                  , TraceMessagesBuffer originalBuffer
                                  )
  {
      // TODO: implement this method!!!
  }

  public void prepareMessageBuffers()
  {
      // 'this' test manager is the main test manager of the test !!!!
      // see assertion in 'closeTestFrame'

      IdentitySet keySet = test.getManagers().keySet();
      for( Object o : keySet )
      {
          TestManager tm = (TestManager)o;

          // we're inside test frame completion, if and only if
          // the main test thread had completed all its invocations

          LinkedList<TracerMessage> messages = tm.traceMessagesBuffer.extractBufferedMessages();
          completeTraceBuffer( messages, tm.traceMessagesBuffer );

          if( tm == this )
          {
              // message buffer of the main test thread should be empty
              // ( all events in that buffer should be completed, so
              //   no messages should be transfered to the next frame in that
              //   message buffer )
              assert this.traceMessagesBuffer.isEmpty();

              setMainMessageBuffer();
          }

          test.getTraceMessageBuffer().addAll( messages );
      }

      if( test.isTransitionPerformed() )
      {
          // here this.traceMessageBuffer == test.getTraceMessageBuffer
          traceMessagesBuffer.switchToImmediateTracing();
      }
  }

  /**
   * ????
   *
   */
  public void closeTestFrame()
  {
    assert test.isManagerTheMain( this );

    boolean was_asynch_mode = test.getMode() == Test.ASYNCHRONOUS;

    // 'prepareMessageBuffers' is invoked inside 'closeFrame' of 'Test'
    test.closeFrame();

    if( was_asynch_mode )
    {
        traceMessagesBuffer.addNew( new TestFrameEndMessage( this ) );
    }
  }

  public void serializationStart()
  {
      traceMessagesBuffer.addNew( new SerializationStartMessage( this ) );
  }

  public void serializationEnd()
  {
      traceMessagesBuffer.addNew( new SerializationEndMessage( this ) );
  }

  private Class<?> currentMediator = null;
  private Stack<Class<?>> mediatorStack = new Stack<Class<?>>();

  public void setCurrentMediator( Class<?> mediatorClass )
  {
      if( currentMediator == null )
          currentMediator = mediatorClass;
      else
          mediatorStack.push( mediatorClass );
  }

  public void unsetCurrentMediator()
  {
      if( mediatorStack.size() == 0 )
          currentMediator = null;
      else
          currentMediator = mediatorStack.pop();
  }

  private Stack<Object> channelIdStack = new Stack<Object>();

  public void pushChannelId( Object channelId )
  { channelIdStack.push( channelId ); }

  public void popChannelId()
  { channelIdStack.pop(); }

  public void pushSelfAsChannelId()
  { channelIdStack.push( this.threadName ); }

  public void registerModelEvent( ModelEvent e, ModelEventProperties e_props )
  {
    if( e_props == null )
      e_props = ModelEventProperties.Empty;

    Object givenChannelId = e_props.getChannelID();

    if( givenChannelId == ModelEventProperties.defaultChannelId )
      givenChannelId = this.threadName;

    try
    {
      // here we store channelId in channel stack to make
      // all nested model events registered in the same channel
      // as the current (enclosing) one
      if( givenChannelId != null )
      {
        pushChannelId( givenChannelId );
      }

      e.addSynchronizationObject( e_props.getSynchronizationObject() );

      e.setManager(this);
      e.setChannelId( channelIdStack.peek() );

      test.registerEvent(e);

      traceMessagesBuffer.addNew( new ModelEventMessage( this, e ) );

      setCurrentEvent(e);

      executeEvent(e);
    }
    finally
    {
      if( givenChannelId != null )
        popChannelId();
    }
  }

  private void executeEvent(ModelEvent e)
  {
    boolean performOracleCheck = true;

    synchronized (test)
    {
      int currentMode = test.getMode();

      if (currentMode == Test.SYNCHRONOUS) performOracleCheck = true;
      else if (currentMode == Test.WEAK_SYNCHRONOUS)
      {
        test.setMode(Test.ASYNCHRONOUS);
        performOracleCheck = false;
      }
      else performOracleCheck = false;
    }

    try
    {
      if(performOracleCheck && e.getObject()._ts_isOracleSet())
      {
        if(!e.check())
        {
          throw getOracleException();
        }
      }
      else
      {
        e.execute();
      }
    }
    finally
    {
      traceMessagesBuffer.addNew( new ModelEventEndMessage( this, currentEvent ) );

      removeCurrentEvent();
    }

  }

  /*
  public void registerModelEvent (ModelEvent e)
  {
    e.setManager(this);
    test.registerEvent(this, e);
    tracer.traceModelReaction(this, e);
  }*/

  /*
   * public void storeModelArgument (String name, Object a) {} public void
   * storeModelArgumentBoolean (String name, boolean a) {} public void
   * storeModelArgumentByte (String name, byte a) {} public void
   * storeModelArgumentShort (String name, short a) {} public void
   * storeModelArgumentCharacter(String name, char a) {} public void
   * storeModelArgumentInteger (String name, int a) {} public void
   * storeModelArgumentLong (String name, long a) {} public void
   * storeModelArgumentFloat (String name, float a) {} public void
   * storeModelArgumentDouble (String name, double a) {}
   *
   * public void storeModelMArgument (String name, Object[] a) {} public void
   * storeModelMArgumentBoolean (String name, boolean[] a) {} public void
   * storeModelMArgumentByte (String name, byte[] a) {} public void
   * storeModelMArgumentShort (String name, short[] a) {} public void
   * storeModelMArgumentCharacter(String name, char[] a) {} public void
   * storeModelMArgumentInteger (String name, int[] a) {} public void
   * storeModelMArgumentLong (String name, long[] a) {} public void
   * storeModelMArgumentFloat (String name, float[] a) {} public void
   * storeModelMArgumentDouble (String name, double[] a) {}
   */
  public void storeModelField (String name, Object f)
  {}

  public void storeModelFieldBoolean (String name, boolean f)
  {}

  public void storeModelFieldByte (String name, byte f)
  {}

  public void storeModelFieldShort (String name, short f)
  {}

  public void storeModelFieldCharacter (String name, char f)
  {}

  public void storeModelFieldInteger (String name, int f)
  {}

  public void storeModelFieldLong (String name, long f)
  {}

  public void storeModelFieldFloat (String name, float f)
  {}

  public void storeModelFieldDouble (String name, double f)
  {}

  /*
   * public void storeModelResult (Object r) {} public void
   * storeModelResultBoolean (boolean r) {} public void storeModelResultByte
   * (byte r) {} public void storeModelResultShort (short r) {} public void
   * storeModelResultCharacter(char r) {} public void storeModelResultInteger
   * (int r) {} public void storeModelResultLong (long r) {} public void
   * storeModelResultFloat (float r) {} public void storeModelResultDouble
   * (double r) {}
   *
   * public void storeModelException (Throwable e) {}
   */

  private boolean isLastEventInSequence = false;

  private ModelEvent currentEvent = null;

  public ModelEvent getCurrentEvent ()
  {
    return currentEvent;
  }

  public void setCurrentEvent (ModelEvent e)
  {
    currentEvent = e;
  }

  public void eventIsTheLastInSequence()
  {
      isLastEventInSequence = true;
  }

  public void removeCurrentEvent ()
  {
    currentEvent = null;
    isLastEventInSequence = false;
  }

  public void storeResult(Object res)
  {
    currentEvent.storeResult(res);
  }

  public void storeException(Throwable exception)
  {
    currentEvent.setException(exception);
  }

  public boolean isImplementationExceptionOccurred()
  {
    return currentEvent.isImplException();
  }

  public void implementationStart()
  {
    ModelEvent e = getCurrentEvent();

    if( e instanceof ModelReaction )
        e.update();
  }

  public void implementationExit()
  {

    if(test.getMode() == Test.SYNCHRONOUS)
    {
      //traceMessage("    PERFORMING COMMON UPDATE");

      // causes invocation of mediators' mapStateUp-s
      test.getSynchronizationManager().update();
    }

    ModelEvent e = getCurrentEvent();

    if( e instanceof ModelCall )
        e.update();
  }

  /**
   * this method is invoked by oracles after appropriate update of model state
   * @param classLevel - equals to the level of the class which oracle is requesting
   *                     invariant checks
   */
  public void checkInvariants( int classLevel )
  {
      // invariants will be checked only by the request from the lowest oracle
      // in model classes hierarchy
      if( classLevel != currentEvent.getInvariantCheckClassLevel() )
          return;

      if( test.getMode() == Test.SYNCHRONOUS )
      {
        // if test is in synchornous mode its status shoud always
        // be equal to 'executing'
        assert test.getStatus() == Test.EXECUTING;

        test.getSynchronizationManager().check();

      } else
      if(    test.getMode() == Test.ASYNCHRONOUS
          && test.getStatus() == Test.CHECKING
        )
      {
        // this means we're in series now.
        if( isLastEventInSequence )
            test.getSynchronizationManager().check();
        else
            test.getSynchronizationManager().checkPrime();
      }
  }

  public void checkAllInvariants ()
  {
    if(test.getMode() == Test.SYNCHRONOUS)
    {
      test.getSynchronizationManager().check();
    }
    // called after the end of mediator method execution or synchronization
    // in synchronious testing should call update during execution
    // in asynchronous - during checking
  }

  public void checkPrimeInvariants ()
  {
    if(test.getMode() == Test.SYNCHRONOUS)
    {
      test.getSynchronizationManager().checkPrime();
    }
    // called after the end of mediator method execution or synchronization
    // in synchronious testing should call update during execution
    // in asynchronous - during checking
  }

  public void traceValue (String exp, String val)
  {
    StringBuffer s = new StringBuffer(exp);
    traceMessagesBuffer.addNew( new ScenarioMessage( this, s.append(" = ").append( val ).toString() ) );
  }
  /*
  public void modelOperationExit ()
  {

  }

  public void modelEventOccurence ()
  {

  }
  public void inverseOperationEnter ()
  {

  }

  public void inverseOperationExit ()
  {

  }

  public void serializationStart ()
  {

  }

  public void serializationEnd ()
  {

  }

  public void oracleEnter (ModelObject obj, String signature)
  {
    tracer.traceOracleEnter(this, signature);
  }

  public void oracleExit ()
  {
    tracer.traceOracleExit(this);
  }*/

  /*
  public void preconditionEnter ()
  {
    System.out.println(id()
        + Integer.toHexString(System.identityHashCode(this))
        + ": enter the precondition: " + getCurrentEvent());

  }*/

  public void preconditionEnter()/*ModelEvent e*/
  {
    traceMessagesBuffer.setSuppressUserInfo( true );
  }

  public void preconditionExit()
  {
    traceMessagesBuffer.setSuppressUserInfo( false );
  }

  public void coverageAnalysisPreconditionEnter ()/*ModelEvent e*/
  {
      traceMessagesBuffer.setSuppressUserInfo( true );
  }

  public void coverageAnalysisPreconditionExit()
  {
  }

  public void coverageAnalysisPostconditionEnter()
  {

  }

  public void coverageAnalysisPostconditionExit()
  {
      traceMessagesBuffer.setSuppressUserInfo( false );
  }

  public void oraclePreconditionEnter( int classLevel )
  {
    SEOperationInfo op = currentEvent.getOperation().getRefinedOperation( classLevel );
    String signature = op.getPrettySignature();

    traceMessagesBuffer.addNew( new OraclePrecodnitionEnterMessage
                                     ( this
                                     , signature
                                     , currentEvent.getId()
                                     )
                           );

    if( !(test.isCalledOperation( op )) )
    {
      traceMessagesBuffer.addNew( new CoverageStructureMessage( this, op ) );
    }
  }

  public void oraclePreconditionExit ( boolean pre_vedict )
  {
    ModelObject mo = currentEvent.getObject();
    if( pre_vedict )
    {
        mo._ts_setAnyPreHolds();
    } else {
        if( !(mo._ts_isAnyPreHolds()) )
        {
            currentEvent.reducedInvariantCheckClassLevel();
        }
    }

    traceMessagesBuffer.addNew( new PreconditionExitMessage( this ) );
  }

//  public void allPreconditionsFailed (ModelObject obj)  { }
/*
  public void coverageElementCalculationEnter (ModelObject obj, String method)
  {

  }

  public void coverageElementCalculationExit ()
  {

  }*/

  public void oraclePostconditionEnter( int classLevel )
  {
    String signature = currentEvent.getOperation().getRefinedOperation( classLevel ).getPrettySignature();

    traceMessagesBuffer.addNew( new OraclePostcodnitionEnterMessage( this, signature ) );
  }

  public void oraclePostconditionExit ()
  {
    traceMessagesBuffer.addNew( new OracleExitMessage( this ) );
  }

  private TestException oracleException = null;

  public boolean isAnyPostFails()
  {
      return currentEvent.getObject()._ts_isAnyPostFails();
  }

  public void addPreVerdictFailureComment( Object fc )
  {
      currentEvent.addPreVerdictFailureComment( fc );
  }

  public void preconditionViolation( boolean isStatic )
  {
      String signature = currentEvent.getOperation().getPrettySignature();
      ModelObject mo = currentEvent.getObject();

      if( mo._ts_isAnyPreHolds() )
      {
          invokeEventImplementation();
      } else {

          String failureComment = currentEvent.getPreVerdictFailureComment();

          if( isStatic )
          {
              if( failureComment == null )
              {
                  oracleException = new PreconditionViolationException
                                              ( currentEvent.getOperation().getDeclaringClass()
                                              , signature
                                              );
              } else {
                  oracleException = new PreconditionViolationException
                                              ( currentEvent.getOperation().getDeclaringClass()
                                              , signature
                                              , failureComment
                                              );
              }
          } else {

              if( failureComment == null )
              {
                  oracleException = new PreconditionViolationException
                                              ( currentEvent.getObject()
                                              , signature
                                              );
              } else {
                  oracleException = new PreconditionViolationException
                                              ( currentEvent.getObject()
                                              , signature
                                              , failureComment
                                              );
              }
          }
      }
  }

  public void addPostVerdictFailureComment( Object fc )
  {
      currentEvent.addPostVerdictFailureComment( fc );
  }

  public void postconditionViolation( int classLevel, boolean isStatic )
  {
      ModelObject mo = currentEvent.getObject();

      mo._ts_setAnyPostFails();
      SEOperationInfo op = currentEvent.getOperation().getRefinedOperation( classLevel );
      String signature = op.getPrettySignature();

      String failureComment = currentEvent.getPostVerdictFailureComment();

      if( isStatic )
      {
          if( failureComment == null )
          {
              oracleException = new PostconditionViolationException( mo, signature );
          } else {
              oracleException = new PostconditionViolationException( mo, signature, failureComment );
          }
      } else {
          if( failureComment == null )
          {
              oracleException = new PostconditionViolationException( op.getDeclaringClass()
                                                                   , signature
                                                                   );
          } else {
              oracleException = new PostconditionViolationException( op.getDeclaringClass()
                                                                   , signature
                                                                   , failureComment
                                                                   );
          }
      }
  }

  public void registerOracleException(TestException error)
  {
    oracleException = error;
  }

  public TestException getOracleException()
  {
    return oracleException;
  }

  /*
  private boolean createResultEvent = false;

  public boolean createResultEvent ()
  {
    return createResultEvent;
  }*/

  public void branchExecution (String branch, String msg )
  {
    traceMessagesBuffer.addNew( new BranchMessage( this, branch, msg ) );
  }

  public void invokeEventImplementation()
  {
    if( isExecuting() )
    {
      try
      {
        implementationStart();
        currentEvent.execute();
        implementationExit();
      }
      catch( jatva.lang.TestException _lv_error )
      {
        if( !(_lv_error.isCritical()) )
          implementationExit();
        throw _lv_error;
      }
      catch( Throwable _lv_error )
      {
        currentEvent.setException( _lv_error );
        implementationExit();
      }
    }
    else
    {
      implementationStart();
      implementationExit();
    }
  }

  public void deferredBranchExecution (String branch, String msg)
  {
    traceMessagesBuffer.addNew( new DeferredBranchMessage( this, branch, msg ) );
    //createResultEvent = true;
  }

  public void markExecution (String id, String label)
  {
    traceMessagesBuffer.addNew( new MarkMessage( this, (id == null)?label:(id + "." + label) ) );
  }

  public boolean formulaValue (int id, boolean value)
  {
    traceMessagesBuffer.addNew( new PrimeFormulaMessage( this, id, value ) );
    return value;
  }

  // auxiliary method. useful for writing manual translataions of
  // experimental constructs
//  public boolean formulaValue(String id, boolean value)
//  {
//    tracer.tracePrimeFormula(this, id, value);
//    return value;
//  }

  public void coverageItem ()
  {

  }
/*
  public void updateEnter ()
  {

  }

  public void updateExit ()
  {

  }
*/

  public void invariantEnter (ModelObject obj, String name, String msg)
  {
    traceMessagesBuffer.addNew( new InstanceInvariantStartMessage( this, obj, name, msg ) );
  }

  public void invariantEnter (String p_class, String name, String msg)
  {
    traceMessagesBuffer.addNew( new StaticInvariantEnterMessage( this, p_class, name, msg ) );
  }

  public void invariantExit ()
  {
      traceMessagesBuffer.addNew( new InvariantEndMessage( this ) );
  }

  public void implementationFrameEnter ()
  {

  }

  public void implementationFrameExit ()
  {

  }

  public void implementationCallEnter( OperationInfo op_info, Object obj, Object[] args )
  {
    traceMessagesBuffer.addNew( new ImplementationCallEnterMessage( this, op_info, obj, args ) );
  }

  public void implementationCallExit (long stime, long etime)
  {
    traceMessagesBuffer.addNew( new ImplementationCallExitMessage( this, stime, etime ) );
  }

  public void registerImplementationResult(String staticResultType, Object result)
  {
    traceMessagesBuffer.addNew( new ImplementationResultMessage( this, staticResultType, result ) );
  }

//  public void registerImplementationResult(boolean result, long time)
//  {
//  }
//
//  public void registerImplementationResult(byte result, long time)
//  {
//  }
//
//  public void registerImplementationResult(char result, long time)
//  {
//  }
//
//  public void registerImplementationResult(short result, long time)
//  {
//  }
//
//  public void registerImplementationResult(int result, long time)
//  {
//  }
//
//  public void registerImplementationResult(long result, long time)
//  {
//  }
//
//  public void registerImplementationResult(float result, long time)
//  {
//  }
//
//  public void registerImplementationResult(double result, long time)
//  {
//  }

  public void registerImplementationException(Throwable error)
  {
    // First state the implementation exception flag in the current event!!!!
    currentEvent.setImplExceptionFlag();
    currentEvent.setImplException(error);
    traceMessagesBuffer.addNew( new ImplementationExceptionMessage( this, error ) );
  }

  public Throwable getImplementationException()
  {
    // Return implementation exception
    return currentEvent.getImplException();
  }

  public void registerScenarioMethod ( TestingModel object, String method, AbstractAction action )
  {
      traceMessagesBuffer.addNew( new ScenarioMethodMessage( this, object, method ) );
  }

  public void registerIterationVariableValue ( jatva.tracer.ValueKind kind, String type, String name, boolean isPrimitive, Object value )
  {
      traceMessagesBuffer.addNew( new ScenarioValueMessage( this, kind.toString(), type, name, isPrimitive, value ) );
  }

//  public void registerIterationVariableValue ( jatva.tracer.ValueKind kind, String type, String name, boolean value)
//  {
//    if(!transitionPerformed)
//      transitionTrace.addLast(new ScenarioValueMessage(kind, type, name, true, Utils.toObject(value)));
//    else
//      tracer.traceScenarioValue(this, kind, type, name, true, Utils.toObject(value));
//  }
//
//  public void registerIterationVariableValue ( jatva.tracer.ValueKind kind, String type, String name, byte value)
//  {
//    if(!transitionPerformed)
//      transitionTrace.addLast(new ScenarioValueMessage(kind, type, name, true, Utils.toObject(value)));
//    else
//      tracer.traceScenarioValue(this, kind, type, name, true, Utils.toObject(value));
//  }
//
//  public void registerIterationVariableValue ( jatva.tracer.ValueKind kind, String type, String name, char value)
//  {
//    if(!transitionPerformed)
//      transitionTrace.addLast(new ScenarioValueMessage(kind, type, name, true, Utils.toObject(value)));
//    else
//      tracer.traceScenarioValue(this, kind, type, name, true, Utils.toObject(value));
//  }
//
//  public void registerIterationVariableValue ( jatva.tracer.ValueKind kind, String type, String name, short value)
//  {
//    if(!transitionPerformed)
//      transitionTrace.addLast(new ScenarioValueMessage(kind, type, name, true, Utils.toObject(value)));
//    else
//      tracer.traceScenarioValue(this, kind, type, name, true, Utils.toObject(value));
//  }
//
//  public void registerIterationVariableValue ( jatva.tracer.ValueKind kind, String type, String name, int value)
//  {
//    if(!transitionPerformed)
//      transitionTrace.addLast(new ScenarioValueMessage(kind, type, name, true, Utils.toObject(value)));
//    else
//      tracer.traceScenarioValue(this, kind, type, name, true, Utils.toObject(value));
//  }
//
//  public void registerIterationVariableValue ( jatva.tracer.ValueKind kind, String type, String name, long value)
//  {
//    if(!transitionPerformed)
//      transitionTrace.addLast(new ScenarioValueMessage(kind, type, name, true, Utils.toObject(value)));
//    else
//      tracer.traceScenarioValue(this, kind, type, name, true, Utils.toObject(value));
//  }
//
//  public void registerIterationVariableValue ( jatva.tracer.ValueKind kind, String type, String name, float value)
//  {
//    if(!transitionPerformed)
//      transitionTrace.addLast(new ScenarioValueMessage(kind, type, name, true, Utils.toObject(value)));
//    else
//      tracer.traceScenarioValue(this, kind, type, name, true, Utils.toObject(value));
//  }
//
//  public void registerIterationVariableValue ( jatva.tracer.ValueKind kind, String type, String name, double value)
//  {
//    if(!transitionPerformed)
//      transitionTrace.addLast(new ScenarioValueMessage(kind, type, name, true, Utils.toObject(value)));
//    else
//      tracer.traceScenarioValue(this, kind, type, name, true, Utils.toObject(value));
//  }

  public void scenarioMethodEnter ( TestingModel object, String method, AbstractAction action )
  {
    //tracer.traceScenarioMethod(this, object, method);
  }

  public void scenarioMethodExit ( )
  {

  }

  public void registerIterationVariable( String type, String name, String value )
  {

  }

  public void registerCollectionIterationVariable( String type, String name, String value )
  {

  }

  public void registerImplicitIterationVariable( String kind, String value )
  {

  }

  /*
  public void storeImplementationArgument (String name, Object a)
  {}

  public void storeImplementationArgumentBoolean (String name, boolean a)
  {}

  public void storeImplementationArgumentByte (String name, byte a)
  {}

  public void storeImplementationArgumentShort (String name, short a)
  {}

  public void storeImplementationArgumentCharacter (String name, char a)
  {}

  public void storeImplementationArgumentInteger (String name, int a)
  {}

  public void storeImplementationArgumentLong (String name, long a)
  {}

  public void storeImplementationArgumentFloat (String name, float a)
  {}

  public void storeImplementationArgumentDouble (String name, double a)
  {}

  public void storeImplementationMArgument (String name, Object[] a)
  {}

  public void storeImplementationMArgumentBoolean (String name, boolean[] a)
  {}

  public void storeImplementationMArgumentByte (String name, byte[] a)
  {}

  public void storeImplementationMArgumentShort (String name, short[] a)
  {}

  public void storeImplementationMArgumentCharacter (String name, char[] a)
  {}

  public void storeImplementationMArgumentInteger (String name, int[] a)
  {}

  public void storeImplementationMArgumentLong (String name, long[] a)
  {}

  public void storeImplementationMArgumentFloat (String name, float[] a)
  {}

  public void storeImplementationMArgumentDouble (String name, double[] a)
  {}
*/
  public void storeImplementationField (String name, Object f)
  {}

  public void storeImplementationFieldBoolean (String name, boolean f)
  {}

  public void storeImplementationFieldByte (String name, byte f)
  {}

  public void storeImplementationFieldShort (String name, short f)
  {}

  public void storeImplementationFieldCharacter (String name, char f)
  {}

  public void storeImplementationFieldInteger (String name, int f)
  {}

  public void storeImplementationFieldLong (String name, long f)
  {}

  public void storeImplementationFieldFloat (String name, float f)
  {}

  public void storeImplementationFieldDouble (String name, double f)
  {}
/*
  public void storeImplementationResult (Object r)
  {}

  public void storeImplementationResultBoolean (boolean r)
  {}

  public void storeImplementationResultByte (byte r)
  {}

  public void storeImplementationResultShort (short r)
  {}

  public void storeImplementationResultCharacter (char r)
  {}

  public void storeImplementationResultInteger (int r)
  {}

  public void storeImplementationResultLong (long r)
  {}

  public void storeImplementationResultFloat (float r)
  {}

  public void storeImplementationResultDouble (double r)
  {}
*/
  public void markImmediateResult ()
  {
    traceMessagesBuffer.addNew( new ModelMessage( this, "MARK IMMEDIATE RESULT CALLED" ) );
    test.markImmediateResult(currentEvent);
  }

  public void testException( TestException error
                           , String info
                           , Throwable cause
                           , boolean isCritical
                           , boolean isInternal
                           , boolean isTraced
                           , String[] tags
                           , String[] values
                           )
  {
    if(isTraced && !isInternal)
    {
      boolean isInterim =    test.getStatus() == Test.CHECKING
                          && !(error instanceof CompositeOracleException)
                          && !(error instanceof NoCorrectSeriesException)
                        ;

      traceMessagesBuffer.addNew( new TestExceptionMessage( this, isInterim, cause, info, tags, values ) );
    }
    else
    if( isTraced )
    {
      traceMessagesBuffer.addNew( new InternalErrorMessage( this, cause, info, tags, values ) );
    }
  }

  public void traceValidException(Throwable ex)
  {
    traceMessagesBuffer.addNew( new ValidExceptionMessage( this, ex ) );
  }

  /*
  private StringBuffer offset()
  {
    StringBuffer r = new StringBuffer();
    for(int i = 0; i < 4*id; i++) r.append(' ');
    return r;
  }

  public void traceMarkedMessage(String s)
  {
    tracer.traceMessage(offset().append(s));
  }*/

  public void traceTestMessage(String s)
  {
    traceMessagesBuffer.addNew( new TestMessage( this, s ) );
  }

  public void traceScenarioMessage(String s)
  {
    traceMessagesBuffer.addNew( new ScenarioMessage( this, s ) );
  }

  public void traceModelMessage(String s)
  {
    traceMessagesBuffer.addNew( new ModelMessage( this, s ) );
  }

  public void traceImplementationMessage(String s)
  {
    traceMessagesBuffer.addNew( new ImplementationMessage( this, s ) );
  }

  //public void

  public void scenarioEnd (String finishReason)
  {
    traceMessagesBuffer.addNew( new ScenarioEndMessage( this, finishReason ) );
  }

  public boolean transitionPerformed ()
  {
    return test.isTransitionPerformed();
  }

    public Object implCtor( String reflectSignature, Object[] args ) throws Throwable
    {
        CtorInfo ctorInfo = OperationInfo.getInvokedCtorInfo( reflectSignature );
        Constructor<?> ctor = ctorInfo.getCtor();

        // if( ctor == null )
        // throw new TranslationError( "Invalid constructor signature"
        // , new String[]{ "class", "constructor signature" }
        // , new String[]{ clazz.getName(), reflect_ctor }
        // );
        //
        // try
        // {
        // ctor.setAccessible( true );
        // }
        // catch( SecurityException e )
        // {
        // throw new InsufficientAccessibilityError(
        // "Insufficient rights to call implementation constructor through reflection"
        // , e
        // , new String[]{"class", "constructor"}
        // , new String[]{clazz.getName(), reflect_ctor}
        // );
        // }

        long stime = -1, etime = -1;
        Object instance = null;

        implementationCallEnter( ctorInfo, null, args );

        stime = System.currentTimeMillis();

        try
        {
            instance = ctor.newInstance( args );

            etime = System.currentTimeMillis();

            try
            {
                registerImplementationResult( ctorInfo.getReturnType(), instance );
            }
            catch( Throwable error )
            {
                throw new ResultTracingException( ctorInfo, instance, error );
            }

            return instance;
        }
        // test system related exceptions
        catch( IllegalAccessException error )
        {
            throw new InsufficientAccessibilityError( "Insufficient rights to call implementation constructor through reflection"
                                                    , error
                                                    , new String[]{ "class", "constructor" }
                                                    , new String[]{ ctorInfo.getDeclaringClass()
                                                                  , ctorInfo.getPrettySignature()
                                                                  }
                                                    );
        }
        catch( IllegalArgumentException error )
        {
            throw new TranslationError( "Incorrect translation of constructor call in mediator"
                                      , error
                                      , new String[]{ "class", "called constructor", "arguments" }
                                      , new String[]{ ctorInfo.getDeclaringClass()
                                                    , ctorInfo.getPrettySignature()
                                                    , Utils.getArraySimpleString( args )
                                                    }
                                      );
        }
        catch( InstantiationException error )
        {
            throw new TranslationError( "Abstract class or interface instantiation"
                                      , error
                                      , new String[]{ "type", "called constructor", "arguments" }
                                      , new String[]{ ctorInfo.getDeclaringClass()
                                                    , ctorInfo.getPrettySignature()
                                                    , Utils.getArraySimpleString( args )
                                                    }
                                      );
        }
        // exception from implementation
        catch( ExceptionInInitializerError error )
        {
            etime = System.currentTimeMillis();

            Throwable ex = error.getCause();
            if( ex == null )
            {
                throw new LostExceptionError( error
                                            , new String[]{ "class", "called constructor", "arguments" }
                                            , new String[]{ ctorInfo.getDeclaringClass()
                                                          , ctorInfo.getPrettySignature()
                                                          , Utils.getArraySimpleString( args )
                                                          }
                                            );
            }
            else if( ex instanceof TestException )
            {
                throw (TestException)ex;
            }
            else
            {
                try
                {
                    registerImplementationException( ex );
                }
                catch( Throwable exception )
                {
                    throw new ResultTracingException( ctorInfo, ex, exception );
                }

                throw ex;
            }
        }
        catch( InvocationTargetException error )
        {
            etime = System.currentTimeMillis();

            Throwable ex = error.getCause();
            if( ex == null )
            {
                throw new LostExceptionError( error
                                            , new String[]{ "class"
                                                          , "called constructor"
                                                          , "arguments"
                                                          }
                                            , new String[]{ ctorInfo.getDeclaringClass()
                                                          , ctorInfo.getPrettySignature()
                                                          , Utils.getArraySimpleString( args )
                                                          }
                                            );
            }
            else if( ex instanceof TestException )
            {
                throw (TestException)ex;
            }
            else
            {
                try
                {
                    registerImplementationException( ex );
                }
                catch( Throwable exception )
                {
                    throw new ResultTracingException( ctorInfo, ex, exception );
                }

                throw ex;
            }
        }
        finally
        {
            implementationCallExit( stime, etime );
        }
    }

    public static void addImplOperations( String ... extendedReflectSignatures )
    {
        OperationInfo.addImplOperations( extendedReflectSignatures );
    }

    public Object implMethod( String reflectSignature, Object obj, Object[] args ) throws Throwable
    {

        MethodInfo methodInfo = OperationInfo.getInvokedMethodInfo( reflectSignature );
        Method method = methodInfo.getMethod();
        //
        // if( method == null )
        // throw new TranslationError( "Invalid method signature"
        // , new String[]{ "class", "method signature" }
        // , new String[]{ clazz.getName(), reflect_mth }
        // );
        //
        // try
        // {
        // method.setAccessible( true );
        // }
        // catch( SecurityException e )
        // {
        // throw new InsufficientAccessibilityError(
        // "Insufficient rights to call implementation method through reflection"
        // , e
        // , new String[]{"class", "method"}
        // , new String[]{clazz.getName(), reflect_mth}
        // );
        // }

        long stime = -1, etime = -1;
        Object result = null;

        boolean isStatic = ( method.getModifiers() & Modifier.STATIC ) != 0;

        implementationCallEnter( methodInfo, obj, args );

        stime = System.currentTimeMillis();

        try
        {
            result = method.invoke( obj, args );

            etime = System.currentTimeMillis();

            try
            {
                registerImplementationResult( methodInfo.getReturnType(), result );
            }
            catch( Throwable error )
            {
                if( isStatic )
                    throw new ResultTracingException( methodInfo, result, error );
                else
                    throw new ResultTracingException( methodInfo, obj, result, error );
            }

            return result;
        }
        // test system related exceptions
        catch( IllegalAccessException error )
        {
            throw new InsufficientAccessibilityError( "Insufficient rights to call implementation method through reflection"
                                                    , error
                                                    , new String[]{ "method" }
                                                    , new String[]{ methodInfo.getPrettySignature() }
                                                    );
        }
        catch( IllegalArgumentException error )
        {
            throw new TranslationError( "Incorrect translation of method call in mediator"
                                      , error
                                      , new String[]{ "class", "object", "called method", "arguments" }
                                      , new String[]{ methodInfo.getDeclaringClass()
                                                    , "" + obj
                                                    , methodInfo.getPrettySignature()
                                                    , Utils.getArraySimpleString( args )
                                                    }
                                      );
        }
        catch( NullPointerException error )
        {
            throw new TranslationError( "Incorrect translation of method call in mediator - call in null instance"
                                      , error
                                      , new String[]{ "class", "object", "called method", "arguments" }
                                      , new String[]{ methodInfo.getDeclaringClass()
                                                    , "" + obj
                                                    , methodInfo.getPrettySignature()
                                                    , Utils.getArraySimpleString( args )
                                                    }
                                      );
        }
        // exceptions from implementation
        catch( ExceptionInInitializerError error )
        {
            etime = System.currentTimeMillis();

            Throwable ex = error.getCause();
            if( ex == null )
            {
                throw new LostExceptionError( error
                                            , new String[]{ "class", "object", "called method", "arguments" }
                                            , new String[]{ methodInfo.getDeclaringClass()
                                                          , "" + obj
                                                          , methodInfo.getPrettySignature()
                                                          , Utils.getArraySimpleString( args )
                                                          }
                                            );
            }
            else if( ex instanceof TestException )
            {
                throw (TestException)ex;
            }
            else
            {
                try
                {
                    registerImplementationException( ex );
                }
                catch( Throwable exception )
                {
                    if( isStatic )
                        throw new ResultTracingException( methodInfo, ex, exception );
                    else
                        throw new ResultTracingException( methodInfo, obj, ex, exception );
                }

                throw ex;
            }
        }
        catch( InvocationTargetException error )
        {
            etime = System.currentTimeMillis();

            Throwable ex = error.getCause();
            if( ex == null )
            {
                throw new LostExceptionError( error
                                            , new String[]{ "class", "object", "called method", "arguments" }
                                            , new String[]{ methodInfo.getDeclaringClass()
                                                          , "" + obj
                                                          , methodInfo.getPrettySignature()
                                                          , Utils.getArraySimpleString( args )
                                                          }
                                            );
            }
            else if( ex instanceof TestException )
            {
                throw (TestException)ex;
            }
            else
            {
                try
                {
                    registerImplementationException( ex );
                }
                catch( Throwable exception )
                {
                    if( isStatic )
                        throw new ResultTracingException( methodInfo, ex, exception );
                    else
                        throw new ResultTracingException( methodInfo, obj, ex, exception );
                }

                throw ex;
            }
        }
        finally
        {
            implementationCallExit( stime, etime );
        }
    }

  // --------------------------------------------------

//  public static class ScenarioValueMessage implements Message
//  {
//    private ValueKind kind;
//    private String    type;
//    private String    name;
//    private boolean   isPrimitive;
//    private Object    value;
//
//    ScenarioValueMessage(ValueKind kind, String type, String name, boolean isPrimitive, Object value)
//    {
//      this.kind = kind;
//      this.type = type;
//      this.name = name;
//      this.isPrimitive = isPrimitive;
//      this.value = value;
//    }
//
//    public ValueKind getKind() { return kind; }
//    public String    getType() { return type; }
//    public String    getName() { return name; }
//    public boolean   isPrimitive() { return isPrimitive; }
//    public Object    getValue() { return value; }
//  }

  public void traceUserInfo( String info )
  {
      traceMessagesBuffer.addNew( new SimpleUserInfoMessage( this, info ) );
  }

  public void traceUserInfo ( String[] tags, String[] values )
  {
      traceMessagesBuffer.addNew( new TaggedUserInfoMessage( this, tags, values ) );
  }
}
