/*
 * Copyright (c) 2000-2003 ISP RAS.
 * B.Communisticheskaya,25, Moscow, Russia.
 * All rights reserved.
 *
 * TraceFormatter
 * $Id: TraceFormatter.java,v 1.22 2008/07/25 10:02:09 almo Exp $
 *
 * Author: Alexei Barantsev <barancev@ispras.ru>
 * Created on December 6, 2002
 */
//=============================================================================

package jatva.tracer;

//=============================================================================

import jatva.lang.ModelEvent;
import jatva.lang.ModelObject;
import jatva.lang.TestingModel;
import jatva.lang.coverage.SEOperationInfo;
import jatva.lang.synchronization.TestManager;
import jatva.util.OperationInfo;

import java.util.Map;
import java.util.Properties;

//=============================================================================

/**
 * <p>A class that implements this interface can be used as formatting strategy
 * for <code>Tracer</code>. Objects of such a class convert messages to text
 * representation according to format rules implemented by the class.</p>
 * <p>This interface consists of a set of methods that correspond to trace
 * messages. When tracer called to trace a message it converts it to string
 * using the corresponding method of the formatting strategy object.</p>
 * <p>In addition to these methods there is two special ones --
 * {@link #getTraceHeader()} and {@link #getTraceFooter()} that provide
 * text representation of trace start and trace end. For example, XML formatter
 * uses these methods to open and close the root tag of the whole trace.</p>
 * <p>Method {@link #configure()} is used to configure strategy object from
 * properties passed by test launcher. To configure strategy object
 * programmatically it is highly recommended to use constructors instead.</p>
 * <p>Tracer pluggable strategies should be clonable because tracer creates
 * defensive unmodifiable copy of all the plugged strategies in the beginning
 * of sending messages and uses this copy till the end of tracing.</p>
 *
 * @see Tracer
 *
 * @author  <a href="mailto:barancev@ispras.ru">Alexei Barantsev</a>
 * @version 1.0
 * @since JavaTESK 1.2
 *
 * @author Victor Kuliamin
 * @version 2.0
 */
public interface TraceFormatter
  extends Cloneable
{
  TraceFormatter clone();

  /**
   * Configures trace formatter from properties passed by test launcher.
   * <strong>WARNING:</strong> This method is designed to be used by
   * <code>TracerConfiguration</code> only. For other purposes use
   * appropriate constructors that create already configured objects.
   *
   * @param formatterProperties Properties passed by test launcher
   * @throws TracerException if the specified properties can't be used to
   *         configure trace formatter.
   */
  void configure(Properties formatterProperties);

  StringBuffer formatTestThreadStart(TestManager manager);

  StringBuffer formatScenarioStart(TestManager manager, TestingModel scenario);

  StringBuffer formatScenarioEnd(TestManager manager, String msg);

  StringBuffer formatEnvironmentProperties(TestManager manager, Map<String,String> props);

  StringBuffer formatException(TestManager manager, boolean isInterim, Throwable cause, String info, String[] tags, String[] values);

  StringBuffer formatValidException(TestManager manager, Throwable exception);

  StringBuffer formatInternalError(TestManager manager, Throwable cause, String info, String[] tags, String[] values);

  StringBuffer formatState(TestManager manager, int stateId, Object state);

  StringBuffer formatTransitionStart(TestManager manager, int transitionId);

  StringBuffer formatTransitionEnd(TestManager manager);

  StringBuffer formatJumpStart(TestManager manager, String jumpKind);

  StringBuffer formatJumpEnd(TestManager manager);

  StringBuffer formatTestFrameStart(TestManager manager);

  StringBuffer formatTestFrameEnter(TestManager manager);

  StringBuffer formatTestFrameExit(TestManager manager);

  StringBuffer formatSerializationStart(TestManager manager);

  StringBuffer formatSerializationEnd(TestManager manager);

  StringBuffer formatTestFrameEnd(TestManager manager);

  StringBuffer formatScenarioMethod(TestManager manager, TestingModel obj, String mth);

  StringBuffer formatScenarioValue(TestManager manager, String kind, String type, String name, boolean isPrimitive, Object value);

  StringBuffer formatScenarioMessage(TestManager manager, String info);

  StringBuffer formatTestMessage(TestManager manager, String info);

  StringBuffer formatModelEvent(TestManager manager, ModelEvent event);

  StringBuffer formatModelEventEnd(TestManager manager, ModelEvent event);

  StringBuffer formatModelValue(TestManager manager, String kind, String type, String name, Object value);

  StringBuffer formatOraclePreconditionEnter(TestManager manager, String signature, int id);

  StringBuffer formatOraclePostconditionEnter(TestManager manager, String signature);

  StringBuffer formatOracleExit(TestManager manager);

  StringBuffer formatPreconditionExit(TestManager manager);

  StringBuffer formatCoverageStructure(TestManager manager, SEOperationInfo disjunctManager);

  StringBuffer formatBranch(TestManager manager, String branch, String msg);

  StringBuffer formatDeferredBranch(TestManager manager, String branch, String msg);

  StringBuffer formatMark(TestManager manager, String mark);

  StringBuffer formatPrimeFormula(TestManager manager, int formulaId, boolean value);

  StringBuffer formatPrimeFormula(TestManager manager, String formula, boolean value);

  StringBuffer formatCoverageElement(TestManager manager, String coverage, int elementId);

  StringBuffer formatInvariantStart(TestManager manager, ModelObject o, String signature, String msg);

  StringBuffer formatInvariantStart(TestManager manager, String p_class, String signature, String msg);

  StringBuffer formatInvariantEnd(TestManager manager);

  StringBuffer formatModelMessage(TestManager manager, String info);

  StringBuffer formatImplementationOperationStart(TestManager manager, OperationInfo operationData, Object obj, Object[] args);

  StringBuffer formatImplementationOperationEnd(TestManager manager, long stime, long etime);

  StringBuffer formatImplementationResult(TestManager m, String resultType, Object result);

  StringBuffer formatImplementationException(TestManager m, Throwable error);

  StringBuffer formatImplementationTimeInterval(TestManager manager, long startTime, long endTime);

  StringBuffer formatImplementationValue(TestManager manager, String kind, String type, String name, Object value);

  StringBuffer formatImplementationMessage(TestManager manager, String info);

  StringBuffer formatUserInfo(TestManager manager, String info);

  StringBuffer formatUserInfo(TestManager manager, String[] tags, String[] values);

  StringBuffer getTraceHeader();

  StringBuffer getTraceFooter();

  void setTestId( int testId );

  void setModelObjectsFormatDepth(int l);

  void setImplementationObjectsFormatDepth(int l);
}
//=============================================================================
// End of file TraceFormatter.java
