/*
 * Copyright (c) 2000-2003 ISP RAS.
 * B.Communisticheskaya,25, Moscow, Russia.
 * All rights reserved.
 *
 * TraceWriter
 * $Id: TraceWriter.java,v 1.4 2008/05/16 12:58:20 almo Exp $
 *
 * Author: Alexei Barantsev <barancev@ispras.ru>
 * Created on December 6, 2002
 */
//=============================================================================

package jatva.tracer;

//=============================================================================

import jatva.lang.synchronization.Test;

import java.util.Properties;

//=============================================================================

/** 
 * <p>A class that implements this interface can be used as trace destination
 * strategy for <code>Tracer</code> - objects of this class write messages
 * in text form to the destination specified by implementation properties.</p>
 * <p>Method {@link #configure()} is used to configure strategy object from
 * properties passed by test launcher. To configure strategy object
 * programmatically it is highly recommended to use constructors instead.</p>
 * <p>Tracer pluggable strategies should be clonable because tracer creates
 * defensive unmodifiable copy of all the plugged strategies in the beginning
 * of sending messages and uses this copy till the end of tracing.</p>
 *
 * @see Tracer
 *
 * @author  <a href="mailto:barancev@ispras.ru">Alexei Barantsev</a>
 * @version 1.0
 * @since JavaTESK 1.2
 */
public interface TraceWriter extends Cloneable 
{

    int DEFAULT_TEST_ID = 0;

    /**
     * Configures trace writer from properties passed by test launcher.<br/>
     * <strong>WARNING:</strong> This method is designed to be used by
     * <code>TracerConfiguration</code> only. For other purposes use 
     * appropriate constructors that create already configured objects.
     * 
     * @param writerProperties Properties passed by test launcher
     * @throws TracerException if the specified properties can't be used to
     *         configure trace writer.
     */
    void configure(Properties writerProperties);

    /** 
     * Sets test id of this tracer. Should have nontrivial implementation for
     * writers that support collecting trace from many tests simultaneously.
     */    
    void setTestId(int testId);

    /**
     * Sets trace header. It is automatically called during tracer
     * preparation (that is hold before sending the first message).<br/>
     * Trace header set by this method should be written to the destination
     * before the first message.
     * 
     * @param header Trace header returned by formatter's method
     *               {@link TraceFormatter#getTraceHeader()}.
     */
    void setTraceHeader(StringBuffer header);

    /**
     * Sets trace footer. It is automatically called during tracer
     * preparation (that is hold before sending the first message).<br/>
     * Trace header set by this method should be written to the destination
     * after the last message.
     * 
     * @param footer Trace footer returned by formatter's method
     *               {@link TraceFormatter#getTraceFooter()}.
     */
    void setTraceFooter(StringBuffer footer);

    void prepare();

    /** 
     * Sends message to the destination.
     * 
     * @param message Trace message to be sent.
     */
    void sendMessage(StringBuffer message);

    /** 
     * Notifies the destination about trace finish. Cliens should call this
     * method to notify writer that there are no more messages to send and
     * writer can release all allocated resources.
     */
    void finish();
    
    TraceWriter clone();
}

//=============================================================================
// End of file TraceWriter.java
