/*
 * Copyright (c) 2000-2003 ISP RAS. B.Communisticheskaya,25, Moscow, Russia. All
 * rights reserved.
 *
 * Tracer $Id: Tracer.java,v 1.32 2008/07/28 13:33:51 almo Exp $
 *
 * Author: Alexei Barantsev <barancev@ispras.ru> Created on November 22, 2002
 */
// =============================================================================
package jatva.tracer;

// =============================================================================

import jatva.lang.ModelEvent;
import jatva.lang.ModelObject;
import jatva.lang.TestingModel;
import jatva.lang.coverage.SEOperationInfo;
import jatva.lang.synchronization.Test;
import jatva.lang.synchronization.TestManager;
import jatva.tracer.impl.AbstractWriter;
import jatva.tracer.impl.FormatAsText;
import jatva.tracer.impl.FormatAsXml;
import jatva.tracer.impl.WriteToConsole;
import jatva.tracer.impl.WriteToDirectory;
import jatva.tracer.impl.WriteToFile;
import jatva.tracer.impl.WriteToMonitor;
import jatva.tracer.impl.WriteToMultiple;
import jatva.tracer.impl.WriteToNull;
import jatva.tracer.impl.WriteToWriter;
import jatva.tracer.messages.TracerMessage;
import jatva.util.OperationInfo;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.Writer;
import java.net.InetAddress;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

// =============================================================================

/**
 * <p>
 * This class implements configurable and plugabble tracer that allows writing
 * trace in different formats to different destinations. Configuration of tracer
 * defines how these messages will be presented (formatting strategy) and where
 * they will be destined (trace destination strategy).
 * </p>
 *
 * <p>
 * Creation of tracer objects is performed automatically either during test
 * system initialization or when the first message send from a Test (this latter
 * case is designed for tests not conforming to JavaTESK technology).
 * </p>
 *
 * <p>
 * To instantiate tracer objects this class uses Prototype pattern. During
 * tracer initialization it creates <code>Tracer</code> object and configure
 * it using <code>TracerConfiguration</code> object passed from test launcher
 * (or default configuration if nothing passed). All other tracer objects are
 * created by cloning this prototype.
 * </p>
 *
 * <p>
 * Starting from version 2.0, tracer object is associated with a Test and most
 * of the information previousely held in a tracer object is now governed by the
 * Test and TestManager.
 * </p>
 *
 * @see TraceFormatter
 * @see AbstractWriter
 * @see TracerConfiguration
 *
 * @author <a href="mailto:barancev@ispras.ru">Alexei Barantsev </a>
 * @version 2.0
 * @since JavaTESK 1.2
 */
public final class Tracer implements TraceFormatter, Cloneable
{
    // Environment properties names
    public static final String ENV_HOST = "Host";

    public static final String ENV_OS = "Operating System";

    public static final String ENV_JVM = "Java VM";

    public static final String ENV_PRODUCT_NAME = "Product Name";

    public static final String ENV_PRODUCT_VERSION = "Product Version";

    public static final String ENV_PRODUCT_BUILD = "Product Build";

    private static final String P_PRODUCT_VERSION = "Specification-Version";

    private static final String P_PRODUCT_BUILD = "Implementation-Version";

    /* Version information for environment */
    public static final String PRODUCT_NAME = "JavaTESK";

    public static final String PRODUCT_VERSION;

    public static final String PRODUCT_BUILD;

    static
    {
        Properties versionProperties = new Properties();
        versionProperties.setProperty( P_PRODUCT_VERSION, "SNAPSHOT" );
        versionProperties.setProperty( P_PRODUCT_BUILD, new Date().toString() );

        InputStream versionResource = Tracer.class.getClassLoader().getResourceAsStream( "META-INF/MANIFEST.MF" );

        if( versionResource != null )
        {
            try
            {
                versionProperties.load( versionResource );
            } catch( IOException e )
            {}
        }

        PRODUCT_VERSION = versionProperties.getProperty( P_PRODUCT_VERSION );
        PRODUCT_BUILD = versionProperties.getProperty( P_PRODUCT_BUILD );
    }

    // -------------------------------------------------------------------------
    // Static part (default tracer + tracer factory)
    // -------------------------------------------------------------------------

    /**
     * Holds prototype of tracer object. All other tracer objects are created by
     * cloning this object.
     */
    private static Tracer tracerPrototype;

    /**
     * Provides access to the tracer prototype object.
     *
     * @return The tracer prototype object.
     */
    public static Tracer getPrototype()
    {
        if( tracerPrototype == null )
        {
            synchronized( Tracer.class )
            {
                if( tracerPrototype == null )
                {
                    tracerPrototype = new Tracer();
                    TracerConfiguration.instantiate().configure( tracerPrototype );
                }
            }
        }

        return tracerPrototype;
    }

    /**
     * Internal factory method for tracer objects.
     */
    public static Tracer createTracerObject()
    {
        synchronized( Tracer.class )
        {
            return getPrototype().clone();
        }
    }

    // -------------------------------------------------------------------------
    // Tracer object
    // -------------------------------------------------------------------------

    /**
     * Only factory can create objects of this class.
     */
    protected Tracer()
    {}

    /**
     * Holds object that implements formatting strategy.
     */
    private TraceFormatter formatter = new FormatAsXml();

    /**
     * Holds object that implements trace destination strategy.
     */
    private TraceWriter writer = new WriteToConsole(); // new WriteToNull();

    /**
     * Holds scenario (test) unique identifier.
     *
     * @see Test#getId()
     */
    private int testId;

    /**
     * Holds information about current state of tracer object -- is it in use or
     * not.
     */
    private boolean tracingStarted = false;

    /** Holds information about environment of test system */
    static private Map<String,String> environment = new HashMap<String,String>();

    /** Holds number of exceptions traced */
    private int numberOfExceptions = 0;

    /*
     * Init environment properties - to be executed only once, when prototype is
     * created.
     */
    static
    {
        // Name and version of tool are set by tool itself
        setEnvironmentProperty( ENV_HOST, getHostInfo() );
        setEnvironmentProperty( ENV_OS, getOSInfo() );
        setEnvironmentProperty( ENV_JVM, getJVMInfo() );
        setEnvironmentProperty( ENV_PRODUCT_NAME, PRODUCT_NAME );
        setEnvironmentProperty( ENV_PRODUCT_VERSION, PRODUCT_VERSION );
        setEnvironmentProperty( ENV_PRODUCT_BUILD, PRODUCT_BUILD );

        // System.out.println(environment);
    }

    /**
     * Creates and returns a copy of this object.
     */
    @Override
    public synchronized Tracer clone()
    {
        Tracer newObj;
        try
        {
            newObj = (Tracer)super.clone();
        }
        catch( CloneNotSupportedException e )
        {
            // Never should happen
            throw new TracerInternalError(   this.getClass().getName()
                                           + ".clone() "
                                           + "Could not clone tracer object"
                                         , e
                                         );
        }

        newObj.setFormatter( formatter.clone() );
        newObj.setTraceTo( writer.clone() );

        newObj.filter = (Tracer.Filter)filter.clone();
        // TODO: ???
        // newObj.environment = (Map)environment.clone();

        return newObj;
    }

    // -------------------------------------------------------------------------
    // Configuration
    // -------------------------------------------------------------------------

    /**
     * Getter for property formatter.
     *
     * @return Value of property formatter.
     */
    private TraceFormatter getFormatter()
    {
        return formatter;
    }

    /**
     * Setter for property formatter.
     *
     * @param formatter
     *            New value of property formatter.
     */
    public synchronized void setFormatter( TraceFormatter formatter )
    {
        if( tracingStarted )
        {
            System.err.println( "WARNING: tracer configuration cannot be changed - tracing already started" );
        } else
        {
            this.formatter = formatter != null ? formatter : new FormatAsXml();
        }
    }

    /**
     * Configure tracer to use text format.
     */
    public void setTextFormat()
    {
        setFormatter( new FormatAsText() );
    }

    /**
     * Configure tracer to use XML format.
     */
    public void setXmlFormat()
    {
        setFormatter( new FormatAsXml() );
    }

    /**
     * Sets the depth of model objects tracing in trace messages 0 - full depth
     * 1 - only type information and unique object code 2 - 1 + fields, elements
     * of collections and elements of collection fields, full arrays
     */
    public void setModelObjectsFormatDepth( int level )
    {
        this.formatter.setModelObjectsFormatDepth( level );
    }

    /**
     * Sets the depth of model objects tracing in trace messages 0 - full depth
     * 1 - only type information and unique object code 2 - 1 + fields, elements
     * of collections and elements of collection fields, full arrays
     */
    public void setImplementationObjectsFormatDepth( int level )
    {
        this.formatter.setImplementationObjectsFormatDepth( level );
    }

    public void finishWriter()
    {
        writer.finish();
    }

    public void setTestId( int testId )
    {
        this.testId = testId;
        this.writer.setTestId( testId );
        this.formatter.setTestId( testId );
    }

    /**
     * Setter for property writer.
     *
     * @param writer
     *            New value of property writer.
     */
    private synchronized void setWriter( TraceWriter writer )
    {
        if( tracingStarted )
        {
            System.err.println( "WARNING: tracer configuration cannot be changed - tracing already started" );
        } else
        {
            this.writer = writer != null ? writer : new WriteToNull();
            this.writer.setTraceHeader( formatter.getTraceHeader() );
            this.writer.setTraceFooter( formatter.getTraceFooter() );
            this.writer.setTestId( this.testId );
        }
    }

    /**
     * Configure tracer to send trace to the specified writer plugin.
     *
     * @param traceWriter
     *            Writer plugin to send trace to.
     */
    public synchronized void setTraceTo( TraceWriter traceWriter )
    {
        setWriter( traceWriter );
    }

    /**
     * Configure tracer to send trace to the specified writer plugins.
     *
     * @param traceWriters
     *            Writer plugins to send trace to.
     */
    public synchronized void setTraceTo( TraceWriter[] traceWriters )
    {
        if( traceWriters != null )
        {
            if( traceWriters.length > 1 )
            {
                WriteToMultiple multiWriter = new WriteToMultiple();
                for( int i = 0; i < traceWriters.length; i++ )
                {
                    multiWriter.addTraceWriter( traceWriters[ i ] );
                }
                setTraceTo( multiWriter );

            } else if( traceWriters.length == 1 )
            {
                setTraceTo( traceWriters[ 0 ] );
            } else
            {
                setTraceTo( new WriteToNull() );
            }
        } else
        {
            setTraceTo( new WriteToNull() );
        }
    }

    /**
     * Configure tracer to send trace to console.
     */
    public void setTraceToConsole()
    {
        setTraceTo( new WriteToConsole() );
    }

    /**
     * Configure tracer to send trace to the specified file.
     *
     * @param file
     *            File to send trace to.
     */
    public void setTraceToFile( File file )
    {
        setTraceTo( new WriteToFile( file ) );
    }

    /**
     * Configure tracer to send trace to a file in the specified directory.
     *
     * @param directory
     *            Directory to send trace to.
     */
    public void setTraceToDirectory( File directory )
    {
        setTraceTo( new WriteToDirectory( directory ) );
    }

    /**
     * Configure tracer to send trace to a network monitor.
     *
     * @param address
     *            Address of the network monitor.
     * @param port
     *            Port the network monitor listens to.
     */
    public void setTraceToMonitor( InetAddress address, int port )
    {
        setTraceTo( new WriteToMonitor( address, port ) );
    }

    /**
     * Configure tracer to send trace to an arbitrary writer.
     *
     * @param writer
     *            Writer to send trace to.
     */
    public void setTraceToWriter( Writer writer )
    {
        setTraceTo( new WriteToWriter( writer ) );
    }

    /**
     * Configure tracer to not send trace anywhere.
     */
    public void setTraceToNull()
    {
        setTraceTo( new WriteToNull() );
    }

    /**
     * Get the number of exceptions traced so far
     */
    public int getNumberOfTracedExceptions()
    {
        return numberOfExceptions;
    }

    private synchronized void prepare( TestManager m )
    {
        if( tracingStarted ) { return; }

        tracingStarted = true;

        writer.prepare();

        writer.sendMessage( formatEnvironmentProperties( m, environment ) );
    }

    // -------------------------------------------------------------------------
    // Environment
    // -------------------------------------------------------------------------

    /**
     * Sets environment property to trace
     */
    static public void setEnvironmentProperty( String propName, String propValue )
    {
        environment.put( propName, propValue );
    }

    /**
     * Gets stored environment property
     */
    public static String getEnvironmentProperty( String propName )
    {
        return environment.get( propName );
    }

    /**
     * Removes environment proerty key
     */
    public static void removeEnvironmentProperty( String propName )
    {
        environment.remove( propName );
    }

    // Default values for environment properties

    public static String getOSInfo()
    {
        return   System.getProperty( "os.name" )
               + " "
               + System.getProperty( "os.arch" )
               + " "
               + System.getProperty( "os.version" )
             ;
    }

    public static String getJVMInfo()
    {
        return   System.getProperty( "java.vm.vendor" )
               + " "
               + System.getProperty( "java.vm.name" )
               + " "
               + System.getProperty( "java.vm.version" )
             ;
    }

    public static String getHostInfo()
    {
        try
        {
            return InetAddress.getLocalHost().toString();
        }
        catch( Throwable e )
        {
            return TraceConstants.UNKNOWN_HOST;
        }
    }

    // -------------------------------------------------------------------------
    // Messages
    // -------------------------------------------------------------------------

    public StringBuffer formatTestThreadStart( TestManager m )
    {
        try
        {
            return formatter.formatTestThreadStart( m );
        }
        catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatThreadStart()", err );
        }
    }

    /**
     * Sends message about scenario start. <br/>WARNING: This method is designed
     * to be used by test system. Users should not use this method. For
     * non-scenario threads use {@link #startTrace()}.
     *
     * @param scenario
     *            Started scenario.
     */
    public StringBuffer formatScenarioStart( TestManager m, TestingModel scenario )
    {
        try
        {
            return formatter.formatScenarioStart( m, scenario );
        }
        catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatScenarioStart()", err );
        }
    }

    /**
     * Sends message about scenario environment. <br/>WARNING: This method is
     * designed to be used by test system. Users should not use this method. For
     * non-scenario threads use {@link #traceEnvironmentProperties()}.
     *
     * @param props
     *            Properties of environment.
     */
    public StringBuffer formatEnvironmentProperties( TestManager m, Map<String,String> props )
    {
        try
        {
            return formatter.formatEnvironmentProperties( m, props );
        } catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatEnvironmentProperties()", err );
        }
    }

    /**
     * Sends message about scenario end. <br/>WARNING: This method is designed
     * to be used by test system. Users should not use this method. For
     * non-scenario threads use {@link #endTrace()}.
     */
    public StringBuffer formatScenarioEnd( TestManager m, String msg )
    {
        try
        {
            return formatter.formatScenarioEnd( m, msg );
        }
        catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatScenarioEnd()", err );
        }
    }

    /**
     * Sends message about exception thrown by test system on account of catched
     * exception. <br/>WARNING: This method is designed to be used by test
     * system. Users should not use this method.
     *
     * @param where
     *            Identification of the place where the problem occurs.
     * @param rootCause
     *            Exception that causes this problem.
     * @param info
     *            Additional information about the problem.
     */
    public StringBuffer formatException( TestManager m, boolean isInterim, Throwable error, String info, String[] tags, String[] vals )
    {
        numberOfExceptions++;

        try
        {
            return formatter.formatException( m, isInterim, error, info, tags, vals );
        }
        catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatException()", err );
        }
    }

    public StringBuffer formatValidException( TestManager m, Throwable ex )
    {
        numberOfExceptions++;

        try
        {
            return formatter.formatValidException( m, ex );
        }
        catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatValidException()", err );
        }
    }

    /**
     * Sends message about internal error thrown by test system on account of
     * catched exception. <br/>WARNING: This method is designed to be used by
     * test system. Users should not use this method.
     *
     * @param where
     *            Identification of the place where the problem occurs.
     * @param rootCause
     *            Exception that causes this problem.
     * @param info
     *            Additional information about the problem.
     */
    public StringBuffer formatInternalError( TestManager m, Throwable error, String info, String[] tags, String[] vals )
    {
        numberOfExceptions++;

        try
        {
            return formatter.formatInternalError( m, error, info, tags, vals );
        }
        catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatInternalError()", err );
        }
    }

    /**
     * Send message about reached state. <br/>WARNING: This method is designed
     * to be used by test system. Users should not use this method.
     *
     * @param stateId
     *            State identifier.
     */
    public StringBuffer formatState( TestManager m, int stateId, Object state )
    {
        try
        {
            if( isTraceScenarioGroup() )
            {
                return formatter.formatState( m, stateId, state );
            } else
                return null;
        } catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatState()", err );
        }
    }

    /**
     * Send message about transition start. <br/>WARNING: This method is
     * designed to be used by test system. Users should not use this method.
     *
     * @param transitionId
     *            Transition identifier.
     */
    public StringBuffer formatTransitionStart( TestManager m, int transitionId )
    {
        try
        {
            if( isTraceScenarioGroup() )
            {
                return formatter.formatTransitionStart( m, transitionId );
            } else {
                return null;
            }
        } catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatTransitionStart()", err );
        }
    }

    /**
     * Send message about transition end. <br/>WARNING: This method is designed
     * to be used by test system. Users should not use this method.
     */
    public StringBuffer formatTransitionEnd( TestManager m )
    {
        try
        {
            if( isTraceScenarioGroup() )
            {
                return formatter.formatTransitionEnd( m );
            } else {
                return null;
            }
        } catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatTransitionEnd()", err );
        }
    }

    /**
     * Send message about special transition start. <br/>WARNING: This method is
     * designed to be used by test system. Users should not use this method.
     *
     * @param jumpKind
     *            Kind of special transition.
     */
    public StringBuffer formatJumpStart( TestManager m, String jumpKind )
    {
        try
        {
            if( isTraceScenarioGroup() )
            {
                return formatter.formatJumpStart( m, jumpKind );
            } else {
                return null;
            }
        } catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatJumpStart()", err );
        }
    }

    /**
     * Send message about special transition end. <br/>WARNING: This method is
     * designed to be used by test system. Users should not use this method.
     */
    public StringBuffer formatJumpEnd( TestManager m )
    {
        try
        {
            if( isTraceScenarioGroup() )
            {
                return formatter.formatJumpEnd( m );
            } else {
                return null;
            }
        } catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatJumpEnd()", err );
        }
    }

    public StringBuffer formatTestFrameStart( TestManager m )
    {
        try
        {
            if( isTraceScenarioGroup() )
            {
                return formatter.formatTestFrameStart( m );
            } else {
                return null;
            }
        } catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatTestFrameStart()", err );
        }
    }

    public StringBuffer formatTestFrameEnter( TestManager m )
    {
        try
        {
            if( isTraceScenarioGroup() )
            {
                return formatter.formatTestFrameEnter( m );
            } else {
                return null;
            }
        } catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatTestFrameEnter()", err );
        }
    }

    public StringBuffer formatTestFrameExit( TestManager m )
    {
        try
        {
            if( isTraceScenarioGroup() )
            {
                return formatter.formatTestFrameExit( m );
            } else {
                return null;
            }
        } catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatTestFrameExit()", err );
        }
    }

    public StringBuffer formatSerializationStart( TestManager m )
    {
        try
        {
            if( isTraceScenarioGroup() )
            {
                return formatter.formatSerializationStart( m );
            } else {
                return null;
            }
        } catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatSerializationStart()", err );
        }
    }

    public StringBuffer formatSerializationEnd( TestManager m )
    {
        try
        {
            if( isTraceScenarioGroup() )
            {
                return formatter.formatSerializationEnd( m );
            } else {
                return null;
            }
        }
        catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatSerializationEnd()", err );
        }
    }

    public StringBuffer formatTestFrameEnd( TestManager m )
    {
        try
        {
            if( isTraceScenarioGroup() )
            {
                return formatter.formatTestFrameEnd( m );
            } else {
                return null;
            }
        }
        catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatTestFrameEnd()", err );
        }
    }

    public StringBuffer formatScenarioMethod( TestManager m, TestingModel obj, String mth )
    {
        try
        {
            if( isTraceScenarioGroup() )
            {
                return formatter.formatScenarioMethod( m, obj, mth );
            } else {
                return null;
            }
        } catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatScenarioMethod()", err );
        }
    }

    /**
     * Sends message that contains information about some scenario-related
     * value. <br/>WARNING: This method is designed to be used by test system.
     * Users should not use this method.
     *
     * @param kind
     *            Kind of the value.
     * @param type
     *            Type of the value.
     * @param name
     *            Name of the value.
     * @param value
     *            Value of the value :)
     */
    public StringBuffer formatScenarioValue( TestManager m
                                           , String kind
                                           , String type
                                           , String name
                                           , boolean isPrimitive
                                           , Object value
                                           )
    {
        try
        {
            if( isTraceScenarioGroup() )
            {
                return formatter.formatScenarioValue( m
                                                    , kind
                                                    , type
                                                    , name
                                                    , isPrimitive
                                                    , value
                                                    );
            } else {
                return null;
            }
        }
        catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatScenarioValue()", err );
        }
    }

    /**
     * Sends message that contains arbitrary scenario-related information. <br/>
     * WARNING: This method is designed to be used by test system. Users should
     * not use this method.
     *
     * @param info
     *            A piece of information.
     */
    public StringBuffer formatScenarioMessage( TestManager m, String info )
    {
        try
        {
            if( isTraceScenarioGroup() )
            {
                return formatter.formatScenarioMessage( m, info );
            } else {
                return null;
            }
        }
        catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatScenarioMessage()", err );
        }
    }

    /**
     * @param manager
     * @param s
     */
    public StringBuffer formatTestMessage( TestManager m, String s )
    {
        try
        {
            if( isTraceScenarioGroup() )
            {
                return formatter.formatTestMessage( m, s );
            } else {
                return null;
            }
        }
        catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatTestMessage()", err );
        }
    }

    public StringBuffer formatModelEvent( TestManager m, ModelEvent e ) // coverage
    {
        try
        {
            if( isTraceModelGroup() )
            {
                return formatter.formatModelEvent( m, e );
            } else {
                return null;
            }
        }
        catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatModelEvent()", err );
        }
    }

    /**
     * Sends message about model operation end. <br/>WARNING: This method is
     * designed to be used by test system. Users should not use this method.
     */
    public StringBuffer formatModelEventEnd( TestManager m, ModelEvent event )
    {
        try
        {
            if( isTraceModelGroup() )
            {
                return formatter.formatModelEventEnd( m, event );
            } else {
                return null;
            }
        } catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatModelEventEnd()", err );
        }
    }

    /**
     * Sends message that contains information about some model-related value.
     * <br/>WARNING: This method is designed to be used by test system. Users
     * should not use this method.
     *
     * @param kind
     *            Kind of the value.
     * @param type
     *            Type of the value.
     * @param name
     *            Name of the value.
     * @param value
     *            Value of the value :)
     */
     public StringBuffer formatModelValue( TestManager m
                                         , String kind
                                         , String type
                                         , String name
                                         , Object value
                                         )
     {
        try
        {
            if( isTraceModelValues() )
            {
                return formatter.formatModelValue( m, kind, type, name, value );
            } else {
                return null;
            }
        } catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in format_model_value()", err );
        }
    }

    /**
     * Sends message about oracle start. <br/>WARNING: This method is designed
     * to be used by test system. Users should not use this method.
     *
     * @param signature
     *            Specification signature.
     * @param disjunctManager
     *            Disjunct manager that contain coverage structure description.
     */
    public StringBuffer formatOraclePreconditionEnter( TestManager m, String signature, int eventId )
    {
        try
        {
            if( isTraceModelGroup() )
            {
                return formatter.formatOraclePreconditionEnter( m, signature, eventId );
            } else {
                return null;
            }
        }
        catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatOraclePreconditionEnter()", err );
        }
        /*
         * if (disjunctManager != null && isTraceCoverageInfo() &&
         * !oracles.contains(signature)) { oracles.add(signature); {
         * getWriter().sendMessage( getFormatter().formatCoverageStructure(wrid,
         * signature, disjunctManager)); } }
         */
    }

    public StringBuffer formatOraclePostconditionEnter( TestManager m, String signature )
    {
        try
        {
            if( isTraceModelGroup() )
            {
                return formatter.formatOraclePostconditionEnter( m, signature );
            } else {
                return null;
            }
        }
        catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatOraclePostconditionEnter()", err );
        }
    }

    public StringBuffer formatCoverageStructure( TestManager m, SEOperationInfo disjunctManager )
    {
        try
        {
            if( isTraceModelGroup() )
            {
                return formatter.formatCoverageStructure( m, disjunctManager );
            } else {
                return null;
            }
        }
        catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatCoverageStructure()", err );
        }
    }

    /**
     * Sends message about oracle end. <br/>WARNING: This method is designed to
     * be used by test system. Users should not use this method.
     */
    public StringBuffer formatOracleExit( TestManager m )
    {
        try
        {
            if( isTraceModelGroup() )
            {
                return formatter.formatOracleExit( m );
            } else {
                return null;
            }
        }
        catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatOracleExit()", err );
        }
    }

    /**
     * Sends message about precondition end. <br/>WARNING: This method is
     * designed to be used by test system. Users should not use this method.
     */
    public StringBuffer formatPreconditionExit( TestManager m )
    {
        try
        {
            if( isTraceModelGroup() )
            {
                return formatter.formatPreconditionExit( m );
            } else {
                return null;
            }
        }
        catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatPreconditionExit()", err );
        }
    }

    /**
     * Sends message about covered branch. <br/>WARNING: This method is designed
     * to be used by test system. Users should not use this method.
     *
     * @param branch
     *            Branch name.
     */
    public StringBuffer formatBranch( TestManager m, String branch, String msg )
    {
        try
        {
            if( isTraceCoverageInfo() )
            {
                return formatter.formatBranch( m, branch, msg );
            } else {
                return null;
            }
        } catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatBranch()", err );
        }
    }

    public StringBuffer formatDeferredBranch( TestManager m, String branch, String msg )
    {
        try
        {
            if( isTraceCoverageInfo() )
            {
                return formatter.formatDeferredBranch( m, branch, msg );
            } else {
                return null;
            }
        } catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatDeferredBranch()", err );
        }
    }

    /**
     * Sends message about covered mark. <br/>WARNING: This method is designed
     * to be used by test system. Users should not use this method.
     *
     * @param mark
     *            Mark name.
     */
    public StringBuffer formatMark( TestManager m, String mark )
    {
        try
        {
            if( isTraceCoverageInfo() )
            {
                return formatter.formatMark( m, mark );
            } else {
                return null;
            }
        } catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatMark()", err );
        }
    }

    /**
     * Sends message about covered prime formula. <br/>WARNING: This method is
     * designed to be used by test system. Users should not use this method.
     *
     * @param value
     *            Value of the prime formula.
     * @param formulaId
     *            Prime formula identifier.
     */
    public StringBuffer formatPrimeFormula( TestManager m, int formulaId, boolean value )
    {
        try
        {
            if( isTraceCoverageInfo() )
            {
                return formatter.formatPrimeFormula( m, formulaId, value );
            } else {
                return null;
            }
        } catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatFormula()", err );
        }
    }

    public StringBuffer formatPrimeFormula( TestManager m, String formula, boolean value )
    {
        try
        {
            if( isTraceCoverageInfo() )
            {
                return formatter.formatPrimeFormula( m, formula, value );
            } else {
                return null;
            }
        }
        catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatFormula()", err );
        }
    }

    /**
     * Sends message about user defined coverage element. <br/>WARNING: This
     * method is designed to be used by test system. Users should not use this
     * method.
     *
     * @param coverage
     *            Coverage identifier.
     * @param elementId
     *            Coverage element identifier.
     */
    public StringBuffer formatCoverageElement( TestManager m, String coverage, int elementId )
    {
        try
        {
            if( isTraceCoverageInfo() )
            {
                return formatter.formatCoverageElement( m, coverage, elementId );
            } else {
                return null;
            }
        } catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatCoverageElement()", err );
        }
    }

    /**
     * Sends message about invariant start. <br/>WARNING: This method is
     * designed to be used by test system. Users should not use this method.
     *
     * @param signature
     *            Invariant signature.
     */
    public StringBuffer formatInvariantStart( TestManager m, ModelObject o, String signature, String msg )
    {
        try
        {
            if( isTraceModelGroup() )
            {
                return formatter.formatInvariantStart( m, o, signature, msg );
            } else {
                return null;
            }
        } catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatInvariantEnter()", err );
        }
    }

    /**
     * Sends message about invariant start. <br/>WARNING: This method is
     * designed to be used by test system. Users should not use this method.
     *
     * @param signature
     *            Invariant signature.
     */
    public StringBuffer formatInvariantStart( TestManager m
                                            , String p_class
                                            , String signature
                                            , String msg
                                            )
    {
        try
        {
            if( isTraceModelGroup() )
            {
                return formatter.formatInvariantStart( m, p_class, signature, msg );
            } else {
                return null;
            }
        } catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatInvariantEnter()", err );
        }
    }

    /**
     * Sends message about invariant end. <br/>WARNING: This method is designed
     * to be used by test system. Users should not use this method.
     */
    public StringBuffer formatInvariantEnd( TestManager m )
    {
        try
        {
            if( isTraceModelGroup() )
            {
                return formatter.formatInvariantEnd( m );
            } else {
                return null;
            }
        } catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatInvariantExit()", err );
        }
    }

    /**
     * Sends message that contains arbitrary model-related information. <br/>
     * WARNING: This method is designed to be used by test system. Users should
     * not use this method.
     *
     * @param info
     *            A piece of information.
     */
    public StringBuffer formatModelMessage( TestManager m, String info )
    {
        try
        {
            if( isTraceModelGroup() )
            {
                return formatter.formatModelMessage( m, info );
            } else {
                return null;
            }
        }
        catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatModelMessage()", err );
        }
    }

    /**
     * Sends message about implementation operation start. <br/>WARNING: This
     * method is designed to be used by test system. Users should not use this
     * method.
     *
     * @param signature
     *            Operation signature.
     */
    public StringBuffer formatImplementationOperationStart( TestManager m
                                                          , OperationInfo operationData
                                                          , Object obj
                                                          , Object[] args
                                                          )
    {
        try
        {
            if( isTraceImplementationGroup() )
            {
                return formatter.formatImplementationOperationStart( m, operationData, obj, args );
            } else {
                return null;
            }
        } catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatImplementationOperationStart()", err );
        }
    }

    /**
     * Sends message about implementation operation end. <br/>WARNING: This
     * method is designed to be used by test system. Users should not use this
     * method.
     */
    public StringBuffer formatImplementationOperationEnd( TestManager m, long stime, long etime )
    {
        try
        {
            if( isTraceImplementationGroup() )
            {
                return formatter.formatImplementationOperationEnd( m, stime, etime );
            } else {
                return null;
            }
        }
        catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatImplementationOperationEnd()", err );
        }
    }

    public StringBuffer formatImplementationResult( TestManager m
                                                  , String resultType
                                                  , Object result
                                                  )
    {
        try
        {
            if( isTraceImplementationGroup() )
            {
                return formatter.formatImplementationResult( m, resultType, result );
            } else {
                return null;
            }
        }
        catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatImplementationResult()", err );
        }
    }

    public StringBuffer formatImplementationException( TestManager m, Throwable error )
    {
        try
        {
            if( isTraceImplementationGroup() )
            {
                return formatter.formatImplementationException( m, error );
            } else {
                return null;
            }
        }
        catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatImplementationException()", err );
        }
    }

    /**
     * Sends message that contains arbitrary implementation-related information.
     * <br/>WARNING: This method is designed to be used by test system. Users
     * should not use this method.
     *
     * @param startTime
     *            Implementation operation start time.
     * @param endTime
     *            Implementation operation end time.
     */
    public StringBuffer formatImplementationTimeInterval( TestManager m, long startTime, long endTime )
    {
        try
        {
            if( isTraceImplementationGroup() )
            {
                return formatter.formatImplementationTimeInterval( m, startTime, endTime );
            } else {
                return null;
            }
        }
        catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatImplementationTimeInterval()", err );
        }
    }

    /**
     * Sends message that contains information about some implementation-related
     * value. <br/>WARNING: This method is designed to be used by test system.
     * Users should not use this method.
     *
     * @param kind
     *            Kind of the value.
     * @param type
     *            Type of the value.
     * @param name
     *            Name of the value.
     * @param value
     *            Value of the value :)
     */
    public StringBuffer formatImplementationValue( TestManager m, String kind, String type, String name, Object value )
    {
        try
        {
            if( isTraceImplementationValues() )
            {
                return formatter.formatImplementationValue( m, kind, type, name, value );
            } else {
                return null;
            }
        }
        catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatImplementationValue()", err );
        }
    }

    /**
     * Sends message that contains arbitrary implementation-related information.
     * <br/>WARNING: This method is designed to be used by test system. Users
     * should not use this method.
     *
     * @param info
     *            A piece of information.
     */
    public StringBuffer formatImplementationMessage( TestManager m, String info )
    {
        try
        {
            if( isTraceImplementationGroup() )
            {
                return formatter.formatImplementationMessage( m, info );
            } else {
                return null;
            }
        }
        catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatImplementationMessage()", err );
        }
    }

    /**
     * Sends message that contains arbitrary user supplied information. WARNING:
     * This method should never be used by test system.
     *
     * @param info
     *            A piece of information.
     */
    public StringBuffer formatUserInfo( TestManager m, String info )
    {
        try
        {
            if( isTraceUserInfo() )
            {
                return formatter.formatUserInfo( m, info );
            } else {
                return null;
            }
        }
        catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatUserInfo()", err );
        }
    }

    /**
     * Sends message that contains arbitrary user supplied information. WARNING:
     * This method should never be used by test system.
     *
     * @param info
     *            A piece of information.
     */
//    public static StringBuffer formatUserInfo( String info )
//    {
//        TestManager m = TestManager.getRegistrar();
//        m.getTracer().traceUserInfo( m, info );
//    }

    /**
     * Sends message that contains arbitrary user supplied information. WARNING:
     * This method should never be used by test system.
     *
     * @param info
     *            A piece of information.
     */
    public StringBuffer formatUserInfo( TestManager m, String[] tags, String[] values )
    {
        try
        {
            if( isTraceUserInfo() )
            {
                return formatter.formatUserInfo( m, tags, values );
            } else {
                return null;
            }
        } catch( Throwable err )
        {
            throw new TracerInternalError( "Exception in formatUserInfo()", err );
        }
    }

    /**
     * Sends message that contains arbitrary user supplied information. WARNING:
     * This method should never be used by test system.
     *
     * @param info
     *            A piece of information.
     */
    // public static StringBuffer formatUserInfo (String[] tags, String[]
    // values)
    // {
    // TestManager m = TestManager.getRegistrar();
    // m.getTracer().traceUserInfo(m, tags, values);
    // }
    /**
     * Holds object that defines trace filtering.
     */
    private Tracer.Filter filter = new Tracer.Filter();

    /**
     * Getter for property filter.
     *
     * @return Value of filter property.
     */
    private Tracer.Filter getFilter()
    {
        return filter;
    }

    /**
     * Allows or denies tracing messages of scenario group.
     *
     * @param setTrace
     *            Whether such messages should be traced or not?
     *
     * @see <a href="package-summary.html#Managing_trace_detail_level">Managing
     *      trace detail level </a>
     */
    public void setTraceScenarioGroup( boolean setTrace )
    {
        getFilter().setTraceScenarioGroup( setTrace );
    }

    /**
     * Checks whether tracing messages of scenario group is allowed or not.
     *
     * @return <code>true</code> if tracing such messages is allowed,
     *         otherwise <code>else</code>
     *
     * @see <a href="package-summary.html#Managing_trace_detail_level">Managing
     *      trace detail level </a>
     */
    private boolean isTraceScenarioGroup()
    {
        return getFilter().isTraceScenarioGroup();
    }

    /**
     * Allows or denies tracing messages of model group.
     *
     * @param setTrace
     *            Whether such messages should be traced or not?
     *
     * @see <a href="package-summary.html#Managing_trace_detail_level">Managing
     *      trace detail level </a>
     */
    public void setTraceModelGroup( boolean setTrace )
    {
        setTraceModelGroup( setTrace, false );
    }

    /**
     * Allows or denies tracing messages of model group and maybe its subgroups.
     *
     * @param setTrace
     *            Whether such messages should be traced or not?
     * @param recursive
     *            Whether the same should be done for subgroups or not?
     *
     * @see <a href="package-summary.html#Managing_trace_detail_level">Managing
     *      trace detail level </a>
     */
    public void setTraceModelGroup( boolean setTrace, boolean recursive )
    {
        getFilter().setTraceModelGroup( setTrace, recursive );
    }

    /**
     * Checks whether tracing messages of model group is allowed or not.
     *
     * @return <code>true</code> if tracing such messages is allowed,
     *         otherwise <code>else</code>
     *
     * @see <a href="package-summary.html#Managing_trace_detail_level">Managing
     *      trace detail level </a>
     */
    private boolean isTraceModelGroup()
    {
        return getFilter().isTraceModelGroup();
    }

    /**
     * Allows or denies tracing messages that contain information about model
     * values.
     *
     * @param setTrace
     *            Whether such messages should be traced or not?
     *
     * @see <a href="package-summary.html#Managing_trace_detail_level">Managing
     *      trace detail level </a>
     */
    public void setTraceModelValues( boolean setTrace )
    {
        getFilter().setTraceModelValues( setTrace );
    }

    /**
     * Checks whether tracing messages that contain information about model
     * values is allowed or not.
     *
     * @return <code>true</code> if tracing such messages is allowed,
     *         otherwise <code>else</code>
     *
     * @see <a href="package-summary.html#Managing_trace_detail_level">Managing
     *      trace detail level </a>
     */
    private boolean isTraceModelValues()
    {
        return getFilter().isTraceModelValues();
    }

    /**
     * Allows or denies tracing messages that contains information about test
     * coverage.
     *
     * @param setTrace
     *            Whether such messages should be traced or not?
     *
     * @see <a href="package-summary.html#Managing_trace_detail_level">Managing
     *      trace detail level </a>
     */
    public void setTraceCoverageInfo( boolean setTrace )
    {
        getFilter().setTraceCoverageInfo( setTrace );
    }

    /**
     * Checks whether tracing messages that contain information about test
     * coverage is allowed or not.
     *
     * @return <code>true</code> if tracing such messages is allowed,
     *         otherwise <code>else</code>
     *
     * @see <a href="package-summary.html#Managing_trace_detail_level">Managing
     *      trace detail level </a>
     */
    private boolean isTraceCoverageInfo()
    {
        return getFilter().isTraceCoverageInfo();
    }

    /**
     * Allows or denies tracing messages of implementation group.
     *
     * @param setTrace
     *            Whether such messages should be traced or not?
     *
     * @see <a href="package-summary.html#Managing_trace_detail_level">Managing
     *      trace detail level </a>
     */
    public void setTraceImplementationGroup( boolean setTrace )
    {
        setTraceImplementationGroup( setTrace, false );
    }

    /**
     * Allows or denies tracing messages of implementation group and maybe its
     * subgroups.
     *
     * @param setTrace
     *            Whether such messages should be traced or not?
     * @param recursive
     *            Whether the same should be done for subgroups or not?
     *
     * @see <a href="package-summary.html#Managing_trace_detail_level">Managing
     *      trace detail level </a>
     */
    public void setTraceImplementationGroup( boolean setTrace, boolean recursive )
    {
        getFilter().setTraceImplementationGroup( setTrace, recursive );
    }

    /**
     * Checks whether tracing messages of implementation group is allowed or
     * not.
     *
     * @return <code>true</code> if tracing such messages is allowed,
     *         otherwise <code>else</code>
     *
     * @see <a href="package-summary.html#Managing_trace_detail_level">Managing
     *      trace detail level </a>
     */
    private boolean isTraceImplementationGroup()
    {
        return getFilter().isTraceImplementationGroup();
    }

    /**
     * Allows or denies tracing messages that contain information about
     * implementation values.
     *
     * @param setTrace
     *            Whether such messages should be traced or not?
     *
     * @see <a href="package-summary.html#Managing_trace_detail_level">Managing
     *      trace detail level </a>
     */
    public void setTraceImplementationValues( boolean setTrace )
    {
        getFilter().setTraceImplementationValues( setTrace );
    }

    /**
     * Checks whether tracing messages that contain information about
     * implementation values is allowed or not.
     *
     * @return <code>true</code> if tracing such messages is allowed,
     *         otherwise <code>else</code>
     *
     * @see <a href="package-summary.html#Managing_trace_detail_level">Managing
     *      trace detail level </a>
     */
    private boolean isTraceImplementationValues()
    {
        return getFilter().isTraceImplementationValues();
    }

    /**
     * Allows or denies tracing user-supplied messages.
     *
     * @param setTrace
     *            Whether such messages should be traced or not?
     *
     * @see <a href="package-summary.html#Managing_trace_detail_level">Managing
     *      trace detail level </a>
     */
    public void setTraceUserInfo( boolean setTrace )
    {
        getFilter().setTraceUserInfo( setTrace );
    }

    /**
     * Checks whether tracing user-supplied messages.
     *
     * @return <code>true</code> if tracing such messages is allowed, otherwise
     *         <code>else</code>
     *
     * @see <a href="package-summary.html#Managing_trace_detail_level">Managing
     *      trace detail level </a>
     */
    private boolean isTraceUserInfo()
    {
        return getFilter().isTraceUserInfo();
    }

    // Pseugo-groups

    /**
     * Allows or denies tracing messages of all groups.
     *
     * @param setTrace Whether all messages should be traced or not?
     *
     * @see <a href="package-summary.html#Managing_trace_detail_level">Managing
     *      trace detail level </a>
     */
    public void setTraceAll( boolean setTrace )
    {
        setTraceScenarioGroup( setTrace );
        setTraceModelGroup( setTrace, true );
        setTraceImplementationGroup( setTrace, true );
        setTraceUserInfo( setTrace );
    }

    /**
     * Allows or denies tracing messages that contain information about values.
     *
     * @param setTrace Whether such messages should be traced or not?
     *
     * @see <a href="package-summary.html#Managing_trace_detail_level">Managing
     *      trace detail level </a>
     */
    public void setTraceValues( boolean setTrace )
    {
        setTraceModelValues( setTrace );
        setTraceImplementationValues( setTrace );
    }

    /**
     * Temporarily denies tracing user-supplied messages. <br/>WARNING: This
     * method is designed to be used by test system to prevent tracing
     * user-supplied messages twice.
     */
    public void set_user_info_off()
    {
        getFilter().setTemporarilyStopTraceUserInfo( true );
    }

    /**
     * Allows temporarily denied tracing user-supplied messages. <br/>WARNING:
     * This method is designed to be used by test system to prevent tracing
     * user-supplied messages twice.
     */
    public void set_user_info_on()
    {
        getFilter().setTemporarilyStopTraceUserInfo( false );
    }

    public void traceMessage( TracerMessage tm )
    {
        if( tm.getStringForm() != null )
        {
            try
            {
                prepare( tm.getRegistrar() );
                writer.sendMessage( tm.getStringForm() );
            } catch( Throwable err )
            {
                throw new TracerInternalError( "Exception in formatUserInfo()", err );
            }
        }
    }

    private class Filter implements Cloneable
    {

        @Override
        public Object clone()
        {
            try
            {
                return super.clone();

            } catch( CloneNotSupportedException e )
            {
                // Never should happen
                throw new TracerInternalError( "Class " + this.getClass().getName() + " cannot clone tracer filter object", e );
            }
        }

        private boolean traceScenarioGroup = true;

        private boolean traceModelGroup = true;

        private boolean traceModelValues = true;

        private boolean traceCoverageInfo = true;

        private boolean traceImplementationGroup = true;

        private boolean traceImplementationValues = true;

        private boolean traceUserInfo = true;

        private boolean temporarilyStopTraceUserInfo = false;

        public Filter()
        {}

        public void setTraceScenarioGroup( boolean setTrace )
        {
            traceScenarioGroup = setTrace;
        }

        public boolean isTraceScenarioGroup()
        {
            return traceScenarioGroup;
        }

        public void setTraceModelGroup( boolean setTrace, boolean recursive )
        {
            traceModelGroup = setTrace;
            if( recursive )
            {
                setTraceModelValues( setTrace );
                setTraceCoverageInfo( setTrace );
            }
        }

        public boolean isTraceModelGroup()
        {
            return traceModelGroup;
        }

        public void setTraceModelValues( boolean setTrace )
        {
            traceModelValues = setTrace;
        }

        public boolean isTraceModelValues()
        {
            return traceModelValues;
        }

        public void setTraceCoverageInfo( boolean setTrace )
        {
            traceCoverageInfo = setTrace;
        }

        public boolean isTraceCoverageInfo()
        {
            return traceCoverageInfo;
        }

        public void setTraceImplementationGroup( boolean setTrace, boolean recursive )
        {
            traceImplementationGroup = setTrace;
            if( recursive )
            {
                setTraceImplementationValues( setTrace );
            }
        }

        public boolean isTraceImplementationGroup()
        {
            return traceImplementationGroup;
        }

        public void setTraceImplementationValues( boolean setTrace )
        {
            traceImplementationValues = setTrace;
        }

        public boolean isTraceImplementationValues()
        {
            return traceImplementationValues;
        }

        public void setTraceUserInfo( boolean setTrace )
        {
            traceUserInfo = setTrace;
        }

        public boolean isTraceUserInfo()
        {
            return traceUserInfo && !temporarilyStopTraceUserInfo;
        }

        public void setTemporarilyStopTraceUserInfo( boolean setTrace )
        {
            temporarilyStopTraceUserInfo = setTrace;
        }
    }

    public void configure( Properties formatterProperties )
    {
        formatter.configure( formatterProperties );
    }

    public StringBuffer getTraceFooter()
    {
        return formatter.getTraceFooter();
    }

    public StringBuffer getTraceHeader()
    {
        return formatter.getTraceHeader();
    }

    /**
     * Sends message that contains arbitrary user supplied information. WARNING:
     * This method should never be used by test system.
     *
     * @param info A piece of information.
     */
    @Deprecated
    public void traceUserInfo (TestManager m, String info)
    {
      TestManager.getRegistrar().traceUserInfo( info );
    }

    /**
     * Sends message that contains arbitrary user supplied information. WARNING:
     * This method should never be used by test system.
     *
     * THIS METHOD IS DEPRECATED!!!
     * USE TestManager.getRegistrar.traceUserInfo instead.
     *
     * @param info A piece of information.
     */
    @Deprecated
    public static void traceUserInfo (String info)
    {
      TestManager.getRegistrar().traceUserInfo(info);
    }
}
//=============================================================================
// End of file Tracer.java
