/*
 * Copyright (c) 2000-2003 ISP RAS.
 * B.Communisticheskaya,25, Moscow, Russia.
 * All rights reserved.
 *
 * TracerConfiguration
 * $Id: TracerConfiguration.java,v 1.4 2008/05/16 13:01:21 almo Exp $
 *
 * Author: Alexei Barantsev <barancev@ispras.ru>
 * Created on January 9, 2003
 */
//=============================================================================

package jatva.tracer;

//=============================================================================

import jatva.tracer.impl.AbstractWriter;
import jatva.tracer.impl.WriteToMultiple;
import jatva.tracer.impl.WriteToNull;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.Properties;

//=============================================================================

/**
 * <p>This class is designed to pass configuration in serialized form from test
 * launcher to test. Inasmuch as tracer can use different formatters and
 * writers this class allowed to hold arbitrary non-specialized parameters.
 * Passed configuration is used to tune tracer objects: formatter properties
 * are used to instantiate and configure formatting strategy, writer properties
 * are used to instantiate and configure tarce destination strategies. Tracer
 * can write trace to several destination and so configuration contains an array
 * of writer properties.</p>
 * 
 * <p>WARNING: This class should be used only by test launcher to create tracer
 * configuration and pass it to test.</p>
 *
 * @author  <a href="mailto:barancev@ispras.ru">Alexei Barantsev</a>
 * @version 1.0
 * @since JavaTESK 1.1
 */
public class TracerConfiguration implements Serializable {

    static final long serialVersionUID = 1515323657448323507L;

    /**
     * Name of the system property that holds path to a file that contains
     * serialized TracerConfiguration object.
     */
    public static final transient String CONFIGURATION_FILE_PROPERTY = "jatva.tracer.configuration";

    /**
     * Name of the property that holds strategy class name.
     */
    public static final transient String CLASS_NAME_PROPERTY = "class";

    /**
     * Holds formatter properties.
     */
    private Properties formatterProperties;

    /**
     * Holds an array of writer properties.
     */
    private Properties[] writerProperties;

    /**
     * Value of the trace level that allows all messages.
     */
    public static final transient String TRACE_ALL = "all";

    /**
     * Value of the trace level that allows model messages and disables
     * implementation messages.
     */
    public static final transient String TRACE_MODEL = "model";

    /**
     * Value of the trace level that allows only scenario structure messages
     * (states and transitions).
     */
    public static final transient String TRACE_SCENARIO = "scenario";

    /**
     * Value of the trace level that disables all messages except for failures.
     */
    public static final transient String TRACE_FAILURES_ONLY = "failures";

    /**
     * Holds property that specifies trace level.
     */
    private String traceLevel = TRACE_ALL;

    /**
     * Holds property that allows tracing user messages.
     */
    private boolean traceUserInfo = true;

    /**
     * Holds the single instance of tracer configuration available to test system.
     * Test launchers can create as many configurations as needed (to run different
     * test with appropriate configurations). But each test uses only one configuration
     * passed by the launcher at start time. Tests can't change configuration, all changes
     * should be done in traced objects (global changes - in prototype tracer object).
     */
    private static transient TracerConfiguration instance;

    /**
     * Creates a new instance of TracerConfiguration.
     */
    public TracerConfiguration() {
    }

    /**
     * Restores serialized configuration object from the file. Used during
     * tracer initialization to obtain configuration created by test launcher
     * and stored in serialized form to a temporarty file. Launcher passes path
     * to these file through system property with name defined by
     * <code>CONFIGURATION_FILE_PROPERTY</code> field.
     *
     * @return Restored configuration or <code>null</code> if nothing to restore.
     * @throws TracerInternalError if any error occured during deserialization.
     */
    static synchronized TracerConfiguration instantiate() {
        if (instance == null) {
            String configurationFilePath = System.getProperty(CONFIGURATION_FILE_PROPERTY);

            if (configurationFilePath != null) {
                File configurationFile = new File(configurationFilePath);
                try {
                    ObjectInputStream configurationStream =
                        new ObjectInputStream(new FileInputStream(configurationFile));

                    instance = (TracerConfiguration) configurationStream.readObject();

                } catch (Throwable e) {
                    throw new TracerInternalError(
                          TracerConfiguration.class.getName()
                        + ".instantiate() "
                        + "cannot restore serialized configuration from file "
                        + configurationFile,
                        e);
                }

            }

            if (instance == null) {
                instance = new TracerConfiguration();
            }
        }

        return instance;
    }

    /**
     * Serializes configuration object to the file with name specified by
     * <code>CONFIGURATION_FILE_PROPERTY</code> field.
     *
     * @throws TracerInternalError if any error occured during serialization.
     */
    public void serialize() {
        String configurationFilePath = System.getProperty(CONFIGURATION_FILE_PROPERTY);

        if (configurationFilePath != null) {
            File configurationFile = new File(configurationFilePath);

            try {
                ObjectOutputStream configurationStream =
                    new ObjectOutputStream(new FileOutputStream(configurationFile));
                configurationStream.writeObject( this );
                configurationStream.close();

            } catch (Throwable e) {
                throw new TracerInternalError(
                      TracerConfiguration.class.getName()
                    + ".serialize() "
                    + "cannot serialize configuration to file "
                    + configurationFile,
                    e);
            }
        }
    }

    /**
     * Gets a set of properties to be used for instantiation and configuration
     * of formatting strategy. Returned value must contain special property with
     * the key defined by the value of <code>CLASS_NAME_PROPERTY</code> field
     * that holds formatter class name.
     *
     * @return Formatter properties.
     */
    public Properties getFormatterProperties() {
        return formatterProperties;
    }

    /**
     * Fills configuration with a set of formatter properties.
     *
     * @param formatterProperties Formatter properties.
     * @throws TracerInternalError if property that specifies formatter class 
     * name is not set.
     */
    public void setFormatterProperties(Properties formatterProperties) {
        if (formatterProperties != null) {
            if (!formatterProperties.containsKey(CLASS_NAME_PROPERTY)) {
                throw new TracerInternalError(
                      TracerConfiguration.class.getName()
                    + ".setFormatterProperties(Properties) "
                    + "formatter properties must contain formatter class name"
                    );
            }
        }

        this.formatterProperties = formatterProperties;
    }

    /**
     * Gets an array where each element contains a set of properties to be used
     * for instantiation and configuration of a single trace destination
     * strategy. Each element must contain special property with the key
     * defined by te value of <code>CLASS_NAME_PROPERTY</code> field that holds
     * writer class name.
     *
     * @return Array of writer properties.
     */
    public Properties[] getWriterProperties() {
        return writerProperties;
    }

    /**
     * Fills configuration with an array of writer properties.
     *
     * @param writerProperties Array of writer properties.
     * @throws TracerInternalError if property that specifies writer class name
     * is not set.
     */
    public void setWriterProperties(Properties[] writerProperties) {
        if (writerProperties != null) {
            for (int i = 0; i < writerProperties.length; i++) {
                if (!writerProperties[i].containsKey(CLASS_NAME_PROPERTY)) {
                    throw new TracerInternalError(
                          TracerConfiguration.class.getName()
                        + ".setWriterProperties(Properties[]) "
                        + " writer properties must contain writer class name"
                        );
                }
            }
        }

        this.writerProperties = writerProperties;
    }

    /**
     * Gets value of trace level.
     * 
     * @return Trace level.
     */
    public String getTraceLevel() {
        return traceLevel;
    }

    /**
     * Sets desired trace level.
     * 
     * @param traceLevel Trace level.
     */
    public void setTraceLevel(String traceLevel) {
        if (TRACE_ALL.equals(traceLevel)
            || TRACE_MODEL.equals(traceLevel)
            || TRACE_SCENARIO.equals(traceLevel)
            || TRACE_FAILURES_ONLY.equals(traceLevel))
        {
            this.traceLevel = traceLevel;

        } else {
            throw new TracerException(
                TracerConfiguration.class.getName() + ".setTraceLevel(String) "
                + "unrecognized trace level specification: " + traceLevel);
        }
    }

    /**
     * Checks whether tracing user messages is enabled or not?
     * 
     * @return <code>true</code> if tracing user messages is enables,
     *         <code>false</code> otherwise.
     */
    public boolean isTraceUserInfo() {
        return traceUserInfo;
    }

    /**
     * Enable or disable tracing user messages.
     * 
     * @param traceUserInfo Specifies whether tracing user messages is enabled or not.
     */
    public void setTraceUserInfo(boolean traceUserInfo) {
        this.traceUserInfo = traceUserInfo;
    }

    //-------------------------------------------------------------------------
    //    Factories
    //-------------------------------------------------------------------------

    /**
     * Formatter factory
     */
    private TraceFormatter createAndConfigureFormatter() {
        return createAndConfigureFormatter(getFormatterProperties());
    }

    /** 
     * Formatter factory.
     */
    private static TraceFormatter createAndConfigureFormatter(Properties formatterProperties) {
        TraceFormatter newFormatter =
            (TraceFormatter) instantiateStrategyObject(formatterProperties, "formatter");
        newFormatter.configure(formatterProperties);
        return newFormatter;
    }

    /**
     * Writer factory.
     */
    private TraceWriter createAndConfigureWriter() {
        return createAndConfigureWriter(getWriterProperties());
    }

    /** 
     * Writer factory.
     */
    private static TraceWriter createAndConfigureWriter(Properties[] writerProperties) {
        if (writerProperties != null && writerProperties.length > 0) {
            if (writerProperties.length > 1) {
                WriteToMultiple multiWriter = new WriteToMultiple();

                for (int i = 0; i < writerProperties.length; i++) {
                    multiWriter.addTraceWriter(
                        createAndConfigureWriter(writerProperties[i]));
                }

                return multiWriter;
        
            } else {
                return createAndConfigureWriter(writerProperties[0]);
            }

        } else {
            return new WriteToNull();
        }
    }

    /** 
     * Writer factory.
     */
    private static TraceWriter createAndConfigureWriter(Properties writerProperties) {
        TraceWriter newWriter =
            (TraceWriter) instantiateStrategyObject(writerProperties, "writer");
        newWriter.configure(writerProperties);
        return newWriter;
    }

    /** 
     * Universal factory.
     */
    private static Object instantiateStrategyObject(Properties properties, String kind) {
        String className = properties.getProperty(
            TracerConfiguration.CLASS_NAME_PROPERTY);

        if (className == null) {
            throw new TracerException(
                TracerConfiguration.class.getName() + ".instantiateStrategyObject() "
                + " class of " + kind + " is not specified");
        }

        try {
            Class clazz =
                TracerConfiguration.class.getClassLoader().loadClass(className);
            return clazz.newInstance();

        } catch (Throwable e) {
            throw new TracerException(
                TracerConfiguration.class.getName() + ".instantiateStrategyObject() "
                + "cannot instantiate " + kind, e);
        }
    }

    /**
     * Configures tracer object as specified in the given configuration. If
     * configuration does not contain information about formatter or an error
     * occured during instantiating and configuring new formatter - current
     * formatter will be kept. If configuration does not contain information
     * about writers or an error occured during instantiating and configuring
     * new writers - current writers will be kept.
     *
     * @param tracer Tracer object to configure.
     */
    synchronized void configure(Tracer tracer) {
        if (getFormatterProperties() != null) {
            tracer.setFormatter(createAndConfigureFormatter());
        }

        if (getWriterProperties()!= null) {
            tracer.setTraceTo(createAndConfigureWriter());
        }

        tracer.setTraceAll(true);
        if (!TRACE_ALL.equals(getTraceLevel())) {
            if (TRACE_MODEL.equals(getTraceLevel())) {
                tracer.setTraceImplementationGroup(false, true);

            } else if (TRACE_SCENARIO.equals(getTraceLevel())) {
                tracer.setTraceImplementationGroup(false, true);
                tracer.setTraceModelGroup(false, true);

            } else if (TRACE_FAILURES_ONLY.equals(getTraceLevel())) {
                tracer.setTraceImplementationGroup(false, true);
                tracer.setTraceModelGroup(false, true);
                tracer.setTraceScenarioGroup(false);

            } else {
                throw new TracerInternalError(
                    TracerConfiguration.class.getName() + ".configure(Tracer) "
                    + "unrecognized trace level specification: " + getTraceLevel());
            }
        }

        tracer.setTraceUserInfo(isTraceUserInfo());
    }

}
//=============================================================================
// End of file TracerConfiguration.java
