/*
 * Copyright (c) 2000 ISP RAS. B.Communisticheskaya,25, Moscow, Russia. All
 * rights reserved.
 * 
 * AbstractWriter $Id: AbstractWriter.java,v 1.3 2006/03/30 12:47:32 barancev
 * Exp $
 * 
 * Author: Alexei Barantsev <barancev@ispras.ru> Created on March 29, 2003
 */
// =============================================================================
package jatva.tracer.impl;

// =============================================================================

import jatva.tracer.TraceFormatter;
import jatva.tracer.TraceWriter;
import jatva.tracer.TracerException;
import jatva.tracer.TracerInternalError;

import java.util.Properties;

// =============================================================================

/**
 * <p>
 * This class provides default implementation of all auxiliary methods of
 * <code>TraceWriter</code> interface. This class is designed for extention.
 * It is recommended to inherit concrete classes that implement
 * <code>TraceWriter</code> interface from this class.
 * </p>
 * 
 * <p>
 * Methods {@link AbstractWriter#sendMessage()} has no default implementation, each
 * subclass must implement it.
 * </p>
 * 
 * <p>
 * This class defines two states of a trace writer object -- "configuration" and
 * "tracing-in-progress" -- indicated by the value of
 * <code>tracingInProgress</code> boolean property. These states correspond to
 * different sets of methods available.
 * </p>
 * 
 * <p>
 * Initially trace writer is in "configuration" state where any trace writer
 * property can be set or changed. When method <code>sendMessage()</code> is
 * called trace writer should go to "tracing-in-progress" state where no
 * property changes allowed because that can confuse destination or corrupt
 * trace. If trace writer implemented to be reusable it is possible to move it
 * back to initial state when <code>finish()</code> method is called.
 * </p>
 * 
 * @author <a href="mailto:barancev@ispras.ru">Alexei Barantsev</a>
 * @version 1.0
 */
public abstract class AbstractWriter implements TraceWriter
{
    protected StringBuffer header;

    protected StringBuffer footer;

    protected boolean tracingInProgress = false;

    protected int testId = DEFAULT_TEST_ID;

    public AbstractWriter clone()
    {
        try
        {
            AbstractWriter newObj = (AbstractWriter)super.clone();
            newObj.footer = footer;
            newObj.header = header;
            newObj.tracingInProgress = tracingInProgress;
            newObj.testId = testId;
            return newObj;
        } catch( CloneNotSupportedException e )
        {
            // Never should happen
            throw new TracerInternalError( "Class " + this.getClass().getName() + " cannot clone writer", e );
        }
    }

    /**
     * Configures trace writer from properties passed by test launcher.<br/>
     * <strong>WARNING:</strong> This method is designed to be used by
     * <code>TracerConfiguration</code> only. For other purposes use
     * appropriate constructors that create already configured objects.
     * 
     * @param writerProperties
     *            Properties passed by test launcher
     * @throws TracerException
     *             if the specified properties can't be used to configure trace
     *             writer.
     */
    public void configure( Properties writerProperties )
    {}

    /**
     * Utility method that can be used by inherited classes to set/unset a
     * property indicating that writer is in "tracing-in-progress" state.
     * 
     * @param tracingInProgress
     *            Value of the "tracing-in-progress" state.
     */
    protected synchronized final void setTracingInProgress( boolean tracingInProgress )
    {
        this.tracingInProgress = tracingInProgress;
    }

    /**
     * Utility method that can be used to check current state of trace writer.
     * 
     * @return Value of the "tracing-in-progress" state.
     */
    protected synchronized final boolean isTracingInProgress()
    {
        return tracingInProgress;
    }

    /**
     * Checks whether changes in configuration are allowed or not.
     * 
     * @return <code>true</code> if configuration changes are allowed (<code>!isTracingInProgress()</code>).
     */
    protected final boolean isConfigurationChangeAllowed()
    {
        return !isTracingInProgress();
    }

    public void setTestId( int testId )
    {
        this.testId = testId;
    }

    /**
     * Provides access to test id property.
     * 
     * @return Test id.
     */
    protected final int getTestId()
    {
        return testId;
    }

    /**
     * Sets trace header. It is automatically called during tracer preparation
     * (that is hold before sending the first message).<br/> Trace header set
     * by this method should be written to the destination before the first
     * message.
     * 
     * @param header
     *            Trace header returned by formatter's method
     *            {@link TraceFormatter#getTraceHeader()}.
     */
    public void setTraceHeader( StringBuffer header )
    {
        this.header = header;
    }

    /**
     * Provides access to trace header.
     * 
     * @return Trace header.
     */
    protected final StringBuffer getTraceHeader()
    {
        return header;
    }

    /**
     * Sets trace footer. It is automatically called during tracer preparation
     * (that is hold before sending the first message).<br/> Trace header set
     * by this method should be written to the destination after the last
     * message.
     * 
     * @param footer
     *            Trace footer returned by formatter's method
     *            {@link TraceFormatter#getTraceFooter()}.
     */
    public void setTraceFooter( StringBuffer footer )
    {
        this.footer = footer;
    }

    /**
     * Provides access to trace footer.
     * 
     * @return Trace footer.
     */
    protected final StringBuffer getTraceFooter()
    {
        return footer;
    }

    /**
     * Implements typical behaviour of <code>sendMessage</code> method in
     * initial state by printing trace header to the destination.
     */
    public synchronized void prepare()
    {
        assert !isTracingInProgress();
        setTracingInProgress( true );
        sendMessage( getTraceHeader() );
    }

    /**
     * Sends message to the destination.
     * 
     * @param message
     *            Trace message to be sent.
     */
    public abstract void sendMessage( StringBuffer message );

    /**
     * Notifies the destination about trace finish. Cliens should call this
     * method to notify writer that there are no more messages to send and
     * writer can release all allocated resources.
     * 
     * Default implementation sends trace footer to destination and changes
     * trace writer's state to initial.
     * 
     * @see AbstractWriter#finish()
     */
    public void finish()
    {
        if( isTracingInProgress() )
        {
            sendMessage( getTraceFooter() );
            setTracingInProgress( false );
        }
    }
}

//=============================================================================
// End of file AbstractWriter.java
