/*
 * Copyright (c) 2000 ISP RAS. B.Communisticheskaya,25, Moscow, Russia. All
 * rights reserved.
 * 
 * WriteToFile $Id: WriteToFile.java,v 1.5 2008/06/20 12:34:16 almo Exp $
 * 
 * Author: Alexei Barantsev <barancev@ispras.ru> Created on December 9, 2002
 */
// =============================================================================
package jatva.tracer.impl;

// =============================================================================

import jatva.tracer.TracerException;
import jatva.tracer.TracerInternalError;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

// =============================================================================

/**
 * This class implements writing trace messages to file. TODO: Write more
 * detailed documentation for this class.
 * 
 * @author <a href="mailto:barancev@ispras.ru">Alexei Barantsev</a>
 * @version 1.0
 */
public class WriteToFile extends AbstractWriter
{

    /**
     * Name of configuration property that specifies file to write trace to.
     */
    public static String FILE_PROPERTY = "file";

    /**
     * Holds mapping from File to Writer. Prevents use of several writers to the
     * same file.
     */
    private static Map<File,Writer> writers = new HashMap<File,Writer>();

    /**
     * Holds mapping from File to Integer counter. Used to count who writes to
     * each file.
     */
    private static Map<File,Integer> counters = new HashMap<File,Integer>();

    private File file;

    private Writer writer;

    /**
     * Creates a new non-configured instance of <code>WriteToFile</code>.<br/>
     * <strong>WARNING:</strong> This constructor creates non-configured
     * instance that can't be used immediately. It is designed to be used by
     * <code>TracerConfiguration</code> only. For other purposes use more
     * specific constructors.
     */
    public WriteToFile()
    {}

    /**
     * Creates a new instance of <code>WriteToFile</code> that writes trace to
     * the specified file.
     * 
     * @param file
     *            File to write trace to.
     */
    public WriteToFile( File file )
    {
        try
        {
            this.file = file.getCanonicalFile();

        } catch( IOException e )
        {
            this.file = file;
        }
    }

    public synchronized WriteToFile clone()
    {
        WriteToFile newObj = (WriteToFile)super.clone();
        newObj.writer = null;
        return newObj;
    }

    public void configure( Properties writerProperties )
    {
        String fileProp = writerProperties.getProperty( FILE_PROPERTY );

        if( fileProp == null )
        {
            throw new TracerException( "File not specified" );

        } else
        {
            setFile( new File( fileProp ) );
        }
    }

    private File getFile()
    {
        return file;
    }

    private synchronized void setFile( File file )
    {
        if( isConfigurationChangeAllowed() )
        {
            try
            {
                this.file = file.getCanonicalFile();

            } catch( IOException e )
            {
                this.file = file;
            }

        } else
        {
            System.err.println( "WARNING: tracer configuration cannot be changed - tracing already started" );
        }
    }

    private void registerFile()
    {
        synchronized( WriteToFile.class )
        {
            Integer counter = (Integer)counters.get( getFile() );
            int c = counter == null ? 0 : counter.intValue();
            counters.put( getFile(), new Integer( c + 1 ) );
        }
    }

    private void unregisterFile()
    {
        synchronized( WriteToFile.class )
        {
            int c = ( (Integer)counters.get( getFile() ) ).intValue();
            counters.put( getFile(), new Integer( c - 1 ) );

            if( c - 1 == 0 && writers.containsKey( getFile() ) )
            {
                writers.put( getFile(), null );

                try
                {
                    writer.flush();
                    writer.close();

                } catch( IOException e )
                {
                    throw new TracerInternalError( "Could not close writer to file " + getFile(), e );
                }
            }
        }
    }

    private Writer createWriter( boolean append )
    {
        try
        {
            Writer writer = new OutputStreamWriter( new FileOutputStream( getFile(), append ), "UTF-8" );
            writer.flush();
            return writer;
        } catch( IOException e )
        {
            throw new TracerException( "Could not open trace file " + getFile(), e );
        }
    }

    public synchronized void prepare()
    {
        if( writer == null )
        {
            synchronized( WriteToFile.class )
            {
                registerFile();

                File parent = getFile().getParentFile();
                if( !parent.exists() )
                {
                    parent.mkdirs();
                }

                writer = writers.get( getFile() );

                if( writer == null )
                {
                    writer = createWriter( false );
                    writers.put( getFile(), writer );
                }
            }
        }
        super.prepare();
    }

    public void sendMessage( StringBuffer message )
    {
        try
        {
            synchronized( writer )
            {
                writer.write( message.toString() );
                writer.flush();
            }
        } catch( IOException e )
        {
            throw new TracerException( "Could not write to file " + getFile(), e );
        }
    }

    public void finish()
    {
        super.finish();

        if( writer != null )
        {
            unregisterFile();
            writer = null;
        }
    }

    /** 
     * Last resort -- if a client forgot to call <code>finish()</code> call it.
     */
    public void finalize()
    {
        finish();
    }

}
//=============================================================================
// End of file WriteToFile.java
