package jatva.util;

import java.util.Hashtable;

public abstract class OperationInfo
{
    /**
     * Associates partial signatures of implementation operations invoked during the test with
     * corresponding 'OperationObject'.
     * Partial signature is formed as follows:
     * * methods:
     *     return_type
     *     " "
     *     declaring_type
     *     "."
     *     method_name
     *     "("
     *     par_type
     *     (
     *        ","
     *        par_type
     *     )*
     *     ")"
     *
     * * constructors:
     *    declaring_type
     *     "("
     *     par_type
     *     (
     *        ","
     *        par_type
     *     )*
     *     ")"
     *
     * ATTENTION! full names of all types are supposed to be in reflection format
     * of Class.getSimpleName(), i.e.
     * nested types delimiter should be '$' rather then '.'.
     */
    private static Hashtable<String, OperationInfo>
            invokedImplOperations = new Hashtable<String, OperationInfo>();

    /**
     *
     * @param extendedReflectSignatures have the following format:
     *        <partial signature (see above )>
     *        (
     *          ":"
     *          arg_name
     *          (
     *            ","
     *            arg_name
     *          )*
     *        )? ( if we know arguments names )
     */
    public static synchronized void addImplOperations( String[] extendedReflectSignatures )
    {
        for( String ext_ref_sig : extendedReflectSignatures )
        {
            String ref_sig = cutExtendedReflectSignature( ext_ref_sig );

            if( invokedImplOperations.containsKey( ref_sig ) )
                continue;
            else {
                int opening_brace = ref_sig.indexOf( '(' );
                int space_after_return_type = ref_sig.lastIndexOf( ' ', opening_brace );
                OperationInfo op_info = null;

                if( space_after_return_type == -1 )
                    op_info = new CtorInfo( ext_ref_sig );
                else
                    op_info = new MethodInfo( ext_ref_sig );
                invokedImplOperations.put( ref_sig, op_info );
            }
        }
    }

    private static String cutExtendedReflectSignature( String extendedReflectSignature )
    {
        int colon_index = extendedReflectSignature.lastIndexOf( ':' );
        if( colon_index != -1 )
        {
            return extendedReflectSignature.substring( 0, colon_index );
        } else {
            return extendedReflectSignature;
        }
    }

    public static CtorInfo getInvokedCtorInfo( String reflectSignature )
    {
        return (CtorInfo)invokedImplOperations.get( reflectSignature );
    }

    public static MethodInfo getInvokedMethodInfo( String reflectSignature )
    {
        return (MethodInfo)invokedImplOperations.get( reflectSignature );
    }

    protected static final String[] emptyPars = new String[ 0 ];

    protected String   reflectSignature;
    protected String   prettySignature;
    protected String   declaringClass;
    protected String   returnType;
    protected String[] parTypes;
    protected String[] parNames;

    public OperationInfo( String extendedReflectSignature )
    {
        int colon_index = extendedReflectSignature.lastIndexOf( ':' );
        this.parNames = emptyPars;

        if( colon_index != -1 )
        {
            String arg_names = extendedReflectSignature.substring( colon_index + 1 );
            parNames = arg_names.split( "," );
            this.reflectSignature = extendedReflectSignature.substring( 0, colon_index );
        } else {
            this.reflectSignature = extendedReflectSignature;
        }

        buildParameterTypes( obtainExecutableMember() );
        buildPrettySignature();
    }

    public String getPrettySignature()
    {
        return prettySignature;
    }

    public String getDeclaringClass()
    {
        return declaringClass;
    }

    public String getReturnType()
    {
        return returnType;
    }

    public String[] getParTypes()
    {
        return parTypes;
    }

    public String[] getParNames()
    {
        return parNames;
    }

    protected final void buildParameterTypes( Class<?>[] par_types )
    {
        if( par_types.length == 0 )
        {
            parTypes = emptyPars;
            return;
        }

        parTypes = new String[ par_types.length ];
        for( int i = 0; i < par_types.length; i ++ )
        {
            parTypes[ i ] = Utils.prettifyClass( par_types[ i ] );
        }
    }

    protected final void appendParTypes( StringBuffer sb )
    {
        sb.append( '(' );
        if( parTypes.length > 0 )
        {
            sb.append( ' ' );
            for( int i = 0; i < parTypes.length; i ++ )
            {
                if( i != 0 )
                    sb.append( ", " );
                sb.append( parTypes[ i ] );
            }
            sb.append( ' ' );
        }
        sb.append( ')' );
    }

    protected abstract Class<?>[] obtainExecutableMember();
    protected abstract void buildPrettySignature();


    @Override
    public String toString()
    {
        return prettySignature;
    }
}
